/* -*-C-*-
 * ###################################################################
 *  occ++ - oc to C++ post-compiler for Esterel
 * 
 *  FILE: "genCxx.c"
 *                                    created: 03/09/92 00:00:00 
 *                                last update: 26/11/97 18:20:19 
 *  Author: Frdric Boulanger
 *  E-mail: Frederic.Boulanger@supelec.fr
 *    mail: Suplec - Service Informatique
 *          Plateau de Moulon, F-91192 Gif-sur-Yvette cedex
 *     www: http://wwwsi.supelec.fr/fb/fb.html
 *  
 * ========================================================================
 * Permission to use, copy and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that:
 *   - the above copyright notice appears in all copies and,
 *   - both the copyright notice and warranty disclaimer appear in
 *     supporting documentation and,
 *   - any product making use of this software or of parts of this 
 *     software displays the following acknowledgment:
 *       "This product includes software developed by Suplec
 *        and its contributors"
 *     in the documentation or other materials provided with the product 
 *     and in all advertising material mentioning features or use of 
 *     this software.
 *     
 * Warranty disclaimer: this software is provided "as is",
 *   Suplec and Frdric Boulanger disclaim all warranties with regard 
 *   to this software, including all implied warranties of merchantability 
 *   and fitness.
 *   In no event shall Suplec or Frdric Boulanger be liable for any 
 *   special, indirect or consequential damages or any damages whatsoever 
 *   resulting from loss of use, data or profits, whether in an action of 
 *   contract, negligence or other tortuous action, arising out of or in 
 *   connection with the use or performance of this software.
 * ========================================================================
 *  See header file for further information
 * ###################################################################
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "env.h"
#include "occxxErrors.h"
#include "occxxUtils.h"
#include "occxxTables.h"
#include "occxx.h"
#include "genCxx.h"


/* Prototypes of functions used and defined in this file. */
/* These functions create various pieces of code for the C++ output file. */

void generateInSigList(FILE *cFile);
void generateOutSigList(FILE *cFile);
#ifdef DEAD_CODE
void generateCopyOperator(FILE *cFile);
#endif
void generateBackup(FILE *cFile);
void generateWaitSigList(FILE *cFile);
void generateActionSwitch(FILE *cFile);
void generateAction(FILE *cFile, actInfo *act, unsigned long n);
void generateCall(FILE *cFile, callAction *act);
void generateProcCall(FILE *cFile, callAction *act);
void generateAssign(FILE *cFile, callAction *act);
void generateFunc(FILE *cFile, argNode *arg);
void generateCombine(FILE *cFile,combineAction *act);
void generateOpCall(FILE *cFile, char * name, args *arglist);
void generateIf(FILE *cFile,ifAction *act);
void generateStart(FILE *cFile, int execNum);
void generateKill(FILE *cFile, int execNum);
void generateSuspend(FILE *cFile, int execNum);
void generateActivate(FILE *cFile, int execNum);
void generateReturn(FILE *cFile, int execNum);

void generateArg(FILE *cFile, argNode *arg);
void generateValArg(FILE *cFile, argNode *arg);
void generateValArgs(FILE *cFile, args *arglist);
int generateRefArgs(FILE *cFile, args *arglist);


static char tmpStr[256];           /* String buffer */

static char **actionLabel;         /* Labels of actions for switch statements */

static int defIncluded = 0;        /* Flag for include of user definitions */


/**********************************************/
/* Static strings for include file generation */
/**********************************************/
/* Header of interface file */
/* Looks like 
  //
  //  foo.H
  //  Created by occ++ v4.0 on Wednesday September 16 1992 at 14:37:32 from oc v3 source file foo.oc
         
  #include <Esterel/libSync.h>"
*/
static char hHeader[] =
  "//\n//  %s\n//  Created by occ++ v%s\n//    on %s\n//    from oc v%d source file %s\n\
\n#include %s\n\n";

/* Header of a module in the interface file */
static char modHeader[] = 
     "//\n//  Interface of module %s\n//       root module: %s, from file: %s\n";

/* Info on a submodule in any file. */
static char subHeader[] =
     "//                   %s, from file: %s\n";

/* Include directive for the user definitions interface file */
static char IncludeDefs[] =
  "#include \"%s\"\n\n";

/* Declaration of the class representing a module */
static char classDef[] =
  "class %s : public Esterel {\n  protected:\n    enum {  // Automaton action numbers\n";

/* Text following the declaration of the action numbers */
static char endActNum[] =
    "    };\n\n\
    static const Esterel::tStateNumber *const *const c%sStates;\n";

static char dagsDecl[] =
    "    static const Esterel::tStateNumber *const *const c%sDags;\n";

static char privMethods[] =
    "\n\
    void doAction(tActionNumber op, Esterel::tActionNumber *i);\n\n";

/* Start of the public part of the class declaration */
static char startPublic[] =
  "  public:\n";

/* Text following the end of the public part */
static char endPublic[] =
  "    virtual void resetOutputs();\n\
    virtual void resetInputs();\n\
    virtual void setInputs();\n\
    virtual const InputSignal** inSigList() const;  // Returns list of input signals\n\
    virtual const OutputSignal** outSigList() const;// Returns list of output signals\n";

/* End of constructor and class */
static char endClass[] =
  "const char* name = \"%s\",\n        const bool schedule = true);  // Constructor\n\
    ~%s();  // Destructor\n";

/******************************************/
/* Static strings for C++ file generation */
/******************************************/
/* Header of the C++ file */
/* Similar to the header of the interface file */
static char cHeader[] =
  "//\n//  %s\n//  Created by occ++ v%s\n//    on %s\n//    from oc v%d source file %s\n//\n";

/* Header of a module in the C++ file */
static char cModHeader[] =
  "//\n//  Code of module %s\n//       root module: %s, from file: %s\n";

/* Include directive for the interface file */
static char IncludeH[] =
  "\n#include \"%s\"\n\n";

/* Definition of the action table of a dag */
static char dagOp[] =
  "static const Esterel::tStateNumber c%sDag%dops[] = {\n  ";

/* Definition of the action table of a dag in a separate file */
static char sepDagOp[] =
  "extern const Esterel::tStateNumber c%sDag%dops[] = {\n  ";

/* Definition of the action table of a dag as extern */
static char extDagOp[] =
  "extern const Esterel::tStateNumber c%sDag%dops[];\n\n";

/* Definition of the dag table of the state machine */
static char dagOps[] =
  "static const Esterel::tStateNumber *const c%sDagOps[] = {\n";

/* Definition of the dag table of the state machine as a class member */
/* This class member is initialized with the dagOps static table. */
static char tabDagDecl[] =
  "// Table of automaton dags\nconst Esterel::tStateNumber *const *const %s::c%sDags = c%sDagOps;\n\n";

/* Definition of the action table of a state */
static char stateOp[] =
  "static const Esterel::tStateNumber c%sState%dops[] = {\n  ";

/* Definition of the action table of a state in a separate file */
static char sepStateOp[] =
  "extern const Esterel::tStateNumber c%sState%dops[] = {\n  ";

/* Definition of the action table of a state as extern */
static char extStateOp[] =
  "extern const Esterel::tStateNumber c%sState%dops[];\n\n";

/* Definition of the action table of the state machine */
static char stateOps[] =
  "static const Esterel::tStateNumber *const c%sStateOps[] = {\n";

/* Definition of the action table of the state machine as a class member */
/* This class member is initialized with the stateOps static table. */
static char tabStateDecl[] =
  "// Table of automaton action numbers\nconst Esterel::tStateNumber *const *const %s::c%sStates = c%sStateOps;\n\n";

/* Definition of the class constructor */
static char constructorEnd[] =
  "const char* name, const bool schedule)    // Constructor\n\
        : Synchronous(name, schedule), Esterel(c%sStates";

/* Definition of the doAction member function. */
static char doActionStartDecl[] =
  "void %s::doAction(tActionNumber _op, Esterel::tActionNumber *_i) {\n";
static char actionSwitchHead[] =
  "  switch (_op) {\n    case kSequence:\n      break;\n\
    case kBranchAlways:\n      branchIfFalse(false,_i);\n      break;\n";

static char actionSwitchHead5[] =
  "  switch (_op) {\n    case kSequence:\n      break;\n\
    case kExecuteODAG:\n      executeOpenDAG(_i);\n      break;\n\
    case kGotoCDAG:\n      gotoClosedDAG(_i);\n      break;\n\
    case kBranchAlways:\n      branchIfFalse(false,_i);\n      break;\n";


/******************************************************************/
/* Procedure generateActLabels fills the action labels table with */
/* symbolic names of actions. The name of an action is built from */
/* what is does, sometime from the variable or signal it uses, and*/
/* from its number. The use of the action number guaranties the   */
/* unicity of each label. These labels are used to produce the    */
/* action switch of the automaton engine, and are declared in an  */
/* enum in the interface file of the module.                      */
/******************************************************************/
void generateActLabels() {
  int i,j = 0;
  
  if (ocVersion < 5)
    actionLabel = (char**)checkAlloc(nbActions+1,sizeof(char*));/* Allocate the table of labels */
  else
    actionLabel = (char**)checkAlloc(nbActions+3,sizeof(char*));/* Allocate the table of labels */
  actionLabel[j++] = "kSequence";                               /* 0 is always the sequence */
  if (ocVersion < 5)
    actionLabel[j++] = "kBranchAlways";                         /* 1 is always a jump in oc < 5 */
  else  {
    actionLabel[j++] = "kExecuteODAG";
    actionLabel[j++] = "kGotoCDAG";
    actionLabel[j++] = "kBranchAlways";
  }
  if (ocVersion < 5)
    i = 1;
  else
    i = 0;
  for ( ; i < nbActions; i++) {
  	char *tmp = 0;
  	int snum;
  	
    switch (actTable[i].type) {                                 /* Look at each action */
      case aGoto:                                               /* If it is a goto */
        sprintf(tmpStr,"kGoto_%d",i+incrAction);                /* call it kGoto_<act num> */
        break;
      case aAct:                                                /* Execution of another action */
        sprintf(tmpStr,"kAct_%d_%d",actTable[i].body.simple.sigNum, i+incrAction);
        break;
      case aPresent:                                            /* If it is a presence test */
        sprintf(tmpStr,"kPresent_%s_%d",sigTable[actTable[i].body.simple.sigNum].name,i+incrAction);
        break;                                                  /* call it kPresent_<signal>_<act num> */
      case aOutput:                                             /* If it is an output */
        sprintf(tmpStr,"kOutput_%s_%d",sigTable[actTable[i].body.simple.sigNum].name,i+incrAction);
        break;                                                  /* call it kOutput_<signal>_<act num> */
      case aCall:                                               /* If it is a proc call */
        sprintf(tmpStr,"kCall_%s_%d",procName(actTable[i].body.call.procNum),i+incrAction);
        break;                                                  /* call it kCall_<proc name>_<act num> */
      case aCombine:                                            /* If it is a value combination */
        sprintf(tmpStr,"kCombine_%s_%d",sigTable[actTable[i].body.combine.sigNum].name,i+incrAction);
        break;                                                  /* call it kCombine_<signal>_<act num> */
      case aReset:                                              /* If it is a variable reset */
        snum = actTable[i].body.simple.sigNum;
        switch(varTable[snum].kind) {
        	case vSigVal:      /* Reset of an initialized signal (Esterel >= 5.01) */
        	  tmp = sigTable[varTable[snum].sigNum].name;
       	    break;
       	  default:           /* Other cases (reset of a register is the only known case) */
       	    tmp = varTable[snum].name;
       	    break;
        }
        sprintf(tmpStr,"kReset_%s_%d",tmp,i+incrAction);
        break;                                                  /* call it kReset_<var name>_<act num> */
      case aIf:                                                 /* If it is a test */
        sprintf(tmpStr,"kIf_%d",i+incrAction);
        break;                                                  /* call it kIf_<act num> */
      case aDsz:                                                /* If it is a decr and jump if zero */
        sprintf(tmpStr,"kDsz_%d",i+incrAction);
        break;                                                  /* call it kDsz_<act num> */
      case aSeq:                                                /* If it is a sequence of actions */
        sprintf(tmpStr,"kSeq_%d",i+incrAction);                 /* call it kSeq_<act num> */
        break;
      case aStart:
        sprintf(tmpStr,"kStart_%s_t%d_e%d_%d",
                taskTable[execTable[actTable[i].body.simple.sigNum].task].name,
                execTable[actTable[i].body.simple.sigNum].task,
                actTable[i].body.simple.sigNum,
                i+incrAction);
        break;
      case aKill:
        sprintf(tmpStr,"kKill_%s_t%d_e%d_%d",
                taskTable[execTable[actTable[i].body.simple.sigNum].task].name,
                execTable[actTable[i].body.simple.sigNum].task,
                actTable[i].body.simple.sigNum,
                i+incrAction);
        break;
      case aSuspend:
        sprintf(tmpStr,"kSuspend_%s_t%d_e%d_%d",
                taskTable[execTable[actTable[i].body.simple.sigNum].task].name,
                execTable[actTable[i].body.simple.sigNum].task,
                actTable[i].body.simple.sigNum,
                i+incrAction);
        break;
      case aActivate:
        sprintf(tmpStr,"kActivate_%s_t%d_e%d_%d",
                taskTable[execTable[actTable[i].body.simple.sigNum].task].name,
                execTable[actTable[i].body.simple.sigNum].task,
                actTable[i].body.simple.sigNum,
                i+incrAction);
        break;
      case aReturn:
        sprintf(tmpStr,"kReturn_%s_t%d_e%d_%d",
                taskTable[execTable[actTable[i].body.simple.sigNum].task].name,
                execTable[actTable[i].body.simple.sigNum].task,
                actTable[i].body.simple.sigNum,
                i+incrAction);
        break;
      default:                                                  /* Unknow actions shouldn't happen */
        sprintf(tmpStr,"kUnknownAction_%d",i+incrAction);       /* call them kUnknownAction_<act num> */
    }
    actionLabel[i+incrAction] = StrDup(tmpStr);                 /* Copy the action name in the table */
  }
}

/***************************************************************/
/* Procedure generateHfile creates the interface file (.H) of  */
/* the compiled module. srcFile is the name of the oc source   */
/* file and hFileName is the name of the interface file. These */
/* global variables are defined in occxx.c                     */
/* nbModules is the number of already parsed modules.          */
/***************************************************************/
void generateHfile(int nbModules) {
  FILE *hFile;                     /* Output file */
  int i, j;
  int firstVar = 0;                /* Flag telling if we already met a variable */
  char* ifdefName;                 /* Name of macro to avoid multiple includes */
  
  if (nbModules == 0)
    hFile = fopen(hFileName,"w");  /* First module : create file */
  else
    hFile = fopen(hFileName,"a");  /* Not first : append to file */
  if (hFile == NULL) {             /* If unable to open output file, fail */
    sprintf(tmpStr,"Can't open %s for write access",hFileName);
    fileError(tmpStr);
  }
  if (nbModules == 0) {            /* First module : generate header */
    ifdefName = strrchr(hFileName, dirSep);
    if (ifdefName == 0) {
      ifdefName = checkAlloc(strlen(hFileName) + 3, sizeof(char));
      j = 0;
    }else {
      ifdefName++;
      j = ifdefName - hFileName;
      ifdefName = checkAlloc(strlen(ifdefName) + 3, sizeof(char));
    }
    ifdefName[0] = '_';
    for(i = j; i < strlen(hFileName); i++) {
      if (hFileName[i] == '.')
        ifdefName[i-j+1] = '_';
      else
        ifdefName[i-j+1] = toupper(hFileName[i]);
    }
    ifdefName[++i-j] = '_';
    ifdefName[++i-j] = 0;
    fprintf(hFile,"#ifndef %s\n", ifdefName);
    fprintf(hFile,"#define %s\n\n", ifdefName);
    free(ifdefName);
    
    if (optIncludeEsterel == 0)
      fprintf(hFile,hHeader, fileName(hFileName), occxxVers, dateAndTime(), 
                             ocVersion, fileName(sourceFile),includeEsterel);
    else
      fprintf(hFile,hHeader, fileName(hFileName), occxxVers, dateAndTime(), 
                             ocVersion, fileName(sourceFile),optIncludeEsterel);
  }
  fprintf(hFile,modHeader,moduleName,
          moduleTable[rootModule].name,moduleTable[rootModule].srcfile);
  for(i = 0; i < nbInstances; i++) { /* Put a header for the module with the name */
    if (i != rootModule) {           /* and source file of each submodule. */
      fprintf(hFile,subHeader, moduleTable[i].name,moduleTable[i].srcfile);
    }
  }
  fprintf(hFile,"\n");
#if 0
  if (((nbTypes != 0) || (nbFunctions != 0) || (nbProcedures != 0) || (nbExecs != 0))
        && (defIncluded == 0) && (noDefFile == 0)) {
    fprintf(hFile, IncludeDefs, hDefFileName); /* If there are user defined types */
    defIncluded = 1;                           /* include user definitions interface file */
  }
#endif

  if ((defIncluded == 0) && (needsExternalH() == 1)) {
		fprintf(hFile, IncludeDefs, hDefFileName);
		defIncluded = 1;
	}

  if ((nbConsts != 0) && (paramClass != 0) && !((nbConsts == 1) && (strcmp(constTable[0].name, "this") == 0))) {
    fprintf(hFile,"struct %sConsts {\n", moduleName);
    for(i = 0; i < nbConsts; i++) {
      if (strcmp(constTable[i].name, "this") != 0)
        fprintf(hFile, "  %s %s;\n", typeName(constTable[i].type),
                                     constTable[i].name);
    }
    fprintf(hFile,"};\n\n");
  }
  fprintf(hFile,classDef,moduleName);          /* Declare the class */
  if (ocVersion < 5)
    j = 1;
  else
    j = 4;
  for(i = 0; i < j; i++)  {
    fprintf(hFile,"      %s,\n",actionLabel[i]);
  }
  j--;
  for(i = 1; i < nbActions; i++) {             /* and the enum of action numbers */
    fprintf(hFile,"      %s,\n",actionLabel[i+j]);
  }                                            /* Last label is not followed by a comma. */
  fprintf(hFile,"      %s\n",actionLabel[nbActions+j]);
  fprintf(hFile,endActNum,moduleName);
  if (nbDags > 0)
    fprintf(hFile, dagsDecl,moduleName);

  if ((nbConsts != 0) && !((nbConsts == 1) && (strcmp(constTable[0].name, "this") == 0)))
    fprintf(hFile,"\n");
  for(i=0;i < nbConsts;i++) {    /* Declare constants */
    if (strcmp(constTable[i].name, "this") != 0) {
      fprintf(hFile,"    ");
      if (paramClass == 0)         /* If Constants are not object parameters */
        fprintf(hFile,"static ");  /* they are static members of the class */
      if (setParam == 0)           /* if they are not settable */
          fprintf(hFile,"const "); /* they are const */
      fprintf(hFile,"%s %s;\n",
                    typeName(constTable[i].type),
                    constTable[i].name);
    }
  }
  fprintf(hFile,privMethods);

  if (noTickReaction)
    fprintf(hFile, "    bool firstTick;  // This module does not react to 'tick'\n\n");
  
  for(i = 0;i < nbSignals; i++) {/* Declare Signal variables */
    switch (sigTable[i].kind) {                
      case sSensor:
      case sReturn:
      case sInput:               /* Input signals and sensors are implemented as InSignal<sig type> */
        if (sigTable[i].type == sPure) /* If signal is pure -> PureInSignal */
          fprintf(hFile,"    PureInSignal f%s;\n", sigTable[i].name);
        else
          fprintf(hFile,"    InSignal<%s> f%s;\n",
                  typeName(varTable[sigTable[i].varId].type),
                  sigTable[i].name);
        break;                   /* This yields "InSignal<int> fDial;" if Dial is an integer input signal. */
      case sInputOutput:         /* InputOutput signals are implemented as InOutSignal<sig type> */
        if (sigTable[i].type == sPure) /* If signal is pure -> PureInOutSignal */
          fprintf(hFile,"    PureInOutSignal f%s;\n", sigTable[i].name);
        else
          fprintf(hFile,"    InOutSignal<%s> f%s;\n",
                  typeName(varTable[sigTable[i].varId].type),
                  sigTable[i].name);
        break;                   /* This yields "InOutSignal<char> fKey;" if Key is a char inputoutput signal. */
      case sOutput:              /* Output signals are implemented as OutSignal<sig type> */
        if (sigTable[i].type == sPure) /* If signal is pure -> PureOutSignal */
          fprintf(hFile,"    PureOutSignal f%s;\n", sigTable[i].name);
        else
          fprintf(hFile,"    OutSignal<%s> f%s;\n",
                  typeName(varTable[sigTable[i].varId].type),
                  sigTable[i].name);
        break;                   /* This yields "OutSignal<float> fCost;" if Cost is a float output signal. */
      default:
        break;
    }
  }
  
  for(i = 0;i < nbVariables; i++) {     /* Declare module variables */
    switch (varTable[i].kind) {
      case vVarCnt:                     /* Variable is internal occurrence counter */
      case vVarVal:                     /* a user variable */
      case vVarWire:                    /* a boolean circuit wire */
      case vVarRegister:                /* a boolean circuit register */
      case vVarUnknown:                 /* or a variable of unknown use */
        if (firstVar == 0) {            /* If first variable, put a comment. */
          fprintf(hFile,"// Variables of module %s\n",moduleName);
          firstVar = 1;                   /* Remember you have seen a variable */
        }
        fprintf(hFile,"    %s %s;\n",typeName(varTable[i].type),varTable[i].name);
                                          /* Declare the variable. */
        break;
      case vSigVal:
        if ((sigTable[varTable[i].sigNum].kind == sLocal)
            &&(sigTable[varTable[i].sigNum].type != sPure)) {
          if (firstVar == 0) {              /* If first variable, put a comment. */
            fprintf(hFile,"// Variables of module %s\n",moduleName);
            firstVar = 1;                   /* Remember you have seen a variable */
          }
          fprintf(hFile,"    Sample<%s> locSig_%s_%d;  // Valued local signal.\n",
                       typeName(varTable[i].type),
                       sigTable[varTable[i].sigNum].name,i);
                        /* Variables for local valued signals must be declared. */
        }
        if ((sigTable[varTable[i].sigNum].kind == sException)
            &&(sigTable[varTable[i].sigNum].type != sPure)) {
          if (firstVar == 0) {              /* If first variable, put a comment. */
            fprintf(hFile,"// Variables of module %s\n",moduleName);
            firstVar = 1;                   /* Remember you have seen a variable */
          }
          fprintf(hFile,"    %s valExc_%s_%d;  // Valued exception.\n",typeName(varTable[i].type),
                        sigTable[varTable[i].sigNum].name,i);
                        /* Variables for valued exceptions must be declared. */
        }
        break;
      default:
        break;
    }
  }
  fprintf(hFile,"\n");
  
  if (firstVar > 0) {
    fprintf(hFile,"    class Data {\n");
    fprintf(hFile,"      public:\n");
    for(i = 0; i < nbVariables; i++) {
      switch (varTable[i].kind) {
        case vVarCnt:
        case vVarVal:
        case vVarWire:
        case vVarRegister:
        case vVarUnknown:
          fprintf(hFile,"        %s %s;\n",typeName(varTable[i].type),varTable[i].name);
          break;
        case vSigVal:
          if (sigTable[varTable[i].sigNum].type != sPure) {
            switch (sigTable[varTable[i].sigNum].kind) {
              case sLocal:
                fprintf(hFile, "        Sample<%s> locSig_%s_%d;\n",
                               typeName(varTable[i].type),
                               sigTable[varTable[i].sigNum].name, i);
                break;
              case sException:
                fprintf(hFile, "        %s valExc_%s_%d;\n",
                               typeName(varTable[i].type),
                               sigTable[varTable[i].sigNum].name, i);
                break;
              default:
                break;
            }
          }
          break;
        default:
          break;
      }
    }
    fprintf(hFile,"    } *pf_backup_copy_;\n");
  }
  
  if (nbExecs != 0) {
    fprintf(hFile,"\n// Execs of module %s\n",moduleName);
    for(i = 0; i < nbExecs; i++) {
      fprintf(hFile,"    %sTask_%s %s_%d_%d;\n",
                    moduleName,
                    taskTable[execTable[i].task].name,
                    taskTable[execTable[i].task].name,
                    execTable[i].task, i);
    }
    fprintf(hFile,"\n");
  }

  fprintf(hFile,"\n");
  fprintf(hFile,startPublic);           /* Public part of the class */
  for(i = 0;i < nbSignals; i++) {
    switch (sigTable[i].kind) {         /* Declare signal access methods */
      case sSensor:                     /* These methods bear the name of the signal they yield. */
      case sInput:
        if (sigTable[i].type == sPure) /* If signal is pure -> PureInSignal */
          fprintf(hFile,"    virtual PureInSignal& %s();\n", sigTable[i].name);
        else
          fprintf(hFile,"    virtual InSignal<%s>& %s();\n",typeName(varTable[sigTable[i].varId].type),sigTable[i].name);
        break;
      case sInputOutput:
        if (sigTable[i].type == sPure) /* If signal is pure -> PureInOutSignal */
          fprintf(hFile,"    virtual PureInOutSignal& %s();\n", sigTable[i].name);
        else
          fprintf(hFile,"    virtual InOutSignal<%s>& %s();\n",typeName(varTable[sigTable[i].varId].type),sigTable[i].name);
        break;
      case sOutput:
        if (sigTable[i].type == sPure) /* If signal is pure -> PureOutSignal */
          fprintf(hFile,"    virtual PureOutSignal& %s();\n", sigTable[i].name);
        else
          fprintf(hFile,"    virtual OutSignal<%s>& %s();\n",typeName(varTable[sigTable[i].varId].type),sigTable[i].name);
        break;
      default:
        break;
    }
  }
  fprintf(hFile,endPublic);
  if (hasAwaitedInfo && wantAwaitedList) /* If we have and want information on awaited signals, overload waitSigList() */
    fprintf(hFile,"    virtual const InputSignal** waitSigList() const;// Returns list of awaited signals\n");
  if (firstVar > 0) {
    fprintf(hFile, "    virtual void backup();\n");
    fprintf(hFile, "    virtual void restore();\n");
  }
#ifdef DEAD_CODE
  fprintf(hFile, "    %s& operator=(const %s&);\n", moduleName, moduleName);
#endif
  if ((nbConsts != 0) && (setParam != 0) && !((nbConsts == 1) && (strcmp(constTable[0].name, "this") == 0))) {
    fprintf(hFile,"    virtual void setParams(const %sConsts& newParams);\n", moduleName);
    fprintf(hFile,"    virtual void getParams(%sConsts& theParams) const;\n", moduleName);
  }
  if (noTickReaction) {
    fprintf(hFile, "    virtual void activate();   // This module does not react to ticks.\n");
  }
  fprintf(hFile,"    %s(",moduleName);  /* Constructor */
  if ((nbConsts != 0) && (paramClass != 0) && (setParam == 0) && !((nbConsts == 1) && (strcmp(constTable[0].name, "this") == 0))) {
    fprintf(hFile,"const %sConsts& initInstance,\n        ", moduleName);
  }
  fprintf(hFile, endClass, moduleName, moduleName);
  fprintf(hFile, "  private:\n    %s(const %s &);  // Do not use copy constructor\n",
                 moduleName, moduleName);
  fprintf(hFile,"};\n");
  
  fprintf(hFile,"\n\n");    /* Put an end to the file */
  fclose(hFile);            /* and close it */
}

/**********************************************************************/
/* Procedure generateCfile produces the C++ file for a module.        */
/**********************************************************************/
void generateCfile(int nbModules) {
  FILE *cFile;                       /* Output C++ file */
  FILE *tabFile;                     /* State table file */
  char tabFileName[256];             /* table file name */
  int i,j;
  int hasData = 0;                   /* Module has data to backup/restore */
  
  if (nbModules == 0)
    cFile = fopen(cFileName,"w");    /* First module : create file */
  else
    cFile = fopen(cFileName,"a");    /* Not first : append to file */
  if (cFile == NULL) {
    sprintf(tmpStr,"Can't open %s for write access",cFileName);
    fileError(tmpStr);
  }
  if (nbModules == 0) {              /* First module : generate header */
    fprintf(cFile,cHeader,fileName(cFileName), occxxVers,
            dateAndTime(), ocVersion, fileName(sourceFile));
    fprintf(cFile,IncludeH, fileName(hFileName));
  }
  fprintf(cFile,cModHeader,moduleName,
          moduleTable[rootModule].name,
          moduleTable[rootModule].srcfile);
  for(i = 0; i < nbInstances; i++) {           /* Put a header for the module with the name */
    if (i != rootModule) {                     /* and source file of each submodule. */
      fprintf(cFile,subHeader, moduleTable[i].name,moduleTable[i].srcfile);
    }
  }
  fprintf(cFile,"\n");

  if (nbDags > 0) {
    for(i = 0;i < nbDags; i++) {          /* Generate table of actions for each dag */
      if (separateTables) {
        fprintf(cFile,extDagOp,moduleName,i);
        sprintf(tabFileName,"%s%sDag%d.C",tabBaseFileName, moduleName,i);
        tabFile = fopen(tabFileName,"w");  /* Open table file */
        if (tabFile == NULL) {
          sprintf(tmpStr,"Can't open %s for write access",tabFileName);
          fileError(tmpStr);
        }
        fprintf(tabFile,"#include %s\n\n",classEsterel);
        fprintf(tabFile,sepDagOp,moduleName,i);
        for(j = 0; j < dagTable[i].length - 1; j++) {
          if ((cutLongLines != 0) && (j > 0) && ((j % cutLongLines) == 0))
            fprintf(tabFile,"\n  ");
          fprintf(tabFile,"%d,",dagTable[i].actions[j]);
        }
        if (dagTable[i].isOpenDAG == 0)
          fprintf(tabFile,"%d\n};\n\n",dagTable[i].actions[dagTable[i].length - 1]);
        else
          fprintf(tabFile,"%d,%d\n};\n\n",dagTable[i].actions[dagTable[i].length - 1],
                                          GotoAction);
        fclose(tabFile);
      } else {
        fprintf(cFile,dagOp,moduleName,i);
        for(j = 0; j < dagTable[i].length - 1; j++) {
          if ((cutLongLines != 0) && (j > 0) && ((j % cutLongLines) == 0))
            fprintf(cFile,"\n  ");
          fprintf(cFile,"%d,",dagTable[i].actions[j]);
        }
        if (dagTable[i].isOpenDAG == 0)
          fprintf(cFile,"%d\n};\n\n",dagTable[i].actions[dagTable[i].length - 1]);
        else
          fprintf(cFile,"%d,%d\n};\n\n",dagTable[i].actions[dagTable[i].length - 1],
                                        GotoAction);
      }
    }                                     /* Last action is a special case : must close brace */
    fprintf(cFile,dagOps,moduleName);     /* Create table of action tables */
    for(i = 0; i < nbDags - 1; i++) {
      fprintf(cFile,"  c%sDag%dops,\n",moduleName,i);
    }
    fprintf(cFile,"  c%sDag%dops\n};\n\n",moduleName,nbDags - 1);
                                      /* Last table is a special case : no comma, close brace */
  }                              

  for(i = 0;i < nbStates; i++) {          /* Generate table of actions for each state */
    if (separateTables) {
      fprintf(cFile,extStateOp,moduleName,i);
      sprintf(tabFileName,"%s%sTab%d.C",tabBaseFileName, moduleName,i);
      tabFile = fopen(tabFileName,"w");  /* Open table file */
      if (tabFile == NULL) {
        sprintf(tmpStr,"Can't open %s for write access",tabFileName);
        fileError(tmpStr);
      }
      fprintf(tabFile,"#include %s\n\n",classEsterel);
      fprintf(tabFile,sepStateOp,moduleName,i);
      for(j = 0; j < stateTable[i].length - 1; j++) {
        if ((cutLongLines != 0) && (j > 0) && ((j % cutLongLines) == 0))
          fprintf(tabFile,"\n  ");
        fprintf(tabFile,"%d,",stateTable[i].actions[j]);
      }
      fprintf(tabFile,"%d\n};\n\n",stateTable[i].actions[stateTable[i].length - 1]);
      fclose(tabFile);
    } else {
      fprintf(cFile,stateOp,moduleName,i);
      for(j = 0; j < stateTable[i].length - 1; j++) {
        if ((cutLongLines != 0) && (j > 0) && ((j % cutLongLines) == 0))
          fprintf(cFile,"\n  ");
        fprintf(cFile,"%d,",stateTable[i].actions[j]);
      }
      fprintf(cFile,"%d\n};\n\n",stateTable[i].actions[stateTable[i].length - 1]);
    }
  }                                       /* Last action is a special case : must close brace */
  fprintf(cFile,stateOps,moduleName);     /* Create table of action tables */
  for(i = 0; i < nbStates - 1; i++) {
    fprintf(cFile,"  c%sState%dops,\n",moduleName,i);
  }
  fprintf(cFile,"  c%sState%dops\n};\n\n",moduleName,nbStates - 1);
                                    /* Last table is a special case : no comma, close brace */
  if (allowDebug) {                 /* Table of action name for debug info */
    fprintf(cFile,"#ifdef _SYNC_DEBUG_\n");
    fprintf(cFile,"static const char* _%s_ActLabels[] = {\n", moduleName);
    if (ocVersion < 5)
      j = 1;
    else
      j = 4;
    for(i = 0; i < j; i++)  {
      fprintf(cFile,"  \"%s\",\n",actionLabel[i]);
    }
    j--;
    for(i = 1; i < nbActions; i++) {
      fprintf(cFile,"  \"%s\",\n",actionLabel[i+j]);
    }                                /* Last label is not followed by a comma. */
    fprintf(cFile,"  \"%s\"\n};\n",actionLabel[nbActions+j]);
    fprintf(cFile,"#endif\n\n");
  }

  if (nbDags > 0)
    fprintf(cFile,tabDagDecl,moduleName,moduleName,moduleName);
                                     /* Declare table of dag tables as a class member */
  fprintf(cFile,tabStateDecl,moduleName,moduleName,moduleName);
                                     /* Declare table of action tables as a class member */


  for (i = 0; i < nbSignals; i++) {  /* Generate code for access methods to signals */
    switch (sigTable[i].kind) {      /* These methods simply return a reference on the */
      case sSensor:                  /* corresponding member signal. */
      case sInput:
        if (sigTable[i].type == sPure)
          fprintf(cFile,"PureInSignal& %s::%s() {\n", moduleName,sigTable[i].name);
        else
          fprintf(cFile,"InSignal<%s>& %s::%s() {\n",typeName(varTable[sigTable[i].varId].type),moduleName,sigTable[i].name);
        fprintf(cFile,"  return f%s;\n}\n\n",sigTable[i].name);
        break;
      case sInputOutput:
        if (sigTable[i].type == sPure)
          fprintf(cFile,"PureInOutSignal& %s::%s() {\n", moduleName,sigTable[i].name);
        else
          fprintf(cFile,"InOutSignal<%s>& %s::%s() {\n",typeName(varTable[sigTable[i].varId].type),moduleName,sigTable[i].name);
        fprintf(cFile,"  return f%s;\n}\n\n",sigTable[i].name);
        break;
      case sOutput:
        if (sigTable[i].type == sPure)
          fprintf(cFile,"PureOutSignal& %s::%s() {\n", moduleName,sigTable[i].name);
        else
          fprintf(cFile,"OutSignal<%s>& %s::%s() {\n",typeName(varTable[sigTable[i].varId].type),moduleName,sigTable[i].name);
        fprintf(cFile,"  return f%s;\n}\n\n",sigTable[i].name);
        break;
      default:
        break;
    }
  }
  fprintf(cFile,"void %s::resetOutputs() {\n", moduleName);       /* Code for 'resetOutputs()' */
  if (allowDebug) {
    fprintf(cFile,"#ifdef _SYNC_DEBUG_\n");
    fprintf(cFile,"  cerr << name() << \".resetOuputs()\" << endl;\n");
    fprintf(cFile,"#endif\n");
  }
  for(i = 0; i < nbSignals; i++) {
    switch (sigTable[i].kind) {
      case sOutput:          /* Only output and inputoutput signals */
      case sInputOutput:     /* need to be reset. */
        fprintf(cFile,"  f%s.value().reset();\n", sigTable[i].name);
        break;
      case sLocal:
        if (sigTable[i].type != sPure) { /* Valued local signals need to be reset too */
          fprintf(cFile,"  locSig_%s_%d.reset();\n",
                  sigTable[i].name, sigTable[i].varId);
        }
        break;
      default:
        break;
    }
  }
  fprintf(cFile,"}\n\n");  /* Close function definition */

  fprintf(cFile,"void %s::resetInputs() {\n", moduleName); /* Code for 'resetInputs()' */
  if (allowDebug) {
    fprintf(cFile,"#ifdef _SYNC_DEBUG_\n");
    fprintf(cFile,"  cerr << name() << \".resetInputs()\" << endl;\n");
    fprintf(cFile,"#endif\n");
  }
  for(i = 0; i < nbSignals; i++) {
    switch (sigTable[i].kind) {
      case sSensor:
      case sInput:
      case sInputOutput:
      case sReturn:
        fprintf(cFile,"  f%s.value().reset();\n",sigTable[i].name);
        break;
      default:
        break;
    }
  }
  fprintf(cFile,"}\n\n");  /* Close function definition */

  fprintf(cFile,"void %s::setInputs() {\n", moduleName); /* Code for 'setInputs()' */
  if (allowDebug) {
    fprintf(cFile,"#ifdef _SYNC_DEBUG_\n");
    fprintf(cFile,"  cerr << name() << \".setInputs()\" << endl;\n");
    fprintf(cFile,"#endif\n");
  }
  if (nbExecs > 0)
    fprintf(cFile,"// Ask for the return status of the execs\n");
  for(i = 0; i < nbExecs; i++) {  /* Set execs return signals */
    fprintf(cFile,"  %s_%d_%d.setReturn();\n",
                  taskTable[execTable[i].task].name,
                  execTable[i].task, i);
  }
  
  for(i = 0; i < nbSignals; i++) {
    switch (sigTable[i].kind) {
      case sSensor:
      case sInput:
      case sReturn:
      case sInputOutput:  /* Only input and inputoutput signals need to update */
        fprintf(cFile,"  f%s.update();\n",sigTable[i].name);
        break;
      default:
        break;
    }
  }
  if (allowDebug) {
    fprintf(cFile,"#ifdef _SYNC_DEBUG_\n");
  fprintf(cFile,"#ifndef __GNUG__\n");
    for(i = 0; i < nbSignals; i++) {
      switch (sigTable[i].kind) {
        case sSensor:
        case sInput:
        case sReturn:
        case sInputOutput:
          fprintf(cFile,"  cerr << f%s << \"=\" << f%s.value() << endl;\n",sigTable[i].name,sigTable[i].name);
          break;
        default:
          break;
      }
    }
    fprintf(cFile,"#else\n// GNU g++ 2.5.8 requires ::operator<<(...) instead of ..<<.. for templates\n");
    for(i = 0; i < nbSignals; i++) {
      switch (sigTable[i].kind) {
        case sSensor:
        case sInput:
        case sReturn:
        case sInputOutput:
          fprintf(cFile,"  ::operator<<(cerr, f%s) << \"=\";\n",sigTable[i].name);
          fprintf(cFile,"  ::operator<<(cerr, f%s.value()) << endl;\n", sigTable[i].name);
          break;
        default:
          break;
      }
    }
    fprintf(cFile,"#endif\n");
    fprintf(cFile,"#endif\n");
  }
  fprintf(cFile,"}\n\n");

  generateInSigList(cFile);  /* Generate code for the 'inSigList()' method */
  generateOutSigList(cFile);  /* Generate code for the 'outSigList()' method */
  generateWaitSigList(cFile);  /* Generate code for the 'waitSigList()' method */

#ifdef DEAD_CODE
  generateCopyOperator(cFile); /* Code for operator=() */
#endif

  if ((nbConsts != 0) && (setParam != 0)       /* Code for setParams and getParams */
      && !((nbConsts == 1) && (strcmp(constTable[0].name, "this") == 0))) {
    fprintf(cFile,"void %s::setParams(const %sConsts& newParams) {\n", moduleName, moduleName);
    for(i = 0; i < nbConsts; i++) { /* set parameters value */
      if (strcmp(constTable[i].name, "this") != 0)
        fprintf(cFile,"  %s = newParams.%s;\n", constTable[i].name,
                                                constTable[i].name);
    }
    fprintf(cFile,"}\n\n");
    
    fprintf(cFile,"void %s::getParams(%sConsts& theParams) const {\n", moduleName, moduleName);
    for(i = 0; i < nbConsts; i++) { /* get parameters value */
      if (strcmp(constTable[i].name, "this") != 0)
        fprintf(cFile,"  theParams.%s = %s;\n", constTable[i].name,
                                                constTable[i].name);
    }
    fprintf(cFile,"}\n\n");
  }

  /* Overload activate for modules that do not react to the 'tick' signal' */
  if (noTickReaction) {
    fprintf(cFile, "void %s::activate()  {\n", moduleName);
    fprintf(cFile, "  bool someInputPresent = false;\n\n");
    fprintf(cFile, "  if (firstTick) {   // First tick is an initialization step.\n");
    fprintf(cFile, "    firstTick = false;\n");
    fprintf(cFile, "    Esterel::activate();\n");
    fprintf(cFile, "    return;\n");
    fprintf(cFile, "  }\n");
    
    for (i=0; i < nbSignals; i++) {
      switch (sigTable[i].kind) {            /* First, count input signals */
        case sSensor:
        case sInput:
        case sInputOutput:
        case sReturn:
          fprintf(cFile,"  someInputPresent = someInputPresent || f%s.value().isPresent();\n",sigTable[i].name);
          break;
        default:
        break;
      }
    }
    fprintf(cFile, "\n  if (someInputPresent)\n");
    fprintf(cFile, "    Esterel::activate();\n");
    fprintf(cFile, "}\n\n");
  }

                             /* Generate code for the constructor */
  fprintf(cFile,"%s::%s(", moduleName, moduleName);
  if ((nbConsts != 0) && (paramClass != 0) && (setParam == 0)
      && !((nbConsts == 1) && (strcmp(constTable[0].name, "this") == 0)))
    fprintf(cFile,"const %sConsts& initInstance,\n    ", moduleName);
  fprintf(cFile,constructorEnd, moduleName);
  if (nbDags > 0) {
    fprintf(cFile,", c%sDags)", moduleName);
  } else {
    fprintf(cFile,")");
  }
  
  for(i = 0; i < nbSignals; i++) {
    switch (sigTable[i].kind) {
      case sSensor:
      case sInput:
      case sOutput:
      case sReturn:
      case sInputOutput:
        fprintf(cFile,",\n          f%s(this, \"%s\")",sigTable[i].name,sigTable[i].name);
                     /* Initialization of signals with a pointer to their */
                     /* module, and their name (so they can say who they are) */
        break;
      default:
        break;
    }
  }
  for(i = 0; i < nbVariables; i++) {
    switch(varTable[i].kind) {
      case vVarRegister:
        hasData = 1;
        if (varTable[i].initValue != 0) {
          fprintf(cFile,",\n          %s(",varTable[i].name);
          generateArg(cFile,varTable[i].initValue);
          fprintf(cFile,")");
        }
        break;
      case vVarCnt:
      case vVarVal:
      case vVarWire:
      case vVarUnknown:
        hasData = 1;
        break;
      case vSigVal:
        if (hasData == 0) { /* Do not test if we know we have data to backup */
          if (sigTable[varTable[i].sigNum].type != sPure) {
            if ((sigTable[varTable[i].sigNum].kind == sLocal)
              ||(sigTable[varTable[i].sigNum].kind == sException)) {
              hasData = 1;
             }
          }
        }
        break;
      default:
        break;
    }
  }
  
  if ((nbConsts != 0) && (paramClass != 0) && (setParam == 0)
      && !((nbConsts == 1) && (strcmp(constTable[0].name, "this") == 0))) {
    for(i = 0; i < nbConsts; i++) { /* Initialise const members */
      if (strcmp(constTable[i].name, "this") != 0)
        fprintf(cFile,",\n          %s(initInstance.%s)", constTable[i].name,
                                                          constTable[i].name);
    }
  }
  if (nbExecs != 0) {
    for(i = 0; i < nbExecs; i++) {
      sprintf(tmpStr,"%s_%d_%d",taskTable[execTable[i].task].name,execTable[i].task, i);
      fprintf(cFile,",\n          %s_%d_%d(%d,%d,\"%s\")",
                    taskTable[execTable[i].task].name,
                    execTable[i].task, i,
                    execTable[i].task, i,
                    tmpStr);
    }
  }
  if (noTickReaction)
    fprintf(cFile, ",\n          firstTick(true)");
  
  if (hasData == 1)
    fprintf(cFile, ",\n          pf_backup_copy_(0)");
  
  fprintf(cFile,"\n{\n  gotoState(%d);    // Set initial state\n",initState);
                                  /* Code to set initial state of module */
  /* In Esterel v5, signals may have an initial value */
  for (i = 0; i < nbVariables; i++) {
    switch(varTable[i].kind) {
      case vSigVal:
        if (varTable[i].initValue != 0) {
          fprintf(cFile,"\n  f%s.value() = ",
                  sigTable[varTable[i].sigNum].name);
          generateArg(cFile,varTable[i].initValue);
          fprintf(cFile,";\n");
          fprintf(cFile,"  f%s.value().reset();\n",
                  sigTable[varTable[i].sigNum].name);
        }
        break;
      default:
      	break;
    }
  }
  if (nbExecs != 0) {
    for(i = 0; i < nbExecs; i++) {
      fprintf(cFile,"  f%s << %s_%d_%d.returnSig();\n",
                    sigTable[execTable[i].retsig].name,
                    taskTable[execTable[i].task].name,
                    execTable[i].task, i);
    }
  }
  fprintf(cFile,"}\n\n");
  
  fprintf(cFile,"%s::~%s() {\n", moduleName, moduleName);
  if (hasData == 1) {
    fprintf(cFile,"  if (pf_backup_copy_ != 0)\n");
    fprintf(cFile,"    delete pf_backup_copy_;\n");
  }
  fprintf(cFile, "}\n\n");
  
  if (hasData == 1)
    generateBackup(cFile);

  fprintf(cFile,doActionStartDecl,moduleName);
                                  /* Code of the action switch of the automaton */
  if (allowDebug) {
    fprintf(cFile,"#ifdef _SYNC_DEBUG_\n");
    fprintf(cFile,"  cerr << name() << \" in state #\" << fState << \", doing action #\" << (*_i) ");
    fprintf(cFile,"<< \" : \" << _op << \" (\" << _%s_ActLabels[_op] << \")\" << endl;\n", moduleName);
    fprintf(cFile,"#endif\n");
  }
  if (ocVersion < 5)
    fprintf(cFile,actionSwitchHead);
  else
    fprintf(cFile,actionSwitchHead5);
  generateActionSwitch(cFile);
  fprintf(cFile,"  }\n}\n\n\n");  /* Put a clean end to the file */
  fclose(cFile);                  /* and close it. */
}


/*********************************************************************/
/* Procedure generateInSigList produce the code for the inSigList()  */
/* method. This method returns a null terminated list of the input   */
/* signals of the module.                                            */
/*********************************************************************/
void generateInSigList(FILE *cFile) {
  int i, nbInSig = 0, inSigNum = 0;
  for (i=0; i < nbSignals; i++)
    switch (sigTable[i].kind) {            /* First, count input signals */
      case sSensor:
      case sInput:
      case sInputOutput:
/*      case sReturn: */                   /* Return signals should not appear outside the module */
        nbInSig++;
      default:
        break;
    }
  fprintf(cFile,"const InputSignal** %s::inSigList() const {\n",moduleName);  /* Method header */
  fprintf(cFile,"  const InputSignal** inList;\n");       /* Pointer to null terminated list */
  fprintf(cFile,"  inList = (const InputSignal**) new (InputSignal*[%d]);\n",nbInSig+1);/* Allocate the list */
  for (i=0; i < nbSignals; i++)                           /* Fill it */
    switch (sigTable[i].kind) {
      case sSensor:
      case sInput:
      case sInputOutput:
/*      case sReturn: */
        fprintf(cFile,"  inList[%d] = &f%s;\n",inSigNum,sigTable[i].name);
        inSigNum++;
        break;
      default:
        break;
    }
  fprintf(cFile,"  inList[%d] = 0;\n",nbInSig);             /* Terminate the list with a null pointer */
  fprintf(cFile,"  return inList;\n}\n\n");                 /* Return the list */
}

/***********************************************************************/
/* Procedure generateOutSigList produce the code for the outSigList()  */
/* method. This method returns a null terminated list of the input     */
/* signals of the module.                                              */
/***********************************************************************/
void generateOutSigList(FILE *cFile) {
  int i, nbOutSig = 0, outSigNum = 0;
  for (i=0; i < nbSignals; i++)
    switch (sigTable[i].kind) {            /* First, count output signals */
      case sOutput:
      case sInputOutput:
        nbOutSig++;
      default:
        break;
    }
  fprintf(cFile,"const OutputSignal** %s::outSigList() const {\n",moduleName);  /* Method header */
  fprintf(cFile,"  const OutputSignal** outList;\n");       /* Pointer to null terminated list */
  fprintf(cFile,"  outList = (const OutputSignal**) new (OutputSignal*[%d]);\n",nbOutSig+1);/* Allocate the list */
  for (i=0; i < nbSignals; i++)                     /* Fill it */
    switch (sigTable[i].kind) {
      case sOutput:
      case sInputOutput:
        fprintf(cFile,"  outList[%d] = &f%s;\n",outSigNum,sigTable[i].name);
        outSigNum++;
        break;
      default:
        break;
    }
  fprintf(cFile,"  outList[%d] = 0;\n",nbOutSig);   /* Terminate the list with a null pointer */
  fprintf(cFile,"  return outList;\n}\n\n");        /* Return the list */
}

/*********************************************************************/
/* Procedure generateCopyOperator produce the code for the           */
/* assignment operator and the backup() and restore() methods.       */
/*********************************************************************/
#ifdef DEAD_CODE
void generateCopyOperator(FILE *cFile) {
  int i;
  
  fprintf(cFile, "void %s::backup() {\n", moduleName);
  fprintf(cFile, "  if (pf_backup_copy_ == 0) {\n");
  if ((nbConsts != 0) && (paramClass != 0) && (setParam == 0)
      && !((nbConsts == 1) && (strcmp(constTable[0].name, "this") == 0))) {
    fprintf(cFile,"    %sConsts constParms;\n", moduleName);
    for(i = 0; i < nbConsts; i++) {
      if (strcmp(constTable[i].name, "this") != 0)
        fprintf(cFile,"    constParms.%s = %s;\n",
                      constTable[i].name,
                      constTable[i].name);
    }
    fprintf(cFile,"    pf_backup_copy_ = new %s(constParms,\"backup\",false);\n", moduleName);
  } else
    fprintf(cFile, "    pf_backup_copy_ = new %s(\"backup\",false);\n", moduleName);
  fprintf(cFile, "  }\n  *pf_backup_copy_ = *this;\n}\n\n");
  
  fprintf(cFile, "void %s::restore() {\n", moduleName);
  fprintf(cFile, "  if (pf_backup_copy_ != 0)\n");
  fprintf(cFile, "    *this = *pf_backup_copy_;\n}\n\n");
  
  fprintf(cFile, "%s& %s::operator=(const %s& m) {\n", moduleName, moduleName, moduleName);
  fprintf(cFile, "  fActTable = m.fActTable;\n");
  fprintf(cFile, "  fState = m.fState;\n");
  if ((nbConsts != 0) && (setParam != 0)
      && !((nbConsts == 1) && (strcmp(constTable[0].name, "this") == 0))) {
    fprintf(cFile,"  %sConsts constParms;\n", moduleName);
    fprintf(cFile,"  m.getParams(constParms);\n");
    fprintf(cFile,"  setParams(constParms);\n");
  }
  for(i = 0; i < nbVariables; i++) {
    switch (varTable[i].kind) {
      case vVarCnt:              /* Variable is internal occurrence counter */
      case vVarVal:
      case vVarWire:
      case vVarRegister:
      case vVarUnknown:
        fprintf(cFile,"  %s = m.%s;\n",varTable[i].name,varTable[i].name);
        break;
      case vSigVal:
        if ((sigTable[varTable[i].sigNum].kind == sLocal)
            &&(sigTable[varTable[i].sigNum].type != sPure)) {
          fprintf(cFile,"  locSig_%s_%d = m.locSig_%s_%d;\n",
                        sigTable[varTable[i].sigNum].name,i,
                        sigTable[varTable[i].sigNum].name,i);
        }
        if ((sigTable[varTable[i].sigNum].kind == sException)
            &&(sigTable[varTable[i].sigNum].type != sPure)) {
          fprintf(cFile,"  valExc_%s_%d = m.valExc_%s_%d;\n",
                        sigTable[varTable[i].sigNum].name,i,
                        sigTable[varTable[i].sigNum].name,i);
        }
        break;
      default:
        break;
    }
  }
  for(i = 0;i < nbSignals; i++) {
    switch (sigTable[i].kind) {                
      case sSensor:
      case sInput:
      case sInputOutput:
      case sOutput:
        fprintf(cFile,"  f%s.value() = ((%s&)m).f%s.value();\n",
                            sigTable[i].name,
                            moduleName,
                            sigTable[i].name);
        break;
      default:
        break;
    }
  }
  fprintf(cFile,"  return *this;\n}\n\n");
}
#endif

/*********************************************************************/
/* Procedure generateBackup produces the code for the backup()       */
/* and restore() methods.                                            */
/*********************************************************************/
void generateBackup(FILE *cFile) {
  int i;
  
  fprintf(cFile, "void %s::backup() {\n", moduleName);
  fprintf(cFile, "  Esterel::backup();\n");
  fprintf(cFile, "  if (pf_backup_copy_ == 0) {\n");
  fprintf(cFile, "    pf_backup_copy_ = new %s::Data;\n", moduleName);
  fprintf(cFile, "  }\n");
  for(i = 0; i < nbVariables; i++) {
    switch (varTable[i].kind) {
      case vVarCnt:
      case vVarVal:
      case vVarWire:
      case vVarRegister:
      case vVarUnknown:
        fprintf(cFile,"  pf_backup_copy_->%s = %s;\n",varTable[i].name,varTable[i].name);
        break;
      case vSigVal:
        if (sigTable[varTable[i].sigNum].type != sPure) {
          switch (sigTable[varTable[i].sigNum].kind) {
            case sLocal:
              fprintf(cFile, "  pf_backup_copy_->locSig_%s_%d = locSig_%s_%d;\n",
                      sigTable[varTable[i].sigNum].name, i,
                      sigTable[varTable[i].sigNum].name, i);
              break;
            case sException:
              fprintf(cFile, "  pf_backup_copy_->valExc_%s_%d = valExc_%s_%d;\n",
                      sigTable[varTable[i].sigNum].name, i,
                      sigTable[varTable[i].sigNum].name, i);
              break;
            default:
              break;
          }
        }
        break;
    }
  }
  fprintf(cFile, "}\n\n");

  fprintf(cFile, "void %s::restore() {\n", moduleName);
  fprintf(cFile, "  if (pf_backup_copy_ != 0) {\n");
  fprintf(cFile, "    Esterel::restore();\n");
  for(i = 0; i < nbVariables; i++) {
    switch (varTable[i].kind) {
      case vVarCnt:
      case vVarVal:
      case vVarWire:
      case vVarRegister:
      case vVarUnknown:
        fprintf(cFile,"    %s = pf_backup_copy_->%s;\n",varTable[i].name,varTable[i].name);
        break;
      case vSigVal:
        if (sigTable[varTable[i].sigNum].type != sPure) {
          switch (sigTable[varTable[i].sigNum].kind) {
            case sLocal:
              fprintf(cFile, "    locSig_%s_%d = pf_backup_copy_->locSig_%s_%d;\n",
                      sigTable[varTable[i].sigNum].name, i,
                      sigTable[varTable[i].sigNum].name, i);
              break;
            case sException:
              fprintf(cFile, "    valExc_%s_%d = pf_backup_copy_->valExc_%s_%d;\n",
                      sigTable[varTable[i].sigNum].name, i,
                      sigTable[varTable[i].sigNum].name, i);
              break;
            default:
              break;
          }
        }
        break;
    }
  }
  fprintf(cFile, "  }\n");
  fprintf(cFile, "}\n\n");
}

/*********************************************************************/
/* Procedure generateWaitSigList produce the code for the            */
/* waitSigList() method. This method returns a null terminated list  */
/* of the input signals of the module that are awaited in the        */
/* current state.                                                    */
/*********************************************************************/
void generateWaitSigList(FILE *cFile) {
  int i, state;
  
  if (hasAwaitedInfo && wantAwaitedList) { /* If we have and want information on awaited signals, overload waitSigList() */
    fprintf(cFile,"const InputSignal** %s::waitSigList() const {\n",moduleName);  /* Method header */
    fprintf(cFile,"  const InputSignal** waitList;\n\n");  /* Pointer to null terminated list */
    fprintf(cFile,"  switch(fState) {\n");
    for(state=0; state < nbStates; state++) {
      fprintf(cFile,"    case %d:\n", state);
      fprintf(cFile,"      waitList = (const InputSignal**) new (InputSignal*[%d]);\n",
              stateTable[state].nbAwaited+1);
      for(i=0; i < stateTable[state].nbAwaited; i++)
        fprintf(cFile,"      waitList[%d] = &f%s;\n",
                i, sigTable[stateTable[state].awaited[i]].name);
      fprintf(cFile,"      waitList[%d] = 0;\n      break;\n", i);
    }
    fprintf(cFile,"  }\n");
    fprintf(cFile,"  return waitList;\n}\n\n");       /* Return the list */
  }
}

/*********************************************************************/
/* Procedure generateActionSwitch produces the code for the automaton*/
/* engine of the module which interprets action codes. This code is  */
/* a switch statement driven by the action code read from the current*/
/* state action table.                                               */
/*********************************************************************/
void generateActionSwitch(FILE *cFile) {
  int i,j;

  /* At this point, the "switch(_op){" and the first two cases have been */
  /* written to the file. We have only to handle the changing part of the switch. */
  if (ocVersion < 5)
    j = 1;
  else
    j = 3;
  for(i = j+1 ; i < nbActions + incrAction; i++) {
    fprintf(cFile,"    case %s:\n",actionLabel[i]);
    generateAction(cFile,&(actTable[i-incrAction]), i-incrAction);
    fprintf(cFile,"      break;\n");
  }
}

/*******************************************************************/
/* Procedure generateAction produces the code for an action        */
/* This code is a switch statement driven by the action code read  */
/* from the current state action table.                            */
/*******************************************************************/
void generateAction(FILE *cFile, actInfo *act, unsigned long n) {
  switch (act->type) {
    case aGoto:            /* a goto is handled by the state machine */
      break;               /* by the activate() method of class Esterel. */
    case aAct:             /* Execute another action */
      fprintf(cFile,"      doAction(%s, _i);\n", actionLabel[act->body.simple.sigNum + incrAction]);
      break;
    case aPresent:
      fprintf(cFile,"      branchIfFalse(f%s.value().isPresent(), _i);\n",
                         sigTable[act->body.simple.sigNum].name);
      break;               /* A test is implemented by the branchIfFalse() */
                           /* method of class Esterel. */
    case aOutput:          /* If output of a pure signal, we need a fake assignement to set present to true */
      if (sigTable[act->body.simple.sigNum].type == sPure) {
        fprintf(cFile,"      f%s.value() = true;\n", sigTable[act->body.simple.sigNum].name);
      }
      break;
    case aCall:
      generateCall(cFile,&(act->body.call));
      break;               /* For a proc call, we produce the calling code. */
    case aCombine:
      generateCombine(cFile,&(act->body.combine));
      break;               /* Similar to case aCall: */
    case aReset:
      break;               /* ???. */
    case aIf:
      generateIf(cFile,&(act->body.ifAct));
      break;               /* A if is like a present, but with a more complex condition. */
    case aDsz:
      fprintf(cFile,"      branchIfFalse((-- %s <= 0), _i);\n",
                         varTable[act->body.simple.sigNum].name);
      break;               /* A test is implemented by the branchIfFalse() */
                           /* method of class Esterel. */
    case aSeq:             /* A sequence of actions */
      while(act != 0) {
        generateAction(cFile, act->body.sequence.action, n);
        act = act->body.sequence.next;
      }
      break;
    case aStart:
      generateStart(cFile, act->body.simple.sigNum);
      break;
    case aKill:
      generateKill(cFile, act->body.simple.sigNum);
      break;
    case aSuspend:
      generateSuspend(cFile, act->body.simple.sigNum);
      break;
    case aActivate:
      generateActivate(cFile, act->body.simple.sigNum);
      break;
    case aReturn:
      generateReturn(cFile, act->body.simple.sigNum);
      break;
    default:               /* If we don't know how to implement the action */
      fclose(cFile);       /* stop everything */
      sprintf(tmpStr,"Unknown action type %d for action %ld.",act->type, n);
      intError(tmpStr);    /* and say something went wrong. */
      break;
  }
}

/*****************************************************************/
/* Procedure generateCall produces the code for a procedure call.*/
/* It checks for a leading '_' in the procedure name, and if so, */
/* produces an assignement rather than a call. See generateAssign*/
/* for details.                                                  */
/*****************************************************************/
void generateCall(FILE *cFile, callAction *act) {
  if (procName(act->procNum)[0] == '_')    /* If proc name has a leading '_' */
    generateAssign(cFile, act);            /* it is an assignement. */
  else
    generateProcCall(cFile, act);          /* else it is a regular proc call. */
}

/*********************************************************************/
/* Procedure generateProcCall produces the code for a procedure call */
/* assuming it is really a procedure call (not an assignement). See  */
/* generateCall for details on procedure calls and assignements.     */
/*********************************************************************/
void generateProcCall(FILE *cFile, callAction *act) {
  fprintf(cFile,"      %s(",procName(act->procNum));
  if (
      generateRefArgs(cFile,act->refArgList) /* If there are reference arguments */
      &&(
         (act->valArgList != 0)&&(act->valArgList->node != 0) /* and value arguments */
        )
     ) {
    fprintf(cFile,",");                     /* put a comma */
  }
  generateValArgs(cFile,act->valArgList);   /* followed by by value arguments. */
  fprintf(cFile,");\n");
}

/****************************************************************************/
/* Procedure generateAssign handles assignements. Assignements are in fact  */
/* procedure calls, but assignement procedure have a leading '_' to their   */
/* name. This is detected by the generateCall procedure which calls         */
/* generateAssign in this case. If a variable is assigned to, generateAssign*/
/* produces an assignement statement with infix = operator. This allows     */
/* overloading of the = operator to work fine with occxx generated code.    */
/* If the assignement pertains to a signal, the value() method of the       */
/* signal is activated. In order to detect assignements to signal values, we*/
/* must check for a cSigVal type of argument or a cVarVal pertaining to the */
/* variable storing the value of a signal.                                  */
/****************************************************************************/
void generateAssign(FILE *cFile, callAction *act) {
  argNode *assignedTo;
  int resetDone = 0;
  
  if (act->refArgList == 0) {
    fclose(cFile);
    fatalError("Assignement to nothing!");
  }
  assignedTo = act->refArgList->node;  /* Get the thing that receives a value */
  if (
    (assignedTo->type == cVarVal)
  &&(
      (varTable[assignedTo->value].kind==vSigBool)
    ||
      (varTable[assignedTo->value].kind==vSigVal)
    )
  ) {   /* Assignement to a var that holds a signal */
    switch(varTable[assignedTo->value].kind) {
      case vSigVal:   /* Holds signal value */
        if (sigTable[varTable[assignedTo->value].sigNum].type == sPure) {
          sprintf(tmpStr,"Attempt to assign a value to pure signal %s.",
                  sigTable[varTable[assignedTo->value].sigNum].name);
          fclose(cFile);
          fatalError(tmpStr);
        }
        switch (sigTable[varTable[assignedTo->value].sigNum].kind) {
          case sLocal:
            fprintf(cFile,"      locSig_%s_%d = ", /* assignement to var holding value of signal. */
            sigTable[varTable[assignedTo->value].sigNum].name, assignedTo->value);
            break;
          case sException:
            fprintf(cFile,"      valExc_%s_%d = ", /* assignement to var holding value of exception */
                    sigTable[varTable[assignedTo->value].sigNum].name,assignedTo->value);
            break;
          default:
            fprintf(cFile,"      f%s.value() = ", /* assignement to the value of the signal. */
                    sigTable[varTable[assignedTo->value].sigNum].name);
            break;
        }
        break;
      case vSigBool:   /* Holds signal boolean */
      /* Only 'false' can be assigned to the boolean of a signal, which means 'reset' */
        if (!((act->valArgList->node->type == cSysConstVal)&&(act->valArgList->node->value == 0))) {
          sprintf(tmpStr,"Assignment of something else than 'false' to the boolean of signal %s.",
                  sigTable[varTable[assignedTo->value].sigNum].name);
          fclose(cFile);
          fatalError(tmpStr);
        }
      /* Assignement to the boolean of a signal can only be a reset of the signal */
        resetDone = 1;
        switch (sigTable[varTable[assignedTo->value].sigNum].kind) {
          case sLocal:
            fprintf(cFile,"      locSig_%s_%d.reset()", /* assignement to boolean = reset. */
                    sigTable[varTable[assignedTo->value].sigNum].name,
                    sigTable[varTable[assignedTo->value].sigNum].varId); /* %d is nb of 'val' variable */
            break;
          case sException:
            sprintf(tmpStr,"Attempt to assign a value to the boolean of exception %s.",
                    sigTable[varTable[assignedTo->value].sigNum].name);
            fclose(cFile);
            fatalError(tmpStr);
            break;
          default:
            fprintf(cFile,"      f%s.value().reset()",
                    sigTable[varTable[assignedTo->value].sigNum].name);
            break;
        }
        break;
      default:
        break;
    }
  }else {                  /* If it is a regular assignement */
    fprintf(cFile,"      ");
    generateValArgs(cFile,act->refArgList);
                           /* Assign to ref args of assign proc */
    fprintf(cFile," = ");
  }
  if (!resetDone)
    generateValArgs(cFile,act->valArgList); /* Assigned value is val arg of assign proc. */
  fprintf(cFile,";\n");
}

/*********************************************************************/
/* Procedure generateFunc produces code for a function call.         */
/* It checks for a leading '_' in the function name. Function name   */
/* having a leading '_' are in fact operators. In this case,         */
/* generateFunc produces an operator statement. This allows operator */
/* overloading to work fine with occxx produced code. In the case    */
/* of a regular function, a regular function call is produced.       */
/*********************************************************************/
void generateFunc(FILE *cFile, argNode *arg) {
	char *p;
	
  if (funcName(arg->value)[0] == '_'){  /* Leading '_' identifies an operator */
    if ((funcArity(arg->value) == 3)
        && (strcmp(opFromTag(funcName(arg->value)), "?:") == 0)) {
      fprintf(cFile,"(");      /* Operator bool ? Yes : No */
      generateValArg(cFile,arg->arguments->node);
      fprintf(cFile,") ? (");
      generateValArg(cFile,arg->arguments->next->node);
      fprintf(cFile,") : (");
      generateValArg(cFile,arg->arguments->next->next->node);
      fprintf(cFile,")");
      return;
    }
    if (funcArity(arg->value) == 1) {   /* If it is a unary operator */
      if (strcmp(opFromTag(funcName(arg->value)), "(--)<=0") == 0) { /* dsz op */
        fprintf(cFile,"((--");
        generateValArg(cFile,arg->arguments->node);
        fprintf(cFile,") <= 0)");
        return;
      }
      fprintf(cFile,"%s",opFromTag(funcName(arg->value)));
                                        /* Write the operator name */
      generateValArg(cFile, arg->arguments->node);/* followed by its argument*/
    }else {                     /* We have nothing else than binary operators */
      if (strcmp(opFromTag(funcName(arg->value)), "|||") == 0) {
        fprintf(cFile,"EvalBoth_OR(");
        generateValArg(cFile,arg->arguments->node);
        fprintf(cFile,",");
        generateValArg(cFile,arg->arguments->next->node);
        fprintf(cFile,")");
        return;
      }
      if (strcmp(opFromTag(funcName(arg->value)), "&&&") == 0) {
        fprintf(cFile,"EvalBoth_AND(");
        generateValArg(cFile,arg->arguments->node);
        fprintf(cFile,",");
        generateValArg(cFile,arg->arguments->next->node);
        fprintf(cFile,")");
        return;
      }
      fprintf(cFile,"(");
      generateValArg(cFile,arg->arguments->node);   /* Write first argument */
      fprintf(cFile,"%s",opFromTag(funcName(arg->value))); /* operator name */
      generateValArg(cFile,arg->arguments->next->node);/* and second argument*/
      fprintf(cFile,")");  /* The parenthesis take over a possibly misleading */
                           /* associativity or priority conflict with */
                           /* surrounding operators. */
    }
/* mangled named to C++ operator stuff */
  } else if (strncmp(funcName(arg->value),"CC_op_",6) == 0) {/* It's a C++ operator */
    generateOpCall(cFile,funcName(arg->value),arg->arguments);
  } else if (strncmp(funcName(arg->value),"CC_cons_",8) == 0) { /* Constructor */
    fprintf(cFile,"%s(",funcName(arg->value) + 8);
    generateValArgs(cFile,arg->arguments);/* followed by the list of arguments*/
    fprintf(cFile,")");
  } else if (strncmp(funcName(arg->value),"CC_arrow_",9) == 0) {/* -> member access */
  	if ((arg->arguments->node->type != cConstVal)
  	    ||(strcmp(constTable[arg->arguments->node->value].name,"this") != 0)) {
  	  generateValArg(cFile,arg->arguments->node);
  	  fprintf(cFile,"->");
  	}
  	p = funcName(arg->value) + 9;
  	p = strchr(p,'_');
  	if (p == 0) {
  		sprintf(tmpStr,"Bad member access operator name: \"%s\"",
  		               funcName(arg->value));
  		fatalError(tmpStr);
  	}
  	p++;
  	fprintf(cFile,"%s",p);
  	if (arg->arguments->next != 0) {
  		fprintf(cFile,"(");
  		generateValArgs(cFile,arg->arguments->next);
  		fprintf(cFile,")");
  	}
  } else if (strncmp(funcName(arg->value),"CC_dot_",7) == 0) {/* . member access */
 	  generateValArg(cFile,arg->arguments->node);
 	  fprintf(cFile,".");
  	p = funcName(arg->value) + 7;
  	p = strchr(p,'_');
  	if (p == 0) {
  		sprintf(tmpStr,"Bad member access operator name: \"%s\"",
  		               funcName(arg->value));
  		fatalError(tmpStr);
  	}
  	p++;
  	fprintf(cFile,"%s",p);
  	if (arg->arguments->next != 0) {
  		fprintf(cFile,"(");
  		generateValArgs(cFile,arg->arguments->next);
  		fprintf(cFile,")");
  	}
/* end of special stuff */
  } else { /* If we get here, we have a plain vanilia function */
    fprintf(cFile,"%s(",funcName(arg->value));  /* Write function name */
    generateValArgs(cFile,arg->arguments);/* followed by the list of arguments*/
    fprintf(cFile,")");
  }
}

/*********************************************************************/
/* Procedure generateOpCall produces code for a C++ operator call.   */
/*********************************************************************/
void generateOpCall(FILE *cFile, char * name, args *arglist) {
	char buf[255];
	char opname[15];
	char *p,*q;
	char *op;
	int i;
	args *a;
	
  strcpy(buf, name);
  p = buf+6;           /* Skip CC_op_ */
  q = strchr(p,'_');   /* Look for next '_' to get type name */
  if (q == 0) {
    sprintf(tmpStr,"\"%s\" begins with the reserved prefix 'CC_op_' but is not a valid operator name.",buf);
    fclose(cFile);
    fatalError(tmpStr);
  }
  *q = 0;       /* Now, p is the type name */
  for (i = -nbSysTypes; i < nbTypes; i++) {
    if (strcmp(p, typeName(i)) == 0) {
      break;
    }
  }
  *q = '_';            /* restore op name */
  if (i == nbTypes) {  /* We did not find a type */
    sprintf(tmpStr,"\"%s\" operator refers to type %s which is unknown.",buf,p);
    fclose(cFile);
    fatalError(tmpStr);
  }
  p = q+1;          /* Skip '_' to get op tag */
  op = opFromName(p);
  if (op == 0) {    /* Not a valid operator name */
    sprintf(tmpStr,"\"%s\" operator refers to operator %s which is unknown.",buf,p);
    fclose(cFile);
    fatalError(tmpStr);
  }
    
  strcpy(opname,op); /* Local copy of operator name to avoid writing into table */
  op = opname;
    
 	while((p = strchr(op,'#')) != 0) {      /* Look for the next argument */
 	  *p = 0;
 	  fprintf(cFile,"%s",op);
 	  *p = '#';
 	  p++;
 	  a = arglist;
 	  for (i = 0; i < (*p - '1'); i++) {         /* Go to nth argument */
 	  	a = a->next;
 	  }
 	  generateValArg(cFile,a->node);       /* generate code for this arg */
 	  op = p+1;     /* Next chunk of operator syntax */
 	}
 	if (*op != 0) {
    fprintf(cFile,"%s",op);  /* Write last chunk of operator syntax */
  }
}

/******************************************************************************/
/* Procedure generateCombine produces the code for a combine action.          */
/* A combine action occurs for multiple emmision of a valued signal.          */
/* The emmited values must be combined by a specified function.               */
/* The produced code checks to see if the signal already has a value. If not, */
/* the new value is merely affected to the signal. If the signal already has a*/
/* value, the combination function is called with the old value and the new   */
/* one and the result is affected to the signal.                              */
/******************************************************************************/
void generateCombine(FILE *cFile,combineAction *act) {
  char *sigName;
  
  sigName = sigTable[act->sigNum].name;        /* Get the name of the signal */
  if (sigTable[act->sigNum].kind == sLocal) {  /* Local signals are just samples */
    fprintf(cFile,"      if (locSig_%s_%d.isPresent())\n",sigName, sigTable[act->sigNum].varId);
                                                 /* Create the presence test */
    fprintf(cFile,"        locSig_%s_%d = %s(locSig_%s_%d.value(),",
                                          /* and the call of the combination */
                  sigName, sigTable[act->sigNum].varId,   /* function in the 'then' bloc. */
                  funcName(sigTable[act->sigNum].combFunc),sigName,
                  sigTable[act->sigNum].varId);
  } else {
    fprintf(cFile,"      if (f%s.value().isPresent())\n",sigName);
                                                 /* Create the presence test */
    fprintf(cFile,"        f%s.value() = %s(f%s.value().value(),",
                                          /* and the call of the combination */
                  sigName,                /* function in the 'then' bloc. */
                  funcName(sigTable[act->sigNum].combFunc),sigName);
  }
  generateValArg(cFile,act->oneArg);
  fprintf(cFile,");\n");
  fprintf(cFile,"      else\n");           /* Create a simple assignement in the 'else' bloc.*/
  if (sigTable[act->sigNum].kind == sLocal) {  /* Local signals are just samples */
    fprintf(cFile,"        locSig_%s_%d = ",sigName, sigTable[act->sigNum].varId);
  } else {
    fprintf(cFile,"        f%s.value() = ",sigName);
  }
  generateValArg(cFile,act->oneArg);
  fprintf(cFile,";\n");
}

/****************************************************************/
/* Procedure generateIf produces code for a conditional action. */
/* It consists merely in a call to branchIfFalse with the       */
/* condition (produced by generateArg) as argument.             */
/****************************************************************/
void generateIf(FILE *cFile,ifAction *act) {
  fprintf(cFile,"      branchIfFalse((");  /* Create a call to branchIfFalse() */
  generateValArg(cFile,act->oneArg);    /* with the condition as argument. */
  fprintf(cFile,"), _i);\n");           /* variable '_i' holds a pointer to the */
}                             /* current action number. see Esterel::activate() */

/*******************************************************************/
/* Procedure generateStart produces code for the start of an exec. */
/*******************************************************************/
void generateStart(FILE *cFile, int execNum) {
  fprintf(cFile, "      %s_%d_%d.tstart(",
                  taskTable[execTable[execNum].task].name,
                  execTable[execNum].task,
                  execNum);
    if (
        generateRefArgs(cFile,execTable[execNum].refArgList) /* If there are reference arguments */
        &&(
           (execTable[execNum].valArgList != 0)
         &&(execTable[execNum].valArgList->node != 0) /* and value arguments */
          )
       ) {
      fprintf(cFile,",");                     /* put a comma */
    }
  generateValArgs(cFile, execTable[execNum].valArgList);
  fprintf(cFile, ");\n");
}

/********************************************************************/
/* Procedure generateKill produces code for the killing of an exec. */
/********************************************************************/
void generateKill(FILE *cFile, int execNum) {
  fprintf(cFile, "      %s_%d_%d.tkill();\n",
                  taskTable[execTable[execNum].task].name,
                  execTable[execNum].task,
                  execNum);
}

/***************************************************************/
/* Procedure generateSuspend produces code to suspend an exec. */
/***************************************************************/
void generateSuspend(FILE *cFile, int execNum) {
  fprintf(cFile, "      %s_%d_%d.tsuspend();\n",
                  taskTable[execTable[execNum].task].name,
                  execTable[execNum].task,
                  execNum);
}

/****************************************************************/
/* Procedure generateActivate produces code to activate an exec. */
/****************************************************************/
void generateActivate(FILE *cFile, int execNum) {
  fprintf(cFile, "      %s_%d_%d.tactivate();\n",
                  taskTable[execTable[execNum].task].name,
                  execTable[execNum].task,
                  execNum);
}

/***********************************************************************/
/* Procedure generateReturn produces code for the return from an exec. */
/***********************************************************************/
void generateReturn(FILE *cFile, int execNum) {
  fprintf(cFile, "      %s_%d_%d.treturn();\n",
                  taskTable[execTable[execNum].task].name,
                  execTable[execNum].task,
                  execNum);
}

/***************************************************************************/
/* Procedure generateArg produces the C++ form of the argument described   */
/* by the argNode arg. If this argument is a variable, we must look inside */
/* to see if it is really a variable or if it is the storage of a signal.  */
/* If it is a function call, a recursive call will occur to produce the    */
/* argument list of the function.                                          */
/***************************************************************************/
void generateArg(FILE *cFile, argNode *arg) {
  switch (arg->type) {                             /* Look at the argument */
    case cVarVal:                                  /* If it is a variable */
      switch (varTable[arg->value].kind) {         /* look closer... */
        case vSigBool:                    /* If it is the boolean of a signal */
          fprintf(cFile,"f%s.value().isPresent()", /* get it with isPresent() */
            sigTable[varTable[arg->value].sigNum].name);
          break;
        case vSigVal:                       /* If it is the value of a signal */
          if (sigTable[varTable[arg->value].sigNum].type == sPure) {
            sprintf(tmpStr,"Attempt to get the value of pure signal %s",
                    sigTable[varTable[arg->value].sigNum].name);
            fclose(cFile);
            fatalError(tmpStr);
          }
          switch (sigTable[varTable[arg->value].sigNum].kind) {
            case sLocal:
              fprintf(cFile,"locSig_%s_%d",   /* stored in sample locSig_name */
                    sigTable[varTable[arg->value].sigNum].name,arg->value);
              break;
            case sException:
              fprintf(cFile,"valExc_%s_%d",      /* stored in var valExc_name */
                    sigTable[varTable[arg->value].sigNum].name, arg->value);
              break;
            default:
              fprintf(cFile,"f%s.value()",    /* get it with value() */
                    sigTable[varTable[arg->value].sigNum].name);
              break;
          }
          break;
        case vVarVal:                       /* If it is really a variable */
        case vVarCnt:                       /* or an occurrence counter */
        case vVarWire:                      /* or a boolean circuit wire */
        case vVarRegister:                  /* or a boolean circuit register */
        case vVarUnknown:                   /* or a variable of unknown use */
          fprintf(cFile,"%s",varTable[arg->value].name);
          break;                          /* just get it by its name */
        default:                            /* Should never happen. */
          sprintf(tmpStr,"Unknown variable type %d for variable %d",
                              varTable[arg->value].kind,arg->value);
          fclose(cFile);
          intError(tmpStr);
      }
      break;
    case cSysVal:                      /* We know about no system values */
      break;
    case cConstVal:                    /* Constants are known by their names */
      fprintf(cFile,"%s",constTable[arg->value].name);
      break;
    case cSysConstVal:                 /* System constants too... */
      fprintf(cFile,"%s",sysConstTable[arg->value].name);
      break;
    case cImmedVal:                /* Immediate values are immediate for C++ */
      fprintf(cFile,"%d",arg->value); /* because we only know about integers. */
      break;
    case cStringVal:                 /* String values are litteral strings */
      fprintf(cFile,"%s",(char*)arg->ptrValue);
      break;
    case cSigVal:                    /* value() yields the value of a signal. */
      if (sigTable[arg->value].kind == sLocal) /* Local signals are samples */
        fprintf(cFile,"locSig_%s_%d", sigTable[arg->value].name, sigTable[arg->value].varId);
      else
        fprintf(cFile,"f%s.value()", sigTable[arg->value].name);
      break;
    case cFuncVal:                   /* We get function values through a call */
      generateFunc(cFile, arg);
      break;
    default:                         /* Should never happen! */
      sprintf(tmpStr,"Unknown argument type %d.",arg->type);
      fclose(cFile);
      intError(tmpStr);
  }
}

/*********************************************************************/
/* Procedure generateValArg produces code for a 'by value' argument. */
/* Signals are special because their value is a Sample. Esterel      */
/* understands their value as the value of the sample. So we need    */
/* to add a .value() to get the Esterel value. We do not add it      */
/* systematically in generateArg, because when a signal is taken as a*/
/* lvalue, we need the Sample, not the sample value.                 */
/*********************************************************************/
void generateValArg(FILE *cFile, argNode *arg) {

  generateArg(cFile,arg);                 /* Create code to get arg value */
  if (((arg->type == cVarVal)             /* If argument is a signal value */
       &&(varTable[arg->value].kind == vSigVal)
       &&(sigTable[varTable[arg->value].sigNum].kind != sException))
       ||(arg->type == cSigVal)) {
    fprintf(cFile,".value()");   /* we need a ".value()" to get its value. */
  }
}


/*********************************************************************/
/* Procedure generateValArgs produces code for 'by value' arguments. */
/* It merely iterates through the argument list, calling generateArg */
/* for each argument.                                                */
/*********************************************************************/
void generateValArgs(FILE *cFile, args *arglist) {
  args *tmp;
  
  tmp = arglist;
  while ((tmp != (args*)0)&&(tmp->node != (argNode*)0)) {
    generateValArg(cFile,tmp->node);     /* Create code to get their value */
    if (tmp->next != (args*)0)
      fprintf(cFile,", ");               /* Arguments are separated by commas */
    tmp = tmp->next;
  }
}

/************************************************************************/
/* Procedure generateRefArgs produces code for 'by reference' arguments.*/
/* It does so by prefixing the code produced by generateArg for each    */
/* argument, with a & ('address of' operator).                          */
/************************************************************************/
int generateRefArgs(FILE *cFile, args *arglist) {
  args *tmp;
  int argSeen = 0;
  
  tmp = arglist;
  while ((tmp != (args*)0)&&(tmp->node != (argNode*)0)) {
    argSeen = 1;
    if (refArePointers) {               /* If 'by reference' arguments are pointers */
      fprintf(cFile,"&(");              /* Same as val arg with a preceding '&' */
      generateArg(cFile,tmp->node);     /* '()'s override any misleading precedence */
      fprintf(cFile,")");
    } else                              /* 'by reference' arguments are C++ references */
      generateArg(cFile, tmp->node);
    if (tmp->next != (args*)0)
      fprintf(cFile,", ");
    tmp = tmp->next;
  }
  return argSeen;
}
