/* -*-C-*-
 * ###################################################################
 *  occ++ - oc to C++ post-compiler for Esterel
 * 
 *  FILE: "genMdf.c"
 *                                    created: 03/09/92 00:00:00 
 *                                last update: 26/11/97 18:25:10 
 *  Author: Frdric Boulanger
 *  E-mail: Frederic.Boulanger@supelec.fr
 *    mail: Suplec - Service Informatique
 *          Plateau de Moulon, F-91192 Gif-sur-Yvette cedex
 *     www: http://wwwsi.supelec.fr/fb/fb.html
 *  
 * ========================================================================
 * Permission to use, copy and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that:
 *   - the above copyright notice appears in all copies and,
 *   - both the copyright notice and warranty disclaimer appear in
 *     supporting documentation and,
 *   - any product making use of this software or of parts of this 
 *     software displays the following acknowledgment:
 *       "This product includes software developed by Suplec
 *        and its contributors"
 *     in the documentation or other materials provided with the product 
 *     and in all advertising material mentioning features or use of 
 *     this software.
 *     
 * Warranty disclaimer: this software is provided "as is",
 *   Suplec and Frdric Boulanger disclaim all warranties with regard 
 *   to this software, including all implied warranties of merchantability 
 *   and fitness.
 *   In no event shall Suplec or Frdric Boulanger be liable for any 
 *   special, indirect or consequential damages or any damages whatsoever 
 *   resulting from loss of use, data or profits, whether in an action of 
 *   contract, negligence or other tortuous action, arising out of or in 
 *   connection with the use or performance of this software.
 * ========================================================================
 *  Description: 
 *    This file contains code to produce Module Description files.
 *    mdf files are used to build compound or derived synchronous
 *    classes from previously compiled oc modules.
 * 
 *  History
 *  
 *  modified by  rev reason
 *  -------- --- --- -----------
 *  03/09/92 FBO 1.0 original
 *  26/11/97 FBO 4.8 support for C++ types and operators (mangled)
 * ###################################################################
 */

/* This file contains code to produce Module Description files.     */
/* mdf files are used to build compound or derived synchronous      */
/* classes from previously compiled oc modules.                     */
/********************************************************************/

#include <stdio.h>
#include <string.h>
#include "env.h"
#include "occxxErrors.h"
#include "occxxUtils.h"
#include "occxxTables.h"
#include "occxx.h"

void generateMdfFile() {
  FILE *mdfFile;
  int i;
  int first;
  char tmpStr[256];
  
  mdfFile = fopen(mdfFileName,"w");
  if (mdfFile == 0) {
    sprintf(tmpStr,"Can't open %s for write access",mdfFileName);
    fileError(tmpStr);
  }
  fprintf(mdfFile,"//\n//  %s\n// Created by occ++ v%s on %s\n//\n",
          mdfFileName, occxxVers, dateAndTime());
  fprintf(mdfFile,"#source \"%s\"\n", moduleTable[rootModule].srcfile);
  fprintf(mdfFile,"#oc \"%s\"\n", sourceFile);
  fprintf(mdfFile,"#c++ \"%s\"\n\n", hFileName);

  if (((nbTypes != 0) || (nbFunctions != 0) || (nbProcedures != 0) || (nbExecs != 0))
        && (noDefFile == 0)) {
    fprintf(mdfFile,"#import \"%s\"\n\n", hDefFileName); /* include user definitions */
  }

  if ((nbConsts != 0) && (paramClass != 0) &&
      !((nbConsts == 1) && (strcmp(constTable[0].name, "this") == 0))) {
    fprintf(mdfFile,"#param {\n");
    for(i = 0; i < nbConsts; i++) {
      if (strcmp(constTable[i].name, "this") != 0)
        fprintf(mdfFile, "  %s %s;\n", typeName(constTable[i].type),
                                       constTable[i].name);
    }
    fprintf(mdfFile,"}\n\n");
  }

  if ((nbConsts != 0) && (paramClass != 0) && (setParam == 0) &&
      !((nbConsts == 1) && (strcmp(constTable[0].name, "this") == 0))) {
    fprintf(mdfFile,"%s(#param)  {\n", moduleTable[rootModule].name);
  } else
    fprintf(mdfFile,"%s {\n", moduleTable[rootModule].name);

  first = 1;
  for (i = 0; i < nbSignals; i++) {
    switch(sigTable[i].kind) {
      case sSensor:
      case sInput:
      case sInputOutput:
        if (!first)
          fprintf(mdfFile,",\n");
        else {
          fprintf(mdfFile,"  input:\n");
          first = 0;
        }
        if (sigTable[i].type == sPure)
          fprintf(mdfFile,"    %s", sigTable[i].name);
        else
          fprintf(mdfFile,"    %s %s",
                  typeName(varTable[sigTable[i].varId].type),
                  sigTable[i].name);
      default:
	      break;
    }
  }
  if (!first)
    fprintf(mdfFile,";\n");

  first = 1;
  for (i = 0; i < nbSignals; i++) {
    switch(sigTable[i].kind) {
      case sInputOutput:
      case sOutput:
        if (!first)
          fprintf(mdfFile,",\n");
        else {
          fprintf(mdfFile,"  output:\n");
          first = 0;
        }
        if (sigTable[i].type == sPure)
          fprintf(mdfFile,"    %s", sigTable[i].name);
        else
          fprintf(mdfFile,"    %s %s",
                  typeName(varTable[sigTable[i].varId].type),
                  sigTable[i].name);
	    break;
	  default:
	    break;
    }
  }
  if (!first)
    fprintf(mdfFile,";\n");
  fprintf(mdfFile,"}\n");
  fclose(mdfFile);
}
