/* -*-C-*-
 * ###################################################################
 *  occ++ - oc to C++ post-compiler for Esterel
 * 
 *  FILE: "occxx.c"
 *                                    created: 03/09/92 00:00:00 
 *                                last update: 26/11/97 18:28:35 
 *  Author: Frdric Boulanger
 *  E-mail: Frederic.Boulanger@supelec.fr
 *    mail: Suplec - Service Informatique
 *          Plateau de Moulon, F-91192 Gif-sur-Yvette cedex
 *     www: http://wwwsi.supelec.fr/fb/fb.html
 *  
 * ========================================================================
 * Permission to use, copy and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that:
 *   - the above copyright notice appears in all copies and,
 *   - both the copyright notice and warranty disclaimer appear in
 *     supporting documentation and,
 *   - any product making use of this software or of parts of this 
 *     software displays the following acknowledgment:
 *       "This product includes software developed by Suplec
 *        and its contributors"
 *     in the documentation or other materials provided with the product 
 *     and in all advertising material mentioning features or use of 
 *     this software.
 *     
 * Warranty disclaimer: this software is provided "as is",
 *   Suplec and Frdric Boulanger disclaim all warranties with regard 
 *   to this software, including all implied warranties of merchantability 
 *   and fitness.
 *   In no event shall Suplec or Frdric Boulanger be liable for any 
 *   special, indirect or consequential damages or any damages whatsoever 
 *   resulting from loss of use, data or profits, whether in an action of 
 *   contract, negligence or other tortuous action, arising out of or in 
 *   connection with the use or performance of this software.
 * ========================================================================
 *  See header file for further information
 * ###################################################################
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#ifdef THINK_C
#include <console.h>
#endif

#include "env.h"
#include "occxxErrors.h"
#include "occxxUtils.h"
#include "occxxTables.h"
#include "genCxx.h"
#include "occxxParse.h"

#define SYMBOLTOSTRING(s) #s
#define SYMTOSTR(s) SYMBOLTOSTRING(s)

extern void generateMdfFile();

#if YYDEBUG != 0
extern int yydebug;            /* Yacc debug flag */
#endif

char hFileName[256];           /* Name of output interface file */
char cFileName[256];           /* Name of output C++ file */
char mdfFileName[256];         /* Name of Module Description File */
char mdfBaseFileName[256] = "";/* Base name for MDF files */
char hDefFileName[256] = "";   /* Name of user definitions interface file */
char baseName[256] = "";       /* Base name of the files */
char outDirectory[256] = "";   /* Directory for output files */
char *sourceFile = NULL;       /* Name of source file */
char tabBaseFileName[256] = "";/* Base name of state table files */
char *optIncludeEsterel = 0;   /* Esterel Include file given by -inc */

int verbose = 0;               /* Verbose flag. Default is not verbose */
int noDefFile = 0;             /* No .H user definitions file flag. Default */
                               /* is to include it if there are user defined */
                               /* types, constants, functions or procedures */
int cutLongLines = 0;          /* Cut automaton state tables into chunks */
                               /* (avoids long lines) */
int paramClass = 0;            /* Esterel constants are object parameters */
int setParam = 0;              /* Esterel constants are settable object parameters */
int nbModules = 0;             /* Number of parsed modules */
int allowDebug = 0;            /* Stuff C++ code with debug prompts on cerr */
int separateTables = 0;        /* Create separate files for state tables */
int doMDF = 0;                 /* Create an mdf file for each module */
int refArePointers = 0;        /* 'by reference' arguments are passed as pointers */
int wantAwaitedList = 0;       /* Create method to get awaited signals */
int noTickReaction = 0;        /* activate() does nothing on ticks without signals */

static char errStr[256];       /* Buffer to format error messages */

/***************************************************************************/
/* Function yywrap is called by the lex input function when it sees EOF.   */
/* It must return 1 if this is really the end of source file, 0 if not.    */
/* We always return 1 because our input is a single file without includes. */
/***************************************************************************/
int yywrap() {
  return 1;
}


/******************************************************************************/
/* Main entry point of occxx. Parses the argument list to find the oc source  */
/* file to compile and options. Note that only one source file is allowed.    */
/******************************************************************************/
int main(int argc, char **argv){
  FILE *ocSrc;                           /* oc source file */
  FILE *hFile;                           /* C++ interface file */
  char errStr[256];                      /* Buffer for error messages */
  int i;
  
#ifdef THINK_C
  argc = ccommand(&argv);                /* Stand alone THINK C application */
#endif
  i = 1;
  while (i < argc) {                     /* Decipher the command line */
    switch (argv[i][0]) {
      case '-':                          /* If argument is an option */
        if (strcmp(argv[i],"-v") == 0) { /* -v : Verbose option */
          verbose = 1;
          i++;
        }else if (strcmp(argv[i],"-version") == 0) {
          printf("--- occ++: version %s, %s\n",occxxVers, copyrightStr);
          i++;                           /* -version : Give version option */
          exit(0);
        }else if (strcmp(argv[i],"-access") == 0) {
          printf("--- occ++: free distribution, %s\n", copyrightStr);
          i++;                           /* -access : Give access rights */
          exit(0);
        }else if (strcmp(argv[i],"-info") == 0) {
          printf("--- occ++:  version %s\n", occxxVers);
          printf("         released: %s\n",SYMTOSTR(THE_RELEASE_DATE));
          printf("          formats: oc2,oc3,oc5 --> C++\n");
          printf("      compiled on: %s/%s\n",SYMTOSTR(ARCHI), SYMTOSTR(SYSTEM));
          printf("             with: %s\n",SYMTOSTR(C_COMPILER));
          printf("             date: %s\n",SYMTOSTR(THE_DATE));
          printf("               by: %s\n",SYMTOSTR(THE_GUY));
          i++;                           /* -info : Give info on occ++ */
          exit(0);
        }else if (strcmp(argv[i], "-B") == 0) {
          i++;                           /* -B myBaseName : Set basename option */
          if (argc <= i)
            usageError("base name expected after -B option.");
          strcpy(baseName,argv[i]);
          i++;
        }else if (strcmp(argv[i], "-D") == 0) {
          i++;                           /* -D directory : Set output dir option */
          if (argc <= i)
            usageError("directory name expected after -D option.");
          strcpy(outDirectory,argv[i]);
          i++;
        }else if (strcmp(argv[i], "-H") == 0) {
          i++;                           /* -H fileName : Set user interface file name option */
          if (argc <= i)
            usageError("file name expected after -H option.");
          strcpy(hDefFileName,argv[i]);
          i++;
          if (noDefFile != 0)
            fprintf(stderr,"# occ++ : Warning -H %s option cancelled by -Hnone option.\n",hDefFileName);
        }else if (strcmp(argv[i], "-Hnone") == 0) {
          noDefFile = 1;
          i++;
          if (strlen(hDefFileName) != 0)
            fprintf(stderr,"# occ++ : Warning -H %s option cancelled by -Hnone option.\n",hDefFileName);
        }else if (strcmp(argv[i], "-cut") == 0) {
          i++;
          if (argc <= i)
            usageError("Number expected after -cut option.");
          cutLongLines = atoi(argv[i]);
          i++;
          if (cutLongLines <= 0) {
            cutLongLines = 0;
            fprintf(stderr,"# occ++ : Warning, line length of -cut option is <= 0. Ignored.\n");
          }
        }else if (strcmp(argv[i], "-debug") == 0) {
          i++;
          allowDebug = 1;
#if YYDEBUG != 0
        }else if (strcmp(argv[i], "-yaccdbg") == 0) { /* Private yacc debug option */
          i++;
	  yydebug = 1;  /* Yacc debug flag, available if compiled with YYDEBUG != 0 */
#endif
        }else if (strcmp(argv[i], "-tab") == 0) {
          i++;
          separateTables = 1;
        }else if (strcmp(argv[i], "-tabB") == 0) {
          i++;
          if (argc <= i)
            usageError("state table file base name expected after -tabB option.");
          strcpy(tabBaseFileName,argv[i]);
          separateTables = 1;
          i++;
        }else if (strcmp(argv[i], "-mdf") == 0) {
          i++;
          doMDF = 1;
        }else if (strcmp(argv[i], "-mdfB") == 0) {
          i++;
          if (argc <= i)
            usageError("module description file base name expected after -mdfB option.");
          strcpy(mdfBaseFileName, argv[i]);
          doMDF = 1;
          i++;
        }else if (strcmp(argv[i], "-param") == 0) {
          i++;
          paramClass = 1;
        }else if (strcmp(argv[i], "-setparam") == 0) {
          i++;
          paramClass = 1;
          setParam = 1;
        }else if (strcmp(argv[i], "-noref") == 0) {
          i++;
          refArePointers = 1;
        }else if (strcmp(argv[i], "-inc") == 0) {
          i++;
          if (argc <= i)
            usageError("include file name expected after -inc option.");
          if ((argv[i][0] != '"') && (argv[i][0] != '<')) {
            optIncludeEsterel = checkAlloc(strlen(argv[i]) + 3, sizeof(char));
            sprintf(optIncludeEsterel, "\"%s\"", argv[i]);
          }else {
            optIncludeEsterel = checkAlloc(strlen(argv[i]) + 1, sizeof(char));
            strcpy(optIncludeEsterel, argv[i]);
          }
          i++;
        }else if (strcmp(argv[i], "-awaited") == 0) {
          i++;
          wantAwaitedList = 1;
        }else if (strcmp(argv[i], "-notick") == 0) {
          i++;
          noTickReaction = 1;
        }else {
          sprintf(errStr,"Unknown option %s",argv[i]);
          i++;
          usageError(errStr);
        }
        break;
      default:
        if (sourceFile == NULL) {        /* If we don't have a source file, that's ours */
          sourceFile = argv[i];
          i++;
        }else {                          /* else there is more than one source file. */
          i++;
          usageError("Only one oc source file can be specified");
        }
        break;
    }
  }
  if ((i = strlen(outDirectory)) != 0) {    /* If output directory specified */
    if (outDirectory[i-1] != dirSep) {      /* If it doesn't end with dir sep char */
      outDirectory[i] = dirSep;             /* Append dir sep char to dir name */
      outDirectory[i+1] = '\000';
    }
  }
  if (sourceFile == NULL)                   /* We NEED a source file */
    usageError("An oc source file must be specified");
  if ((ocSrc = fopen(sourceFile,"r")) == NULL) {
    sprintf(errStr,"Can't open file %s for input.",sourceFile);
    fileError(errStr);
  }
  yyin = ocSrc;                             /* Let yylex() read from source file */
  i = yyparse();                            /* Let Yacc do the job */
  fclose(ocSrc);                            /* and close the source file */
  if (i == 0) {
    hFile = fopen(hFileName,"a");    /* Append a closing #endif to x.H */
    if (hFile == NULL) {             /* If unable to open output file, fail */
      sprintf(errStr,"Can't open %s for write access",hFileName);
      fileError(errStr);
    }
    fprintf(hFile,"#endif\n");
  }
  if (verbose)
    printf("  ... done.\n");
  return i;                                 /* Status is Yacc status */
}


/*******************************************************************/
/* This function is called by the parser when it has succesfully   */
/* parsed a module and filled the parse tables.                    */
/* Here, we choose the name of output files, call the functions    */
/* that produce those files and clean up before returning to Yacc. */
/*******************************************************************/
void occxx() {
  if (ocVersion == 2) {
    stateInfo *curState, *oldState;
    int i;
    
    curState = stateTable;
    stateTable = (stateInfo*)checkAlloc(nbStates, sizeof(stateInfo));
    for(i = 0; i < nbStates; i++) {
      stateTable[i] = *curState;
      oldState = curState;
      curState = curState->next;
      free(oldState);
    }
  }
  
  checkVarsAndSigs();       /* Fill var table from sig table and check consistency */
  
  if (nbModules == 0) {                   /* First module gives file names */
    if (strcmp(baseName,"") == 0)         /* If no specified base name */
      strcpy(baseName,moduleName);        /* base name is module name */
    if (strlen(hDefFileName) == 0) {      /* If no specified def include file */
      sprintf(hDefFileName,hDefFileTemplate,baseName);
                                          /* Give one according to base name */
    }
    sprintf(hFileName,hFileTemplate,outDirectory,baseName);
    sprintf(cFileName,cFileTemplate,outDirectory,baseName);
                                          /* Build output file names */
  }
  if (verbose)                            /* If verbose option chosed */
    printModule();                        /* be verbose. */
  generateActLabels();                    /* Build action labels for module */
  if (verbose)
    printf("  building interface file \"%s\", current module is %s ...\n",
                                              hFileName, moduleName);
  generateHfile(nbModules);               /* Generate interface file */
  if (verbose)
    printf("  building C++ file \"%s\", current module is %s ...\n",
                                        cFileName, moduleName);
  generateCfile(nbModules);               /* Generate C++ file */
  if (doMDF) {                            /* If mdf files are requested */
    sprintf(mdfFileName,mdfFileTemplate,mdfBaseFileName, moduleName);
    if (verbose)
      printf("  building mdf file \"%s\" ...\n", mdfFileName);
    generateMdfFile();
  }
  freeStorage();                          /* Free storage allocated by the parser */
  nbModules ++;                           /* We parsed one more module. */
}
