/* -*-C-*-
 * ###################################################################
 *  occ++ - oc to C++ post-compiler for Esterel
 * 
 *  FILE: "occxxErrors.c"
 *                                    created: 03/09/92 00:00:00
 *                                last update: 26/11/97 18:30:29 
 *  Author: Frdric Boulanger
 *  E-mail: Frederic.Boulanger@supelec.fr
 *    mail: Suplec - Service Informatique
 *          Plateau de Moulon, F-91192 Gif-sur-Yvette cedex
 *     www: http://wwwsi.supelec.fr/fb/fb.html
 *  
 * ========================================================================
 * Permission to use, copy and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that:
 *   - the above copyright notice appears in all copies and,
 *   - both the copyright notice and warranty disclaimer appear in
 *     supporting documentation and,
 *   - any product making use of this software or of parts of this 
 *     software displays the following acknowledgment:
 *       "This product includes software developed by Suplec
 *        and its contributors"
 *     in the documentation or other materials provided with the product 
 *     and in all advertising material mentioning features or use of 
 *     this software.
 *     
 * Warranty disclaimer: this software is provided "as is",
 *   Suplec and Frdric Boulanger disclaim all warranties with regard 
 *   to this software, including all implied warranties of merchantability 
 *   and fitness.
 *   In no event shall Suplec or Frdric Boulanger be liable for any 
 *   special, indirect or consequential damages or any damages whatsoever 
 *   resulting from loss of use, data or profits, whether in an action of 
 *   contract, negligence or other tortuous action, arising out of or in 
 *   connection with the use or performance of this software.
 * ========================================================================
 *  See header file for further information
 * ###################################################################
 */

#include <stdlib.h>
#include <stdio.h>
#include "occxxParse.h"

/*************************************************************/
/* Procedure usageError handles occxx parameter list errors. */
/* It prints an error message followed by the correct usage  */
/* on stderr and then exits with a status of 1.              */
/*************************************************************/
void usageError(char *msg) {
  if (msg == 0)
    msg = "(null message)";
  fprintf(stderr,"### occ++ - Fatal usage error\n");
  fprintf(stderr,"# %s\n",msg);
  fprintf(stderr,"# Usage : occ++ [-v][-version][-B baseName][-D outputDirectory]\n\
#               [-H userDefsFileName] [-Hnone] [-cut n] [-debug]\n\
#               [-tab] [-tabB stateTableBaseName] [-mdf] [-mdfB]\n\
#               [-setparam] [-param] [-noref] [-inc EsterelIncludeName]\n\
#               [-awaited] [-notick] [-access] [-info]\n\
#               ocSourceFileName\n");
  exit(1);
}

/*********************************************************/
/* Procedure fileError handle file related fatal errors. */
/* It displays the given error message and then exits    */
/* with an error status of 1                             */
/*********************************************************/
void fileError(char *msg) {
  if (msg == 0)
    msg = "(null message)";
  fprintf(stderr,"### occ++ - Fatal I/O error : %s\n",msg);
  exit(1);
}

/*********************************************************/
/* Procedure fatalError handle unqualified fatal errors. */
/* It displays the given error message and then exits    */
/* with an error status of 1                             */
/*********************************************************/
void fatalError(char *msg) {
  if (msg == 0)
    msg = "(null message)";
  fprintf(stderr,"### occ++ - Fatal error : %s\n",msg);
  exit(1);
}

/*********************************************************************/
/* Procedure ocError handles errors due to inconsistencies in the oc */
/* source code. It called when something is declared with an id which*/
/* is out of range according to previous declarations for example.   */
/* Perhaps we should check for the version of oc!                    */
/* It can also be called because of a bug in the compiler !          */
/*********************************************************************/
void ocError(char *msg) {
  if (msg == 0)
    msg = "(null message)";
  fprintf(stderr,"### occ++ - Fatal : oc syntax error\n");
  fprintf(stderr," File %s; Line %d # %s\n",sourceFile,lineNb,msg);
  exit(1);
}

/****************************************************************************/
/* Procedure intError handles internal errors (table inconsistency etc...). */
/* It prints the message describing the error and exits with status 1. This */
/* procedure should never be called !!!                                     */
/****************************************************************************/
void intError(char *msg) {
  if (msg == 0)
    msg = "(null message)";
  fprintf(stderr,"### occ++ - Fatal Internal Error : %s\n",msg);
  exit(1);
}

/***********************************************************************/
/* Function yyerror is used by the yacc parser to signal a parse error.*/
/* It receives an error message and is expected to return 0 if we      */
/* cannot recover from the error, 1 if we can.                         */
/***********************************************************************/
int yyerror(char *s) {
  if (s == 0)
    s = "(null message)";
  fprintf(stderr,"### occ++ - Fatal error\n");
  fprintf(stderr," File '%s'; Line %d - Offset %d\n",sourceFile,lineNb, colNb);
  fprintf(stderr," # %s, offending token is '%s'\n",s,yytext);
  return 0;    /* We are not enough clever to recover from a parse error.*/
}

/**************************************************************/
/* Procedure ocWarning handles conditions where the compiler. */
/* is not sure of what it should do. So it warns the user that*/
/* it has made a choice which is perhaps not the good one.    */
/**************************************************************/
void ocWarning(char *msg) {
  if (msg == 0)
    msg = "(null message)";
  fprintf(stderr,"### occ++ - Warning : %s\n",msg);
}
