/* -*-C-*- */
%{
/* -*-C-*-
 * ###################################################################
 *  occ++ - oc to C++ post-compiler for Esterel
 * 
 *  FILE: "occxxGram.y"
 *                                    created: 03/09/92 00:00:00 
 *                                last update: 26/11/97 18:33:22 
 *  Author: Frdric Boulanger
 *  E-mail: Frederic.Boulanger@supelec.fr
 *    mail: Suplec - Service Informatique
 *          Plateau de Moulon, F-91192 Gif-sur-Yvette cedex
 *     www: http://wwwsi.supelec.fr/fb/fb.html
 *  
 * ========================================================================
 * Permission to use, copy and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that:
 *   - the above copyright notice appears in all copies and,
 *   - both the copyright notice and warranty disclaimer appear in
 *     supporting documentation and,
 *   - any product making use of this software or of parts of this 
 *     software displays the following acknowledgment:
 *       "This product includes software developed by Suplec
 *        and its contributors"
 *     in the documentation or other materials provided with the product 
 *     and in all advertising material mentioning features or use of 
 *     this software.
 *     
 * Warranty disclaimer: this software is provided "as is",
 *   Suplec and Frdric Boulanger disclaim all warranties with regard 
 *   to this software, including all implied warranties of merchantability 
 *   and fitness.
 *   In no event shall Suplec or Frdric Boulanger be liable for any 
 *   special, indirect or consequential damages or any damages whatsoever 
 *   resulting from loss of use, data or profits, whether in an action of 
 *   contract, negligence or other tortuous action, arising out of or in 
 *   connection with the use or performance of this software.
 * ========================================================================
 *  Description: 
 *    Grammar used to parse oc code from version 2 to version 5.
 *    This grammar contains 28 shift/reduce conflicts.
 * 
 *  History
 * 
 *  modified by  rev reason
 *  -------- --- --- -----------
 *  03/09/92 FBO 1.0 original
 *  26/11/97 FBO 4.8 support for C++ types and operators (mangled)
 * ###################################################################
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "occxxErrors.h"
#include "occxxTables.h"
#include "occxxParse.h"
#include "occxx.h"

char errBuf[256];

#ifdef THINK_C
#define bcopy(src,dst,n) memmove(src,dst,(long)n)
#endif
%}

%union{
  int intval;
  char *strval;
  argInfo *argval;
  sigInfo *sigval;
  varInfo *varval;
  actInfo *actval;
  args *lstval;
  argNode *nodval;
  stateStep *stpval;
  typeInfo *tppval;
  sigList *sltval;
}

%token  tLeftPar     tRightPar     tComma        tColon
%token  tOcVersion3  tModule       tInstances    tRoot
%token  tEnd         tEnum         tTypes        tInteger
%token  tSysInteger  tConstant     tImmedValue   tConstDecl
%token  tFuncDecl    tProcDecl     tSignals      tInputSig
%token  tOutput      tInOutputSig  tSensorSig    tPureSig
%token  tBoolSig     tSingleSig    tMultipleSig  tVariables
%token  tSigBool     tSigVal       tVarName      tVarCnt
%token  tSigNum      tSigName      tActions      tGoto         
%token  tPresent     tCall         tCombine      tCondition    
%token  tDash        tReset        tDecSautZero  tHalts                
%token  tCalls       tStates       tInitState    tEndModule            
%token  tIdent       tRenIdent     tString       tSigValue     
%token  tLocalSig    tExceptSig    tSysConstant  tLitString
%token  tImplications
%token  tExclusions  tOcVersion2   tAutomaton    tState
%token  tSensors     tAwaited      tOcVersion    tOcVersion4
%token  tOcVersion5  tSink         tLeftAngle    tRightAngle
%token  tValue       tVarWire      tVarRegister  tLeftSquare
%token  tRightSquare tDags         tReturn       tTasks
%token  tExecs       tStart        tKill         tSuspend
%token  tActivate    tAct          tLeftCurly    tRightCurly
%token  tSemiColon

%type <intval> tOcVersion3 tEnum tInteger tSysInteger tSysConstant
%type <intval> tConstant tImmedValue tSigNum tSigValue tOcVersion2
%type <intval> sig.boolstore tOcVersion tOcVersion4 tOcVersion5

%type <strval> tIdent tRenIdent tString tSigName tLitString
%type <intval> type.number func.restype typ.assign typ.eq typ.ne
%type <intval> typ.cond func.proc
%type <argval> argtypes arglist

%type <sigval> sig.decl.qual sig.decl inputsig.decl inoutputsig.decl
%type <sigval> outputsig.decl sensorsig.decl localsig.decl exceptsig.decl
%type <sigval> sig.type returnsig.decl

%type <varval> decl.vars.sig decl.vars.user decl.vars.count decl.vars.wire
%type <varval> decl.vars.register decl.vars.elem2

%type <actval> act.desc act.goto act.present act.output act.call act.combine
%type <actval> act.reset act.if act.dsz act.desc5.list act.exec.desc
%type <actval> act.desc5 act.start act.kill act.suspend act.activate
%type <actval> act.return act.act act.supp.desc

%type <lstval> ref.args ref.arglist val.args val.arglist 
%type <nodval> lvalue rvalue func.call 

%type <stpval> state.act state.act.elem state.act.sequence
%type <stpval> state.act.ifte prim.state.act
%type <stpval> state5.act state5.act.goto state5.act.elem state5.act.sequence
%type <stpval> state5.act.cdag  state5.act.odag state5.act.ifte prim.state5.act

%type <tppval> typ.funcproc

%type <sltval> excl.sig.list awaited.sig.list

%start oc.prog

%%
/* An oc program is either an oc2 or an oc3 program */
oc.prog : oc3.prog
        | oc2.prog
        | oc4.prog
        | oc5.prog
        | ocx.prog
        ;

ocx.prog : tOcVersion
           { 
             sprintf(errBuf,
                     "This version of occ++ does not handle oc%d code.",
                     $1);
             fatalError(errBuf);
           }
         ;

oc3.prog : oc3.modules
         ;

oc3.modules : oc3.module
            | oc3.modules oc3.module
            ;

oc2.prog : oc2.modules
         ;

oc2.modules : oc2.module
            | oc2.modules oc2.module
            ;

oc4.prog : oc4.modules
         ;

oc5.prog : oc5.modules
         ;

oc4.modules : oc4.module
            | oc4.modules oc4.module

oc5.modules : oc5.module
            | oc5.modules oc5.module

/* An oc module is composed of a header, declarations, actions and state machine tables */
oc3.module : header3 declarations3 actions3 state.machine3 tEndModule
             { occxx(); }
           ;

oc2.module : header2 declarations2 actions2 state.machine2 tEnd
             { occxx(); }
           ;

oc4.module : header4 declarations3 actions3 state.machine3 tEndModule
             { occxx(); }
           ;

oc5.module : header5 declarations5 actions5 state.machine5 tEndModule
             { occxx(); }
           ;

/***********************************/
/*            Header               */
/***********************************/
/* The header contains the version number of oc, the name of the module and other info */
header3 : tOcVersion3 module.name module.info3
           {
             ocVersion = $1;
             GotoAction = 0;
             JumpAction = 1;
             incrAction = 1;
           }
        ;

header2 : tOcVersion2 module.name
           {
             ocVersion = $1;
             GotoAction = 0;
             JumpAction = 1;
             incrAction = 1;
             initInstances(1);
             declareModule(0, moduleName, "** Unknown **");
           }
        ;

header4 : tOcVersion4 module.name module.info3
           {
             ocVersion = $1;
             GotoAction = 0;
             JumpAction = 1;
             incrAction = 1;
             strcpy(errBuf,
                    "occ++ does not fully handle oc4 code.\n");
             strcat(errBuf,"# Parsing it as extended oc3...");
             ocWarning(errBuf);
           }
        ;

header5 : tOcVersion5 module.name module.info3
           {
             ocVersion = $1;
             GotoAction = 0;
             ODAGAction = 1;
             CDAGAction = 2;
             JumpAction = 3;
             incrAction = 4;
           }
        ;

/* The name of the module is given by 'module: nameOfModule' */
module.name : tModule tIdent
              {moduleName = $2;}
            ;

/* Info on the module gives the number of submodules, the root module and info on each submodule */
module.info3 : module.info.instances3 module.info.root3 module.info.list3
             | /* may be empty */
               {
                  initInstances(1);
                  rootModule = 0;
                  declareModule(0, moduleName, "<unknown source>");
               }
             ;

/* The number of modules involved is given as 'instance: numberOfInstances' */
module.info.instances3 : tInstances tInteger
                         {initInstances($2);}
                      ;

/* The root module is given as 'root: moduleNumber' */
module.info.root3 : tRoot tInteger
                    {rootModule = $2;}
                  ;

/* Follows a list of pieces of information terminated by 'end:' */
module.info.list3 : module.info.elems3 tEnd
                  ;

module.info.elems3 : module.info.elem3
                   | module.info.elems3 module.info.elem3
                   ;

/* Each piece of information has the form 'modNumber: modName integer string string' */
/* The last string is the name of the source file of the submodule. */
module.info.elem3 : tEnum tIdent tInteger tString tString
                    {declareModule($1,$2,$5);}
                  | tEnum tRenIdent tInteger tString tString
                    {declareModule($1,$2,$5);}
                  ;


/**************************************/
/*           declarations             */
/**************************************/
/* the declaration section declares types, constants, functions, */
/* procedures, signals and variables. */
declarations3 : decl.externs3 decl.signals3 decl.relations3 decl.vars3
              ;

decl.externs3 : decl.types3 decl.consts3 decl.funcs3 decl.procs3
              ;

declarations2 : decl.externs2  decl.signals2 decl.relations2 decl.sensors2 decl.vars2
              ;

decl.externs2 : decl.types2 decl.consts2 decl.funcs2 decl.procs2
              ;

declarations5 : decl.externs5  decl.signals5 decl.relations3 decl.vars3 decl.tasks5 decl.execs5
              ;

decl.externs5 : decl.types3 decl.consts3 decl.funcs3 decl.procs3
              ;

/* Types are declared by 'types: numberOfTypes listOfTypes end:' */
decl.types3 : decl.type.head decl.types.elems3 tEnd
            | /* empty */
              {initTypes(0);}
            ;

decl.types2 : decl.type.head decl.types.elems2
            | /* empty */
              {initTypes(0);}
            ;

decl.type.head : tTypes tInteger
                 {initTypes($2);}
               ;

decl.types.elems3 :   /* empty */
                  | decl.types.elem3
                  | decl.types.elems3 decl.types.elem3
                  ;

decl.types.elems2 :   /* empty */
                  | decl.types.elem2
                  | decl.types.elems2 decl.types.elem2
                  ;

/* Each type is declared by 'typeNumber: typeName func/proc( integer or '-' )' */
decl.types.elem3 : tEnum tIdent typ.funcproc
                    {declareType($1,$2,$3);}
                 ;

decl.types.elem2 : tInteger tIdent typ.funcproc
                    {declareType($1,$2,$3);}
                 ;

typ.funcproc : typ.assign typ.eq typ.ne typ.cond
                 { $$ = newType($1,$2,$3,$4); }
             ;

typ.assign : func.proc
               { $$ = $1; }
           ;

typ.eq : func.proc
           { $$ = $1; }
       ;

typ.ne : func.proc
           { $$ = $1; }
       ;

typ.cond : func.proc
             { $$ = $1; }
         ;

func.proc : type.number
              { $$ = $1; }
          | tDash
              { $$ = -1; }
          ;

/* Constants are declared by 'constants: numberOfConsts listOfConstants end:' */
decl.consts3 : decl.consts.head decl.consts.elems3 tEnd
             | /* empty */
               {initConstants(0);}
             ;

decl.consts2 : decl.consts.head decl.consts.elems2
             | /* empty */
               {initConstants(0);}
             ;

decl.consts.head : tConstDecl tInteger
                   {initConstants($2);}
                 ;

decl.consts.elems3 :  /* empty */
                   | decl.consts.elem3
                   | decl.consts.elems3 decl.consts.elem3
                   ;

decl.consts.elems2 :  /* empty */
                   | decl.consts.elem2
                   | decl.consts.elems2 decl.consts.elem2
                   ;

/* Each constant is declared by 'constNumber: constName constType' */
decl.consts.elem3 : tEnum tIdent type.number
                     {declareConstant($1,$3,$2);}
                  ;

decl.consts.elem2 : tInteger tIdent type.number
                     {declareConstant($1,$3,$2);}
                  ;

/* A type is an integer (user defined type) or $integer (built in type) */
type.number : tInteger {$$ = $1;}
            | tSysInteger {$$ = -($1+2);}
            ;

/* Functions are declared by 'functions: numberOfFunc listOfFunctions end:' */
decl.funcs3 : decl.funcs.head decl.funcs.elems3 tEnd
            | /* empty */
              {initFunctions(0);}
            ;

decl.funcs2 : decl.funcs.head decl.funcs.elems2
            | /* empty */
              {initFunctions(0);}
            ;

decl.funcs.head : tFuncDecl tInteger
                  { initFunctions($2);}
                ;

decl.funcs.elems3 :  /* empty */
                  | decl.funcs.elem3
                  | decl.funcs.elems3 decl.funcs.elem3
                  ;

decl.funcs.elems2 :  /* empty */
                  | decl.funcs.elem2
                  | decl.funcs.elems2 decl.funcs.elem2
                  ;

/* Each function is declared by 'funcNumber: funcName (argType,...) : resType' */
decl.funcs.elem3 : tEnum tIdent arglist func.restype
                  { declareFunction($1,$2,$3,$4); }
                ;

decl.funcs.elem2 : tInteger tIdent arglist func.restype
                  { declareFunction($1,$2,$3,$4); }
                ;

arglist : tLeftPar argtypes tRightPar
          { $$ = $2; }
        ;
argtypes :   /* empty */  {$$ = (argInfo*)0;}
         | type.number
           { $$ = addArgType((argInfo*)0,$1);}
         | argtypes tComma type.number
           { $$ = addArgType($1,$3);}
         ;
func.restype : tColon type.number
               { $$ = $2; }
             ;

/* Procedures are declared by 'procedures: numberOfProcs listOfProcs end:' */
decl.procs3 : decl.proc.head decl.procs.elems3 tEnd
            | /* empty */
              {initProcedures(0);}
            ;

decl.procs2 : decl.proc.head decl.procs.elems2
            | /* empty */
              {initProcedures(0);}
            ;

decl.proc.head : tProcDecl tInteger
                 { initProcedures($2);}
               ;
decl.procs.elems3 :  /* empty */
                  | decl.procs.elem3
                  | decl.procs.elems3 decl.procs.elem3
                  ;

decl.procs.elems2 :  /* empty */
                  | decl.procs.elem2
                  | decl.procs.elems2 decl.procs.elem2
                  ;

/* Each procedure is declared by 'procNumber: procName refArgList valArgList' */
decl.procs.elem3 : tEnum tIdent arglist arglist
                    { declareProcedure($1,$2,$3,$4); }
                 ;

decl.procs.elem2 : tInteger tIdent arglist arglist
                    { declareProcedure($1,$2,$3,$4); }
                 ;

/* Signals are declared by 'signals: numberOfSigs listOfSigs end:' */
decl.signals3 : decl.signals.head decl.signals.elems3 tEnd
              | /* empty */
                {initSignals(0);}
              ;

decl.signals2 : decl.signals.head decl.signals.elems2
              | /* empty */
                {initSignals(0);}
              ;

decl.signals5 : decl.signals.head decl.signals.elems5 tEnd
              | /* empty */
                { initSignals(0); }
              ;

decl.signals.head : tSignals tInteger
                    { initSignals($2); }
                  ;

decl.signals.elems3 : /* empty */
                    | decl.signals.elem3
                    | decl.signals.elems3 decl.signals.elem3
                    ;

decl.signals.elems2 : /* empty */
                    | decl.signals.elem2
                    | decl.signals.elems2 decl.signals.elem2
                    ;

decl.signals.elems5 : /* empty */
                    | decl.signals.elem5
                    | decl.signals.elems5 decl.signals.elem5
                    ;

/* Each signal has the form 'sigNumber: sigDeclaration' */
decl.signals.elem3 : tEnum sig.decl
                      { declareSignal($1,$2); }
                   ;

decl.signals.elem2 : tInteger sig.decl
                      { declareSignal($1,$2); }
                   ;

decl.signals.elem5 : tEnum sig.decl
                      { declareSignal($1,$2); }
                   | tEnum returnsig.decl
                      { declareSignal($1,$2); }
                   ;

/* A signal can be input, input and output, output, sensor or local */
sig.decl : inputsig.decl
         | inoutputsig.decl
         | outputsig.decl
         | sensorsig.decl
         | localsig.decl
         | exceptsig.decl
         ;

inputsig.decl : tInputSig tIdent sig.decl.qual
                {  $3->kind = sInput;
                   $3->name = $2;
                   $$ = $3;
                }
              ;

inoutputsig.decl : tInOutputSig tIdent sig.decl.qual
                   {  $3->kind = sInputOutput;
                      $3->name = $2;
                      $$ = $3;
                   }
                 ;

outputsig.decl : tOutput tIdent sig.decl.qual
                 {  $3->kind = sOutput;
                    $3->name = $2;
                    $$ = $3;
                 }
               ;

sensorsig.decl : tSensorSig tIdent sig.decl.qual
                 {  $3->kind = sSensor;
                    $3->name = $2;
                    $$ = $3;
                 }
               ;

localsig.decl : tLocalSig sig.decl.qual tVarName tSigName
                {  $2->kind = sLocal;
                   $2->name = $4;
                   $$ = $2;
                }
              ;

exceptsig.decl : tExceptSig sig.decl.qual tVarName tSigName
                 {  $2->kind = sException;
                    $2->name = $4;
                    $$ = $2;
                 }
               ;

returnsig.decl : tReturn tIdent sig.decl.qual
                 {
                   $3->kind = sReturn;
                   $3->name = $2;
                   $$ = $3;
                 }

/* After the kind of the signal, we find the signal name, the list */
/* of actions that use it, and its type (pure, single or multiple) */
sig.decl.qual : sig.action.list sig.type sig.boolstore
                { $$ = $2;
                  $$->boolId = $3;
                }
              ;
sig.action.list : /* empty */
                | tDash      /* added for oc5 */
                  { ; /* Override default action to avoid type clash */}
                | tInteger
                  { ; /* Override default action to avoid type clash */}
                | sig.action.list tInteger
                  { ; /* Override default action to avoid type clash */}
                ;

sig.type : tPureSig
           { $$ = newSignal(sPure); }
         | tSingleSig tInteger
           {  $$ = newSignal(sSingle);
              $$->varId = $2; }
         | tMultipleSig tInteger type.number
           {  $$ = newSignal(sMultiple);
              $$->varId = $2;
              $$->combFunc = $3; }
         ;

sig.boolstore : /* empty */
                  { $$ = -1; }
              | tBoolSig tInteger
                  { $$ = $2; }
              ;

/* Relations between signals can be implications or exclusions */
decl.relations3 : decl.implic3 decl.excl3
                ;

decl.relations2 : decl.implic2 decl.excl2
                ;

decl.implic3 :  /* empty */
               { initImplications(0);}
             | decl.implic.head decl.implic.elems3 tEnd
             ;

decl.implic2 : /* empty */
               { initImplications(0);}
             | decl.implic.head
             | decl.implic.head decl.implic.elems2
             ;

decl.implic.head :  tImplications tInteger
                    { initImplications($2); }
                 ;

decl.implic.elems3 : decl.implic.elem3
                   | decl.implic.elems3 decl.implic.elem3
                   ;

decl.implic.elems2 : decl.implic.elem2
                   | decl.implic.elems2 decl.implic.elem2
                   ;

decl.implic.elem3 : tEnum tInteger tInteger
                     { declareImplication($1, $2, $3); }
                  ;

decl.implic.elem2 : tInteger tInteger tInteger
                     { declareImplication($1, $2, $3); }
                  ;

decl.excl3 : /* empty */
             { initExclusions(0); }
           | decl.excl.head decl.excl.elems3 tEnd
           ;

decl.excl2 : /* empty */
             { initExclusions(0); }
           | decl.excl.head
           | decl.excl.head decl.excl.elems2
           ;

decl.excl.head : tExclusions tInteger
                  { initExclusions($2); }
               ;

decl.excl.elems3 : decl.excl.elem3
                 | decl.excl.elems3 decl.excl.elem3
                 ;

decl.excl.elems2 : decl.excl.elem2
                 | decl.excl.elems2 decl.excl.elem2
                 ;

decl.excl.elem3 : tEnum tLeftPar excl.sig.list tRightPar
                   { declareExclusion($1, $3); }
                ;

decl.excl.elem2 : tInteger tLeftPar excl.sig.list tRightPar
                   { declareExclusion($1, $3); }
                ;

excl.sig.list : tInteger
                 { $$ = addSignal((sigList*)0,$1); }
              | excl.sig.list tComma tInteger
                 { $$ = addSignal($1,$3); }
              ;
/* In oc2, sensors are declared apart from signals */
decl.sensors2 : /* empty */
              | tSensors tInteger
              ;

/* Variables are declared by 'variables: numberOfVars listOfVars end:' */
decl.vars3 : decl.vars.head decl.vars.elems3 tEnd
           | /* empty */
             { initVariables(0); }
           ;

decl.vars2 : decl.vars.head decl.vars.elems2
           | /* empty */
             { initVariables(0); }
           ;

decl.vars.head : tVariables tInteger
                 { initVariables($2); }
               ;

decl.vars.elems3 : /* empty */
                 | decl.vars.elem3
                 | decl.vars.elems3 decl.vars.elem3
                 ;

decl.vars.elems2 : /* empty */
                 | decl.vars.elem2
                   { ; }
                 | decl.vars.elems2 decl.vars.elem2
                 ;

/* Variables are either signal variables or user variables */
/* They are declared by 'varNum: varType %...%' */
decl.vars.elem3 : tEnum type.number decl.vars.sig
                   { $3->type = $2;
                     declareVariable($1,$3);
                   }
                | tEnum type.number decl.vars.user
                   { $3->type = $2;
                     declareVariable($1,$3);
                   }
                | tEnum type.number decl.vars.count
                   { $3->type = $2;
                     declareVariable($1,$3);
                   }
                | tEnum type.number decl.vars.wire
                  {
                    $3->type = $2;
                    declareVariable($1,$3);
                  }
                | tEnum type.number decl.vars.register
                  {
                    $3->type = $2;
                    declareVariable($1,$3);
                  }
                | tEnum type.number /* oc5 from Lustre has no variable info */
                  {
                    declareNewVariable($1,$2,vVarUnknown);
                  }
                ;

decl.vars.elem2 : tInteger type.number
                   { $$ = newVariable(vVarUnknown);
                     $$->type = $2;
                     declareVariable($1,$$);
                   }
                ;

decl.vars.sig : tSigBool tSigNum
                {  $$ = newVariable(vSigBool);
                   $$->sigNum = $2;
                }
              | tSigVal tSigNum
                {  $$ = newVariable(vSigVal);
                   $$->sigNum = $2;
                }
              | tValue rvalue tSigVal tSigNum
                {  $$ = newVariable(vSigVal);
                   $$->initValue = $2;
                   $$->sigNum = $4;
                }
              ;

decl.vars.user : tVarName tSigName
                 {  $$ = newVariable(vVarVal);
                    $$->name = $2;
                 }
               ;

decl.vars.count : tVarCnt
                 {  $$ = newVariable(vVarCnt);
                    $$->name = nextVarCntName();
                 }
                ;

decl.vars.wire : tVarWire tVarName tSigName
                 {  $$ = newVariable(vVarWire);
                    $$->name = $3;
                 }
                ;

decl.vars.register : tValue rvalue tVarRegister tVarName tSigName
                 {  $$ = newVariable(vVarRegister);
                    $$->initValue = $2;
                    $$->name = $5;
                 }
                ;

/* Tasks are declared as procedures */
decl.tasks5 : decl.tasks.head decl.tasks.elems tEnd
            | /* empty */
              { initTasks(0); }
            ;

decl.tasks.head : tTasks tInteger
                  { initTasks($2); }
                ;

decl.tasks.elems : /* empty */
                 | decl.tasks.elem
                 | decl.tasks.elems decl.tasks.elem
                 ;
decl.tasks.elem : tEnum tIdent arglist arglist
                  { declareTask($1,$2,$3,$4); }
                ;

/* Execs are declared by 'execs: number list end:' */
decl.execs5 : decl.execs.head decl.execs.elems tEnd
            | /* empty */
              { initExecs(0); }
            ;

decl.execs.head : tExecs tInteger
                  { initExecs($2); }
                ;

decl.execs.elems : /* empty */
                 | decl.execs.elem
                 | decl.execs.elems decl.execs.elem
                 ;

decl.execs.elem : tEnum tInteger tInteger ref.args val.args
                  { declareExec($1,$2,$3,$4,$5); }
                ;

/***************************************/
/*               Actions               */
/***************************************/
/* Actions are declared by 'actions: numberOfAct listOfAct end:' */
actions3 : actions.head actions.elems3 tEnd
         | /* empty */
           { initActions(0); }
         ;

actions2 : actions.head actions.elems2
         | /* empty */
           { initActions(0); }
         ;

actions5 : actions.head actions.elems5 tEnd
         | /* empty */
           { initActions(0); }
         ;

actions.head : tActions tInteger
               { initActions($2); }
             ;

actions.elems3 :  /* empty */
               | actions.elem3
               | actions.elems3 actions.elem3
               ;

actions.elems2 :  /* empty */
               | actions.elem2
               | actions.elems2 actions.elem2
               ;

actions.elems5 :  /* empty */
               | actions.elem5
               | actions.elems5 actions.elem5
               ;

/* Each action has a number, a type and arguments */
actions.elem3 : tEnum act.desc
                 { declareAction($1,$2); }
              ;

actions.elem2 : tInteger act.desc
                 { declareAction($1,$2); }
              ;

actions.elem5 : tEnum act.desc5
                 { declareAction($1,$2); }
              | tEnum act.desc5.list
                 { declareAction($1,$2); }
              ;

act.desc5 : act.desc
          | act.exec.desc
          | act.supp.desc
          ;

act.desc : act.goto
         | act.present
         | act.output
         | act.call
         | act.combine
         | act.reset
         | act.if
         | act.dsz
         ;

act.exec.desc : act.start
              | act.kill
              | act.suspend
              | act.activate
              | act.return
              ;

act.supp.desc : act.act
              ;

act.desc5.list : act.desc5 act.desc5
                 {
                   $$ = newAction(aSeq);
                   $$->body.sequence.action = $1;
                   appendToActionSeq($$,$2);
                 }
              | act.desc5.list act.desc5
                 {
                   $$ = $1;
                   appendToActionSeq($$,$2);
                 }
              ;

/* The goto action is limited to the goto: keyword */
act.goto : tGoto
           { $$ = newAction(aGoto); }
         ;

/* The act action is act: action number */
act.act : tAct tInteger
           {
             $$ = newAction(aAct);
             $$->body.simple.sigNum = $2;
           }
        ;

/* The present action is 'present: sigNum' */
act.present : tPresent tInteger
              {  $$ = newAction(aPresent);
                 $$->body.simple.sigNum = $2;
              }
            ;

/* The output action is 'output: sigNum' */
act.output : tOutput tInteger
             {  $$ = newAction(aOutput);
                $$->body.simple.sigNum = $2;
             }
           ;

/* The call action is 'call: procNumber refArgs valArgs' */
act.call : tCall type.number ref.args val.args
           {  $$ = newAction(aCall);
              $$->body.call.procNum = $2;
              $$->body.call.refArgList = $3;
              $$->body.call.valArgList = $4;
           }
         ;

/* Exec related actions */
act.start : tStart tInteger
            {
              $$ = newAction(aStart);
              $$->body.simple.sigNum = $2;
            }
          ;

act.kill : tKill tInteger
           {
             $$ = newAction(aKill);
             $$->body.simple.sigNum = $2;
           }
         ;

act.suspend : tSuspend tInteger
              {
                $$ = newAction(aSuspend);
                $$->body.simple.sigNum = $2;
              }
            ;

act.activate : tActivate tInteger
               {
                 $$ = newAction(aActivate);
                 $$->body.simple.sigNum = $2;
               }
             ;

act.return : tReturn tInteger
             {
               $$ = newAction(aReturn);
               $$->body.simple.sigNum = $2;
             }
           ;

/* The first list of arguments is passed by reference. */
/* They must be l-values. */
ref.args : tLeftPar ref.arglist tRightPar
           { $$ = $2; }
         | tLeftPar tRightPar
           { $$ = (args*)0; }
         ;

ref.arglist : lvalue
              { $$ = addArgument((args*)0,$1); }
            | ref.arglist tComma lvalue
              { $$ = addArgument($1,$3); }
            ;

lvalue : tInteger
         {  $$ = newArgNode(cVarVal);
            $$->value = $1;
         }
       ;

/* The second list of arguments is passed by value. */
/* Such an argument can even be a function call. */
val.args : tLeftPar val.arglist tRightPar
           { $$ = $2; }
         | tLeftPar tRightPar
           { $$ = (args*)0; }
         ;

val.arglist : rvalue
              { $$ = addArgument((args*)0,$1); }
            | val.arglist tComma rvalue
              { $$ = addArgument($1,$3); }
            ;

rvalue : tInteger
         {  $$ = newArgNode(cVarVal);
            $$->value = $1;
         }
       | tSysInteger
         {  $$ = newArgNode(cSysVal);
            $$->value = $1;
         }
       | tConstant
         {  $$ = newArgNode(cConstVal);
            $$->value = $1;
         }
       | tSysConstant
         {  $$ = newArgNode(cSysConstVal);
            $$->value = $1;
         }
       | tImmedValue
         {  $$ = newArgNode(cImmedVal);
            $$->value = $1;
         }
       | tLitString
         {  $$ = newArgNode(cStringVal);
            $$->ptrValue = (void*)$1;
         }
       | tSigValue
         {  $$ = newArgNode(cSigVal);
            $$->value = $1;
         }
       | func.call
         {  $$ = $1; }
       ;

func.call : type.number val.args
            {  $$ = newArgNode(cFuncVal);
               $$->value = $1;
               $$->arguments = $2;
            }
          ;

/* The combine action is 'combine: sigNum value' */
act.combine : tCombine tInteger rvalue
              {  $$ = newAction(aCombine);
                 $$->body.combine.sigNum = $2;
                 $$->body.combine.oneArg = $3; }
            ;

/* The reset action is 'reset: sigNum' */
act.reset : tReset tInteger
            {  $$ = newAction(aReset);
               $$->body.simple.sigNum = $2;
            }
          ;

/* The if action is 'if: rvalue' */
act.if : tCondition rvalue
         {  $$ = newAction(aIf);
            $$->body.ifAct.oneArg = $2;
         }
       ;

/* The dsz action is 'dsz: var_number' */
act.dsz : tDecSautZero tInteger
         {  $$ = newAction(aDsz);
            $$->body.simple.sigNum = $2;
         }
       ;

/***********************************/
/*         State machine           */
/***********************************/
/* The state machine part describes haltsets, statistics, */
/* start point and transitions between states. */
state.machine3 : halt.desc3 machine.info3
               ;

state.machine2 : tAutomaton states2
               ;

state.machine5 : halt.desc3 dags5 machine.info5
               ;

/* Haltsets are declared by 'halts: numberOfHalts listOfHalts end:' */
halt.desc3 : tHalts tInteger halt.desc.elems tEnd
           | /* empty */
           ;

halt.desc.elems :  /* empty */
                | halt.desc.elem
                | halt.desc.elems halt.desc.elem
                ;

/* Each haltset is described by 'haltNum:' */
halt.desc.elem : tEnum
                 { ; /* Override default action to avoid type clash */}
               ;

dags5 : dags5.head dags5.elems tEnd
      | /* empty */
        { initDags(0); }
      ;

dags5.head : tDags tInteger tCalls tInteger
             { initDags($2); }
           ;

dags5.elems : dags5.elem
            | dags5.elems dags5.elem
            ;

dags5.elem : tEnum state5.act
             { declareDag($1,$2,0); }
           | tEnum state5.act tSemiColon
             { declareDag($1,$2,0); }
           ;

/* The machine info is 'calls: nbCalls states: nbStates' */
machine.info3 : machine.info.head3 startpoint3 states3 tEnd
              ;

machine.info5 : machine.info.head5 states5 tEnd
              ;

machine.info.head3 : tCalls tInteger tStates tInteger
                      { initStates($4); }
                   | tStates tInteger
                      { initStates($2); }
                   ;
          
machine.info.head5 : tStates tInteger tInitState tInteger tSink tInteger tCalls tInteger
                      { initStates($2); setStartState($4); }
                   | tStates tInteger tInitState tInteger tCalls tInteger
                      { initStates($2); setStartState($4); }
                   ;
          
/* The starpoint is 'startpoint: initState' */
startpoint3 : tInitState tInteger
              { setStartState($2); }
            ;

/* States are described as follows : 'stateNum: actionNum ...' */
states3 : state3
        | states3 state3
        ;

states2 : state2
        | states2 state2
        ;

states5 : state5
        | states5 state5
        ;

state3 : tEnum state.act tAwaited awaited.sig.list
          { 
            declareState($1,$2, $4);
            hasAwaitedInfo = 1;
          }
       | tEnum state.act
         {
           declareState($1,$2,0);
         }
       ;

state2 : tState tInteger state.act
          { 
            declareState($2,$3, 0);
            hasAwaitedInfo = 0;
          }
       ;

state5 : tEnum state5.act tAwaited awaited.sig.list
         {
           declareState($1,$2,$4);
           hasAwaitedInfo = 1;
         }
       | tEnum state5.act
         {
           declareState($1,$2,0);
         }
       ;

/* A list of actions follows the state number */
state.act : state.act.elem
            { $$ = $1; }
          | state.act state.act.elem
            { $$ = addStep($1,$2); }
          ;

state5.act : state5.act.elem
             { $$ = $1; }
           | state5.act state5.act.elem
             { $$ = addStep($1,$2); }
           ;

/* Each action can be in sequence or an IFThenElse action */
state.act.elem : state.act.sequence
               | state.act.ifte
               ;

state5.act.elem : state5.act.sequence
                | state5.act.goto
                | state5.act.cdag
                | state5.act.odag
                | state5.act.ifte
                ;

/* A sequential action is merely an action number */
state.act.sequence : tInteger
                     {  $$ = newStateStep(seqStep);
                        $$ -> actNum = $1;
                     }
                   ;

state5.act.sequence : state.act.sequence
                   ;

state5.act.goto : tLeftAngle tInteger tRightAngle
              {
                $$ = newStateStep(gotoStep);
                $$ -> actNum = $2;
              }
            ;

state5.act.cdag : tLeftSquare tInteger tRightSquare
              {
                $$ = newStateStep(cdagStep);
                $$ -> actNum = $2;
              }
            ;

state5.act.odag : tLeftCurly tInteger tRightCurly
              {
                $$ = newStateStep(odagStep);
                $$ -> actNum = $2;
                dagTable[$2].isOpenDAG = 1;
              }
            ;


/* An ifte action is the number of a test action followed by */
/* two lists of actions enclosed in braces. The first list is */
/* the 'then' branch of the test, the second is the 'else' branch */
state.act.ifte : tInteger tLeftPar prim.state.act tRightPar tLeftPar prim.state.act tRightPar
                 {  $$ = newStateStep(condStep);
                    $$ -> actNum = $1;
                    $$ -> thenBranch = $3;
                    $$ -> elseBranch = $6;
                 }
               ;
/* A list of actions in an ifte action can be empty */
prim.state.act : /* empty */
                 { $$ = (stateStep*)0; }
               | state.act
                 { $$ = $1; }
               ;

state5.act.ifte : tInteger tLeftPar prim.state5.act tRightPar tLeftPar prim.state5.act tRightPar
                 {  $$ = newStateStep(condStep);
                    $$ -> actNum = $1;
                    $$ -> thenBranch = $3;
                    $$ -> elseBranch = $6;
                 }
               ;
/* A list of actions in an ifte action can be empty */
prim.state5.act : /* empty */
                  { $$ = (stateStep*)0; }
                | state5.act
                  { $$ = $1; }
                ;

/* List of awaited input signals (Esterel) */
awaited.sig.list : tSigNum
                   { $$=addSignal(0, $1); }
                 | awaited.sig.list tSigNum
                   { $$=addSignal($1, $2); }
                 | /* empty */
                   { $$ = 0; }
                 ;

/* This is the end. */
