/* -*-C-*- */
%{
/* -*-C-*-
 * ###################################################################
 *  occ++ - oc to C++ post-compiler for Esterel
 * 
 *  FILE: "occxxParse.l"
 *                                    created: 03/09/92 00:00:00 
 *                                last update: 26/11/97 18:37:12 
 *  Author: Frdric Boulanger
 *  E-mail: Frederic.Boulanger@supelec.fr
 *    mail: Suplec - Service Informatique
 *          Plateau de Moulon, F-91192 Gif-sur-Yvette cedex
 *     www: http://wwwsi.supelec.fr/fb/fb.html
 *  
 * ========================================================================
 * Permission to use, copy and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that:
 *   - the above copyright notice appears in all copies and,
 *   - both the copyright notice and warranty disclaimer appear in
 *     supporting documentation and,
 *   - any product making use of this software or of parts of this 
 *     software displays the following acknowledgment:
 *       "This product includes software developed by Suplec
 *        and its contributors"
 *     in the documentation or other materials provided with the product 
 *     and in all advertising material mentioning features or use of 
 *     this software.
 *     
 * Warranty disclaimer: this software is provided "as is",
 *   Suplec and Frdric Boulanger disclaim all warranties with regard 
 *   to this software, including all implied warranties of merchantability 
 *   and fitness.
 *   In no event shall Suplec or Frdric Boulanger be liable for any 
 *   special, indirect or consequential damages or any damages whatsoever 
 *   resulting from loss of use, data or profits, whether in an action of 
 *   contract, negligence or other tortuous action, arising out of or in 
 *   connection with the use or performance of this software.
 * ========================================================================
 *  Description: 
 *    Lexical analyser for oc parsing.
 * 
 *  History
 * 
 *  modified by  rev reason
 *  -------- --- --- -----------
 *  03/09/92 FBO 1.0 original
 *  26/11/97 FBO 4.8 support for C++ types and operators (mangled)
 * ###################################################################
 */

#include "parseTypes.h"
#include "occxxGram.tab.h"

#include <stdio.h>
#include <fcntl.h>
#include <string.h>

#ifndef FLEX
#include <stdlib.h>
#endif

/* Prototypes for lex functions */
int yylook();
#ifndef FLEX
int yywrap();
#endif
int yyback(int* p, int m);

/* StrDup duplicates a C string */
#define StrDup(str) ((char*)strcpy(malloc(strlen(str)+1),str))

/* TRASH expands to the null statement. */
/* It is used to explicit the fact that we trash some input characters */
#define TRASH ;

extern YYSTYPE yylval; /* Union used to send values to the Yacc parser. */

int lineNb = 1;        /* Parse begins at line 1 of source file */
int colNb = 0;         /* Char offset on line */

%}

%p 4507
%e 1068

blank        [\ \t]
alpha        [a-zA-Z]
digit        [0-9]
special      [\_]
ident        (({alpha}|{special})({alpha}|{digit}|{special})*)
renident     {ident}\/{ident}
litinteger   ({digit}+)

%START COMMENT PRAGEST PRAGLUS SIGNUM VARNAME AWAITED COMPUTED
%%

<INITIAL>\%                 {colNb++; BEGIN PRAGEST;}
<INITIAL>\^                 {colNb++; BEGIN PRAGLUS;}
<INITIAL>--                 {colNb++; BEGIN COMMENT;}

<COMMENT>\n                 {lineNb++; BEGIN INITIAL;}
<COMMENT>.                  {colNb++; TRASH}
{blank}                     {colNb++; TRASH}
\n                          {lineNb++; colNb = 0;}

<INITIAL>\(                 {colNb++; return tLeftPar;}
<INITIAL>\)                 {colNb++; return tRightPar;}
<INITIAL>\,                 {colNb++; return tComma;}
<INITIAL>\:                 {colNb++; return tColon;}
<INITIAL>\;                 {colNb++; return tSemiColon;}
<INITIAL>\<                 {colNb++; return tLeftAngle;}
<INITIAL>\>                 {colNb++; return tRightAngle;}
<INITIAL>\[                 {colNb++; return tLeftSquare;}
<INITIAL>\]                 {colNb++; return tRightSquare;}
<INITIAL>\{                 {colNb++; return tLeftCurly;}
<INITIAL>\}                 {colNb++; return tRightCurly;}

<INITIAL>-                  {colNb++; return tDash;}

<INITIAL>^oc3:              {
                              colNb += 4;
                              yylval.intval = 3;
                              return tOcVersion3;
                            }

<INITIAL>^oc2:              {
                              colNb += 4;
                              yylval.intval = 2;
                              return tOcVersion2;
                            }

<INITIAL>^oc4:              {
                              colNb += 4;
                              yylval.intval = 4;
                              return tOcVersion4;
                            }

<INITIAL>^oc5:              {
                              colNb += 4;
                              yylval.intval = 5;
                              return tOcVersion5;
                            }

<INITIAL>^oc{litinteger}:   {
                              colNb += strlen((char*)yytext);
                              yylval.intval = atoi((char*)yytext+2);
                              return tOcVersion;
                            }

<INITIAL>^module:           {colNb += 7; return tModule;}
<INITIAL>{renident}		      {
                              colNb += strlen((char*)yytext);
                              yylval.strval = StrDup((char*)yytext);
                              return tRenIdent;
														}
<INITIAL>{ident}            {
                              colNb += strlen((char*)yytext);
                              yylval.strval = StrDup((char*)yytext);
                              return tIdent;
                            }
<INITIAL>^instances:        {colNb += 10; return tInstances;}
<INITIAL>^root:             {colNb += 5; return tRoot;}
<INITIAL>^end:              {colNb += 4; return tEnd;}
<INITIAL>#\"([^"]|\\\")*\"  {
                              colNb += strlen((char*)yytext);
                              yylval.strval = StrDup((char*)yytext+1);
                              return tLitString;
                            }
<INITIAL>\"([^"]|\\\")*\"   {
                              colNb += strlen((char*)yytext);
                              yylval.strval = StrDup((char*)yytext+1);
                              yylval.strval[strlen(yylval.strval)-1] = 0;
                              return tString;
                            }
<INITIAL>^{blank}*{litinteger}: {
                              colNb += strlen((char*)yytext);
                              yylval.intval = atoi((char*)yytext);
                              return tEnum;
                            }

<INITIAL>^types:            {colNb += 6; return tTypes;}
<INITIAL>{litinteger}       {
                              colNb += strlen((char*)yytext);
                              yylval.intval = atoi((char*)yytext);
                              return tInteger;
                            }
<INITIAL>${litinteger}      {
                              colNb += strlen((char*)yytext);
                              yylval.intval = atoi((char*)yytext+1);
                              return tSysInteger;
                            }
<INITIAL>@{litinteger}      {
                              colNb += strlen((char*)yytext);
                              yylval.intval = atoi((char*)yytext+1);
                              return tConstant;
                            }
<INITIAL>@${litinteger}     {
                              colNb += strlen((char*)yytext);
                              yylval.intval = atoi((char*)yytext+2);
                              return tSysConstant;
                            }
<INITIAL>#{litinteger}      {
                              colNb += strlen((char*)yytext);
                              yylval.intval = atoi((char*)yytext+1);
                              return tImmedValue;
                            }
<INITIAL>\?{litinteger}     {
                              colNb += strlen((char*)yytext);
                              yylval.intval = atoi((char*)yytext+1);
                              return tSigValue;
                            }
<INITIAL>^constants:        {colNb += 10; return tConstDecl;}
<INITIAL>^functions:        {colNb += 10; return tFuncDecl;}
<INITIAL>^procedures:       {colNb += 11; return tProcDecl;}
<INITIAL>^signals:          {colNb += 8; return tSignals;}
<INITIAL>input:             {colNb += 6; return tInputSig;}
<INITIAL>output:            {colNb += 7; return tOutput;}
<INITIAL>inputoutput:       {colNb += 12; return tInOutputSig;}
<INITIAL>sensor:            {colNb += 7; return tSensorSig;}
<INITIAL>local:             {colNb += 6; return tLocalSig;}
<INITIAL>exception:         {colNb += 10; return tExceptSig;}
<INITIAL>return:            {colNb += 7; return tReturn;}
<INITIAL>pure:              {colNb += 5; return tPureSig;}
<INITIAL>bool:              {colNb += 5; return tBoolSig;}
<INITIAL>single:            {colNb += 7; return tSingleSig;}
<INITIAL>multiple:          {colNb += 9; return tMultipleSig;}

<INITIAL>^implications:     {colNb += 13; return tImplications;}
<INITIAL>^exclusions:       {colNb += 11; return tExclusions;}
<INITIAL>^sensors:          {colNb += 8; return tSensors;}

<INITIAL>^variables:        {colNb += 10; return tVariables;}
<INITIAL>value:             {colNb += 6; return tValue;}
<PRAGEST>sigbool:           {colNb += 8; BEGIN SIGNUM; return tSigBool;}
<PRAGEST>sigval:            {colNb += 7; BEGIN SIGNUM; return tSigVal;}
<PRAGEST>name:              {colNb += 5; BEGIN VARNAME; return tVarName;}
<PRAGEST>count:             {colNb += 6; return tVarCnt;}
<PRAGEST>wire:              {colNb += 5; return tVarWire;}
<PRAGEST>register:          {colNb += 9; return tVarRegister;}
<PRAGEST>awaited:           {colNb += 8; BEGIN AWAITED; return tAwaited;}
<PRAGEST>computed:          {colNb += 9; BEGIN COMPUTED;}
<PRAGEST>\%                 {colNb ++; BEGIN INITIAL;}
<PRAGEST>.                  {colNb ++; TRASH}

<PRAGLUS>\^                 {colNb ++; BEGIN INITIAL;}
<PRAGLUS>.                  {colNb ++; TRASH}

<SIGNUM>{litinteger}        {
                              colNb += strlen((char*)yytext);
                              yylval.intval = atoi((char*)yytext);
                              BEGIN PRAGEST;
                              return tSigNum;
                            }
<VARNAME>{ident}            {
                              colNb += strlen((char*)yytext);
                              yylval.strval = StrDup((char*)yytext);
                              BEGIN PRAGEST;
                              return tSigName;
                            }

<AWAITED>\%                 { colNb++; BEGIN INITIAL;}

<AWAITED>{litinteger}       {
                              colNb += strlen((char*)yytext);
                              yylval.intval = atoi((char*)yytext);
                              return tSigNum;
                            }

<COMPUTED>\%                { colNb++; BEGIN INITIAL; }
<COMPUTED>.                 { colNb++; TRASH;}

<INITIAL>^tasks:            {colNb += 6; return tTasks;}
<INITIAL>^execs:            {colNb += 6; return tExecs;}
<INITIAL>^actions:          {colNb += 8; return tActions;}
<INITIAL>goto:              {colNb += 5; return tGoto;}
<INITIAL>act:               {colNb += 4; return tAct;}
<INITIAL>present:           {colNb += 8; return tPresent;}
<INITIAL>call:              {colNb += 5; return tCall;}
<INITIAL>combine:           {colNb += 8; return tCombine;}
<INITIAL>if:                {colNb += 3; return tCondition;}
<INITIAL>reset:             {colNb += 6; return tReset;}
<INITIAL>dsz:               {colNb += 4; return tDecSautZero;}
<INITIAL>start:             {colNb += 6; return tStart;}
<INITIAL>kill:              {colNb += 5; return tKill;}
<INITIAL>suspend:           {colNb += 8; return tSuspend;}
<INITIAL>activate:          {colNb += 9; return tActivate;}

<INITIAL>^dags:             {colNb += 5; return tDags;}
<INITIAL>^automaton:        {colNb += 10; return tAutomaton;}
<INITIAL>^halts:            {colNb += 6; return tHalts;}
<INITIAL>^calls:            {colNb += 6; return tCalls;}
<INITIAL>^state:            {colNb += 6; return tState;}
<INITIAL>^states:           {colNb += 7; return tStates;}
<INITIAL>^startpoint:       {colNb += 11; return tInitState;}
<INITIAL>^sink:             {colNb += 5; return tSink;}
<INITIAL>^endmodule:        {colNb += 10; return tEndModule;}
