/* -*-C-*-
 * ###################################################################
 *  occ++ - oc to C++ post-compiler for Esterel
 * 
 *  FILE: "occxxTables.c"
 *                                    created: 03/09/92 00:00:00
 *                                last update: 26/11/97 18:39:57 
 *  Author: Frdric Boulanger
 *  E-mail: Frederic.Boulanger@supelec.fr
 *    mail: Suplec - Service Informatique
 *          Plateau de Moulon, F-91192 Gif-sur-Yvette cedex
 *     www: http://wwwsi.supelec.fr/fb/fb.html
 *  
 * ========================================================================
 * Permission to use, copy and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that:
 *   - the above copyright notice appears in all copies and,
 *   - both the copyright notice and warranty disclaimer appear in
 *     supporting documentation and,
 *   - any product making use of this software or of parts of this 
 *     software displays the following acknowledgment:
 *       "This product includes software developed by Suplec
 *        and its contributors"
 *     in the documentation or other materials provided with the product 
 *     and in all advertising material mentioning features or use of 
 *     this software.
 *     
 * Warranty disclaimer: this software is provided "as is",
 *   Suplec and Frdric Boulanger disclaim all warranties with regard 
 *   to this software, including all implied warranties of merchantability 
 *   and fitness.
 *   In no event shall Suplec or Frdric Boulanger be liable for any 
 *   special, indirect or consequential damages or any damages whatsoever 
 *   resulting from loss of use, data or profits, whether in an action of 
 *   contract, negligence or other tortuous action, arising out of or in 
 *   connection with the use or performance of this software.
 * ========================================================================
 *  See header file for further information
 * ###################################################################
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "occxxErrors.h"
#include "occxxUtils.h"
#include "occxxTables.h"

static char tmpStr[256];       /* Buffer for error messages */

#define StrDup(str) ((char*)strcpy(malloc(strlen(str)+1),str))

/**************************************************************/
/*  See parseTypes.h for a description of the types used here */
/**************************************************************/

/*************************************/
/* Table of Esterel predefined types */
/*************************************/
typeInfo sysTypeTable[] = {    /* Info on predefined types */
  {"boolean",0,0,0,0},
  {"integer",0,0,0,0},
  {"string",0,0,0,0},
  {"float",0,0,0,0},
  {"double",0,0,0,0}
};


/* Predefined types have a negative ID */
/* User defined types have a positive or null ID */

#define boolTyp -1             /* Id of predefined type boolean */
#define intTyp -2              /* Id of predefined type integer */
#define strTyp -3              /* Id of predefined type string */
#define fltTyp -4              /* Id of predefined type float */
#define dblTyp -5              /* Id of predefined type double */

/*****************************************/
/* Table of Esterel predefined constants */
/*****************************************/
constInfo sysConstTable[] = {  /* Info on predefined constants */
  {boolTyp, "false"},
  {boolTyp, "true"}
};

/*****************************************/
/* Table of Esterel predefined functions */
/*****************************************/

static int BoolArgs[] = {      /* Argument list of function taking a boolean */
  boolTyp
};
static int IntArgs[] = {       /* Argument list of function taking an integer */
  intTyp
};
static int StrArgs[] = {       /* Argument list of function taking a string */
  strTyp
};
static int FltArgs[] = {       /* Argument list of function taking a float */
  fltTyp
};
static int DblArgs[] = {       /* Argument list of function taking a double */
  dblTyp
};
static int BoolBoolArgs[] = {  /* Argument list of function taking two booleans */
  boolTyp, boolTyp
};
static int BoolBoolBoolArgs[] = {  /* Argument list of function taking three booleans */
  boolTyp, boolTyp, boolTyp
};
static int BoolIntIntArgs[] = {    /* Argument list of COND for integers */
  boolTyp, intTyp, intTyp
};
static int IntIntArgs[] = {    /* Argument list of function taking two integers */
  intTyp, intTyp
};
static int StrStrArgs[] = {    /* Argument list of function taking two strings */
  strTyp, strTyp
};
static int BoolStrStrArgs[] = { /* Argument list of function COND for strings */
  boolTyp, strTyp, strTyp
};
static int FltFltArgs[] = {     /* Argument list of function taking two floats */
  fltTyp, fltTyp
};
static int BoolFltFltArgs[] = {     /* Argument list of function COND for floats */
  boolTyp, fltTyp, fltTyp
};
static int DblDblArgs[] = {     /* Argument list of function taking two doubles */
  dblTyp, dblTyp
};
static int BoolDblDblArgs[] = {     /* Argument list of function COND for doubles */
  boolTyp, dblTyp, dblTyp
};

static funcInfo equalBoolInfo = {       /* Descriptor of equality for booleans */
  boolTyp,"_eq_boolean",2,BoolBoolArgs
};
static funcInfo nonEqualBoolInfo = {    /* Descriptor of non equality for booleans */
  boolTyp,"_ne_boolean",2,BoolBoolArgs
};
static funcInfo condBoolInfo = {        /* Descriptor of COND for booleans */
  boolTyp,"_cond_boolean",3,BoolBoolBoolArgs
};
static funcInfo orBoolInfo = {          /* Descriptor of OR for booleans */
  boolTyp,"_or_boolean",2,BoolBoolArgs
};
static funcInfo andBoolInfo = {         /* Descriptor of AND for booleans */
  boolTyp,"_and_boolean",2,BoolBoolArgs
};
static funcInfo notBoolInfo = {         /* Descriptor of NOT for a boolean */
  boolTyp,"_not_boolean",1,BoolArgs
};

static funcInfo equalIntInfo = {        /* Descriptor of equality for integers */
  boolTyp,"_eq_integer",2,IntIntArgs
};
static funcInfo nonEqualIntInfo = {     /* Descriptor of non equality for integers */
  boolTyp,"_ne_integer",2,IntIntArgs
};
static funcInfo condIntInfo = {         /* Descriptor of COND for integers */
  boolTyp,"_cond_integer",3,BoolIntIntArgs
};
static funcInfo lowerThanIntInfo = {    /* Descriptor of lower than for integers */
  boolTyp,"_lt_integer",2,IntIntArgs
};
static funcInfo lowerEqualIntInfo = {   /* Descriptor of lower or equal for integers */
  boolTyp,"_le_integer",2,IntIntArgs
};
static funcInfo greaterThanIntInfo = {  /* Descriptor of greater than for integers */
  boolTyp,"_gt_integer",2,IntIntArgs
};
static funcInfo greaterEqualIntInfo = { /* Descriptor of greater or equal for integers */
  boolTyp,"_ge_integer",2,IntIntArgs
};
static funcInfo addIntInfo = {          /* Descriptor of addition for integers */
  intTyp,"_add_integer",2,IntIntArgs
};
static funcInfo subIntInfo = {          /* Descriptor of subtraction for integers */
  intTyp,"_sub_integer",2,IntIntArgs
};
static funcInfo mulIntInfo = {          /* Descriptor of multiplication for integers */
  intTyp,"_mul_integer",2,IntIntArgs
};
static funcInfo modIntInfo = {          /* Descriptor of remainder for integers */
  intTyp,"_mod_integer",2,IntIntArgs
};
static funcInfo divIntInfo = {          /* Descriptor of division for integers */
  intTyp,"_div_integer",2,IntIntArgs
};
static funcInfo oppIntInfo = {          /* Descriptor of unary minus for an integer */
  intTyp,"_opp_integer",1,IntArgs
};
static funcInfo bitorIntInfo = {        /* Descriptor of OR for integers */
  intTyp,"_bitor_integer",2,IntIntArgs
};
static funcInfo bitandIntInfo = {       /* Descriptor of AND for integers */
  intTyp,"_bitand_integer",2,IntIntArgs
};
static funcInfo bitnotIntInfo = {       /* Descriptor of NOT for an integer */
  intTyp,"_bitnot_integer",1,IntArgs
};

static funcInfo equalStrInfo = {        /* Descriptor of equality for strings */
  boolTyp,"_eq_string",2,StrStrArgs
};
static funcInfo nonEqualStrInfo = {     /* Descriptor of non equality for strings */
  boolTyp,"_ne_string",2,StrStrArgs
};
static funcInfo condStrInfo = {         /* Descriptor of COND for strings */
  strTyp,"_cond_string",3,BoolStrStrArgs
};

static funcInfo equalFltInfo = {           /* Descriptor of equality for floats */
  boolTyp,"_eq_float",2,FltFltArgs
};
static funcInfo nonEqualFltInfo = {        /* Descriptor of non equality for floats */
  boolTyp,"_ne_float",2,FltFltArgs
};
static funcInfo condFltInfo = {        /* Descriptor of COND for floats */
  fltTyp,"_cond_float",3,BoolFltFltArgs
};
static funcInfo lowerThanFltInfo = {        /* Descriptor of lower than for floats */
  boolTyp,"_lt_float",2,FltFltArgs
};
static funcInfo lowerEqualFltInfo = {       /* Descriptor of lower or equal for floats */
  boolTyp,"_le_float",2,FltFltArgs
};
static funcInfo greaterThanFltInfo = {      /* Descriptor of greater than for floats */
  boolTyp,"_gt_float",2,FltFltArgs
};
static funcInfo greaterEqualFltInfo = {     /* Descriptor of greater or equal for floats */
  boolTyp,"_ge_float",2,FltFltArgs
};
static funcInfo addFltInfo = {            /* Descriptor of addition for floats */
  fltTyp,"_add_float",2,FltFltArgs
};
static funcInfo subFltInfo = {            /* Descriptor of subtraction for floats */
  fltTyp,"_sub_float",2,FltFltArgs
};
static funcInfo mulFltInfo = {            /* Descriptor of multiplication for floats */
  fltTyp,"_mul_float",2,FltFltArgs
};
static funcInfo divFltInfo = {            /* Descriptor of division for floats */
  fltTyp,"_div_float",2,FltFltArgs
};
static funcInfo oppFltInfo = {            /* Descriptor of unary minus for floats */
  fltTyp,"_opp_float",1,FltArgs
};

static funcInfo equalDblInfo = {           /* Descriptor of equality for doubles */
  boolTyp,"_eq_double",2,DblDblArgs
};
static funcInfo nonEqualDblInfo = {        /* Descriptor of non equality for doubles */
  boolTyp,"_ne_double",2,DblDblArgs
};
static funcInfo condDblInfo = {            /* Descriptor of COND for doubles */
  dblTyp,"_cond_double",3,BoolDblDblArgs
};
static funcInfo lowerThanDblInfo = {        /* Descriptor of lower than for doubles */
  boolTyp,"_lt_double",2,DblDblArgs
};
static funcInfo lowerEqualDblInfo = {       /* Descriptor of lower or equal for doubles */
  boolTyp,"_le_double",2,DblDblArgs
};
static funcInfo greaterThanDblInfo = {      /* Descriptor of greater than for doubles */
  boolTyp,"_gt_double",2,DblDblArgs
};
static funcInfo greaterEqualDblInfo = {     /* Descriptor of greater or equal for doubles */
  boolTyp,"_ge_double",2,DblDblArgs
};
static funcInfo addDblInfo = {            /* Descriptor of addition for doubles */
  dblTyp,"_add_double",2,DblDblArgs
};
static funcInfo subDblInfo = {            /* Descriptor of subtraction for doubles */
  dblTyp,"_sub_double",2,DblDblArgs
};
static funcInfo mulDblInfo = {            /* Descriptor of multiplication for doubles */
  dblTyp,"_mul_double",2,DblDblArgs
};
static funcInfo divDblInfo = {            /* Descriptor of division for doubles */
  dblTyp,"_div_double",2,DblDblArgs
};
static funcInfo oppDblInfo = {            /* Descriptor of unary minus for doubles */
  dblTyp,"_opp_double",1,DblArgs
};
static funcInfo leftOrBoolInfo = {        /* Descriptor of CC OR for booleans */
  boolTyp,"_ccor_boolean",2,BoolBoolArgs
};
static funcInfo leftAndBoolInfo = {       /* Descriptor of CC AND for booleans */
  boolTyp,"_ccand_boolean",2,BoolBoolArgs
};

static funcInfo dszIntInfo = {       /* Descriptor of decrement, return true if <= 0 */
  boolTyp,"_dsz_int",1,IntArgs
};
static funcInfo unknownSysFunc = {        /* Descriptor of any unknown function */
  0,"**unknown sys func**",0,0
};

funcInfo* sysFuncTable[] = {            /* Table of predefined functions */
  &equalBoolInfo,
  &nonEqualBoolInfo,
  &condBoolInfo,
  &orBoolInfo,
  &andBoolInfo,
  &notBoolInfo,
  &equalIntInfo,
  &nonEqualIntInfo,
  &condIntInfo,
  &lowerThanIntInfo,
  &lowerEqualIntInfo,
  &greaterThanIntInfo,
  &greaterEqualIntInfo,
  &addIntInfo,
  &subIntInfo,
  &mulIntInfo,
  &modIntInfo,
  &divIntInfo,
  &oppIntInfo,
  &bitorIntInfo,
  &bitandIntInfo,
  &bitnotIntInfo,
  &equalStrInfo,
  &nonEqualStrInfo,
  &condStrInfo,
  &equalFltInfo,
  &nonEqualFltInfo,
  &condFltInfo,
  &lowerThanFltInfo,
  &lowerEqualFltInfo,
  &greaterThanFltInfo,
  &greaterEqualFltInfo,
  &addFltInfo,
  &subFltInfo,
  &mulFltInfo,
  &divFltInfo,
  &oppFltInfo,
  &equalDblInfo,
  &nonEqualDblInfo,
  &condDblInfo,
  &lowerThanDblInfo,
  &lowerEqualDblInfo,
  &greaterThanDblInfo,
  &greaterEqualDblInfo,
  &addDblInfo,
  &subDblInfo,
  &mulDblInfo,
  &divDblInfo,
  &oppDblInfo,
  &leftAndBoolInfo,
  &leftOrBoolInfo,
  &dszIntInfo
};

/******************************************/
/* Table of Esterel predefined procedures */
/******************************************/
static procInfo assignBoolInfo = {        /* Descriptor of assignement for booleans */
  "_assign_boolean",1,1,BoolArgs,BoolArgs
};
static procInfo assignIntInfo = {         /* Descriptor of assignement for integers */
  "_assign_integer",1,1,IntArgs,IntArgs
};
static procInfo assignStrInfo = {         /* Descriptor of assignement for strings */
  "_assign_string",1,1,StrArgs,StrArgs
};
static procInfo assignFltInfo = {         /* Descriptor of assignement for floats */
  "_assign_float",1,1,FltArgs,FltArgs
};
static procInfo assignDblInfo = {         /* Descriptor of assignement for doubles */
  "_assign_double",1,1,DblArgs,DblArgs
};

static procInfo unknownSysProc = {        /* Descriptor of any unknown procedure */
  "**unknown sys proc**",0,0,(int*)0,(int*)0
};

procInfo* sysProcTable[] = {              /* Table of predefined procedures */
  &assignBoolInfo,
  &assignIntInfo,
  &assignStrInfo,
  &assignFltInfo,
  &assignDblInfo
};

/***************************************/
/* Table of function tag and operators */
/***************************************/

char *tagName[] = {              /* Tag names = name in oc */
  "eq","ne","ccor","ccand","not",
  "or","and",
  "lt","le","gt","ge",
  "add","sub","mul","mod","div","opp",
  "bitor","bitand","bitnot","cond",
  "dsz"
};

char *opName[] = {               /* Corresponding operator names in C or C++*/
  " == "," != "," || "," && "," !",
  "|||","&&&",
  " < "," <= "," > "," >= ",
  " + "," - "," * "," % "," / "," -",
  " | "," & "," ~","?:",
  "(--)<=0"
};

char *cPlusTagName[] = {
  "eq","ne","not","or","and",
  "lt","le","gt","ge",
  "add","sub","mul","mod","div",
  "opp","bitor","bitand","bitnot","index",
  "ls","rs",
  "inc","dec","pinc","pdec","addd","subb",
  "mull","modd","divv",
  "bitorr","bitandd",0
};

char *cPlusOpName[] = {
  "(#1) == (#2)","(#1) != (#2)","!(#1)","(#1) || (#2)","(#1) && (#2)",
  "(#1) < (#2)","(#1) <= (#2)","(#1) > (#2)","(#1) >= (#2)",
  "(#1) + (#2)","(#1) - (#2)","(#1) * (#2)","(#1) % (#2)","(#1) / (#2)",
  "-(#1)","(#1) | (#2)","(#1) & (#2)","~(#1)","(#1)[#2]",
  "(#1) << (#2)","(#1) >> (#2)",
  "(#1)++","(#1)--","++(#1)","--(#1)","(#1) += (#2)","(#1) -= (#2)",
  "(#1) *= (#2)","(#1) %= (#2)","(#1) /= (#2)",
  "(#1) |= (#2)","(#1) &= (#2)",0
};

int cPlusArity[] = {
  2,2,1,2,2,
  2,2,2,2,
  2,2,2,2,2,
  1,2,2,1,2,
  2,2,
  1,1,1,1,2,2,
  2,2,2,
  2,2
};

/*************************************/
/* Information related to the module */
/*************************************/
int ocVersion;                         /* Version of oc language */
char *moduleName;                      /* Name of parsed module */
int rootModule = 0;                    /* Id of root submodule */
int nbInstances = 0;                   /* Number of submodules */
moduleInfo *moduleTable;               /* Table of modules */

/*********************************************/
/* Information related to user defined types */
/*********************************************/
int nbTypes = 0;                       /* Number of user defined types */
typeInfo *typeTable = (typeInfo*)0;    /* Table of user defined types */

/**************************************************/
/* Information related to user supplied constants */
/**************************************************/
int nbConsts = 0;                      /* Number of constants */
constInfo *constTable = (constInfo*)0; /* Table of constants */

/**************************************************/
/* Information related to user defined functions  */
/**************************************************/
int nbFunctions = 0;                   /* Number of user functions */
funcInfo *funcTable = (funcInfo*)0;    /* Table of user functions */

/***************************************************/
/* Information related to user supplied procedures */
/***************************************************/
int nbProcedures = 0;                  /* Number of user procedures */
procInfo *procTable = (procInfo*)0;    /* Table of user procedures */

/**********************************/
/* Information related to signals */
/**********************************/
int nbSignals = 0;                     /* Number of signals */
sigInfo *sigTable = (sigInfo*)0;       /* Table of signals */

/****************************************************/
/* Information related to relations between signals */
/****************************************************/
int nbImplications = 0;                /* Number of implications */
implInfo *implTable = (implInfo*)0;    /* Table of implications */
int nbExclusions = 0;                  /* Number of exclusions */
exclInfo *exclTable = (exclInfo*)0;    /* Table of exclusions */

/************************************/
/* Information related to variables */
/************************************/
int nbVariables = 0;                   /* Number of variables */
varInfo *varTable = (varInfo*)0;       /* Table of variables */

/************************************/
/* Information related to tasks     */
/************************************/
int nbTasks = 0;                       /* Number of tasks */
procInfo *taskTable = (procInfo*)0;    /* Table of tasks (same struct as procs) */

/************************************/
/* Information related to execs     */
/************************************/
int nbExecs = 0;                       /* Number of execs */
execInfo *execTable = (execInfo*)0;    /* Table of execs */

/************************************/
/* Information related to actions   */
/************************************/
int nbActions = 0;                     /* Number of actions */
actInfo *actTable = (actInfo*)0;       /* Table of actions */
int GotoAction;                        /* Code of the Goto action */
int ODAGAction;                        /* Code of the execute open dag action */
int CDAGAction;                        /* Code of the goto closed dag action */
int JumpAction;                        /* Code of the Jump Action */
int incrAction;                        /* increment for automaton actions */

/************************************/
/* Information related to dags      */
/************************************/
int nbDags = 0;
stateInfo *dagTable = (stateInfo*)0;

/************************************/
/* Information related to states    */
/************************************/
int nbStates = 0;                      /* Number of states */
stateInfo *stateTable = (stateInfo*)0; /* Table of states */
int initState = 0;                     /* Initial state */
int hasAwaitedInfo = 0;                /* Awaited signals info present? */


/**************************************************************************/
/**************************************************************************/
/**                                                                      **/
/** Functions used to work on the tables during parse and C++ production **/
/**                                                                      **/
/**************************************************************************/
/**************************************************************************/

/**********************************************************************/
/* Procedure initInstances stores the declared number of instances in */
/* the global variable nbInstances and allocates memory for the table */
/* of modules. It is called when the module header has been parsed in */
/* the oc source file.                                                */
/**********************************************************************/
void initInstances(int nb) {
  nbInstances = nb;    /* nbInstances is the declared number of module in the oc source file */
  if (nb != 0)
    moduleTable = (moduleInfo*)checkAlloc(nb,sizeof(moduleInfo));
}

/*************************************************************************************/
/* Procedure declareModule puts a newly parsed module into the modules' table.       */
/* The parsed id of the module is checked against the declared number of modules.    */
/* modName is the name of the module and srcFile is the name of its Esterel or Lustre*/
/* source file.                                                                      */
/*************************************************************************************/
void declareModule(int modID, char *modName, char *srcFile) {
  if (modID >= nbInstances){
    sprintf(tmpStr,"Module %s has number %d, which makes more than the %d declared modules",
                                                                  modName,modID,nbInstances);
    ocError(tmpStr);
  }
  moduleTable[modID].name = modName; /* modName and srcFile were allocated in the Yacc parser */
  moduleTable[modID].srcfile = srcFile;
}

/***************************************************************************/
/* Procedure initTypes stores the declared number of types in the global   */
/* variable nbTypes and allocates memory for the table of types.           */
/* It is called when the type header has been parsed in the oc source file.*/
/***************************************************************************/
void initTypes(int nb) {
  nbTypes = nb;
  if (nb != 0)
    typeTable = (typeInfo*)checkAlloc(nb,sizeof(typeInfo));
}

/*****************************************************************************/
/* Procedure newType allocates memory for a type information bloc and fills  */
/* it with the numbers of assignement, equality, non equality functions ,and */
/* conditional functions or procedures.                                      */
/*****************************************************************************/
typeInfo *newType(int assignP, int equalF, int diffF, int condF) {
  typeInfo *tmp;
  
  tmp = (typeInfo*)checkAlloc(1,sizeof(typeInfo));
  tmp->assignProc = assignP;
  tmp->equalFunc = equalF;
  tmp->diffFunc = diffF;
  tmp->condFunc = condF;
  return tmp;
}

/*************************************************************************************/
/* Procedure declareType puts a newly parsed user defined type into the types' table.*/
/* The parsed id of the type is checked against the declared number of types.        */
/* typeName is the name of the type.                                                 */
/*************************************************************************************/
void declareType(int typeID, char *typeName, typeInfo *theInfo) {
  char *t,*p;
  char *prefix, *suffix;
  char buf[40];
  
  if (typeID >= nbTypes){
    sprintf(tmpStr,"Type %s has number %d, which makes more than the %d declared types",
                   typeName,typeID,nbTypes);
    ocError(tmpStr);
  }
  if (strncmp(typeName,"CC_",3) == 0) { /* Special type */
    t = typeName+3;
    suffix = "";
    prefix = "";
    
    if (strncmp(t,"ptr_",4) == 0) {
      suffix = "*";
      t += 4;
    }
    if (strncmp(t,"ref_",4) == 0) {
      suffix = "&";
      t += 4;
    }
    while((p = strchr(t,'_')) != 0) {   /* Replace '_' with space in type name */
      *p = ' ';
    }
    sprintf(buf,"%s%s%s", prefix, t, suffix);
    t = StrDup(buf);
    free(typeName);
    typeName = t;
  }
  theInfo->name = typeName;     /* typeName was allocated by the Yacc parser */
  typeTable[typeID] = *theInfo; /* We make a copy of theInfo in our table */
  free(theInfo);                /* so we need to free the original block */
}

/**********************************************************************/
/* Procedure initConstants stores the declared number of constants in */
/* the global variable nbConstants and allocates memory for the table */
/* of constants. It is called when the constant header has been parsed*/
/* in the oc source file.                                             */
/**********************************************************************/
void initConstants(int nb) {
  nbConsts = nb;
  if (nb != 0)
    constTable = (constInfo*) checkAlloc(nb,sizeof(constInfo));
}

/*************************************************************************************/
/* Procedure declareConstant puts a newly parsed constant into the constants' table. */
/* The parsed id of the constant is checked against the declared number of constants.*/
/* constType is the type number of the constant. It is negative for predefined types */
/* and positive for user defined types. constName is the name of the constant.       */
/*************************************************************************************/
void declareConstant(int constID, int constType, char *constName) {
  if (constID >= nbConsts){
    sprintf(tmpStr,"Constant %s has number %d, which makes more than the %d declared constants",
                   constName,constID,nbConsts);
    ocError(tmpStr);
  }
  constTable[constID].type = constType;
  constTable[constID].name = constName;  /* constName was allocated by the Yacc parser */
}

/**********************************************************************/
/* Procedure initFunctions stores the declared number of functions in */
/* the global variable nbFunctions and allocates memory for the table */
/* of functions. It is called when the function header has been parsed*/
/* in the oc source file.                                             */
/**********************************************************************/
void initFunctions(int nb) {
  nbFunctions = nb;
  if (nb != 0)
    funcTable = (funcInfo*)checkAlloc(nb,sizeof(funcInfo));
}

/***********************************************************************/
/* Function addArgType adds an argument type at the end of an argument */
/* list. If the list doesn't exist, it is created. It return a pointer */
/* to the list.                                                        */
/***********************************************************************/
argInfo *addArgType(argInfo* argList,int theType) {
  argInfo *tmp;
  if (argList == 0) {                                   /* If the list is empty */
    argList = (argInfo*)checkAlloc(1,sizeof(argInfo));  /* create a new one */
    tmp = argList;
  } else {                                              /* If not empty */
    tmp = argList;
    while (tmp->next != 0)                              /* Go to the end of the list */
      tmp = tmp->next;
    tmp->next = (argInfo*)checkAlloc(1,sizeof(argInfo));/* Add a new cell */
    tmp = tmp->next;                                    /* Make it the current cell */
  }
  tmp->type = theType;                                  /* Store arg type in cell */
  return argList;                                       /* Return the list */
}

/*********************************************************/
/* Procedure freeArgList iterates through a linked list  */
/* of argument types and frees them.                     */
/*********************************************************/
void freeArgList(argInfo *argList) {
  argInfo *tmp,*f;
  
  tmp = argList;
  while (tmp != 0) {       /* While there are cells in the list */
    f = tmp;               /* Hold the cell to free */
    tmp = tmp->next;       /* Go to the next one */
    free(f);               /* Free the cell */
  }
}

/**************************************************************/
/* Function copyArgList copies an argument list into an       */
/* argument table which is faster to access. The argument     */
/* list was used during parse because it can grow dynamically.*/
/**************************************************************/
int copyArgList(argInfo *list, int **typeTab) {
  argInfo *tmp;
  int nArgs = 0;
  int i;
  int *tmpArgTab;
  
  if (list == 0) {
    *typeTab = 0;
    return 0;
  }
  
  tmp = list;
  while (tmp != 0) {            /* Get the length of the list */
    tmp = tmp->next;
    nArgs++;
  }
  tmpArgTab = (int*)checkAlloc(nArgs,sizeof(int));
                                /* Allocate the table */
  tmp = list;
  for (i = 0; i < nArgs; i++) { /* Fill the table with the list values */
    tmpArgTab[i] = tmp->type;
    tmp = tmp->next;
  }
  *typeTab = tmpArgTab;         /* Give back the table to the caller */
  return nArgs;                 /* Return the size of the table */
}

/*************************************************************************************/
/* Procedure declareFunction puts a newly parsed function into the functions' table. */
/* The parsed id of the function is checked against the declared number of functions.*/
/* funcName is the name of the function, arglist is the chained list of its arguments*/
/* type, and retType is its return type.                                             */
/*************************************************************************************/
void declareFunction(int funcID, char *funcName, argInfo *arglist,int retType) {
  if (funcID >= nbFunctions) {
    sprintf(tmpStr,"Function %s has number %d, which makes more than the %d declared functions",
                   funcName, funcID, nbFunctions);
    ocError(tmpStr);
  }
  funcTable[funcID].resType = retType;
  funcTable[funcID].name = funcName;
  
  funcTable[funcID].nbArgs = copyArgList(arglist,&(funcTable[funcID].argTypes));
                         /* Copy the arg list built during parse into the table */
                         /* of the function descriptor. */

  freeArgList(arglist);  /* After copying the arg list to the arg table, free the list */
}

/************************************************************************/
/* Procedure initProcedures stores the declared number of procedures in */
/* the global variable nbProcedures and allocates memory for the table  */
/* of procedures. It is called when the procedure header has been parsed*/
/* in the oc source file.                                               */
/************************************************************************/
void initProcedures(int nb) {
  nbProcedures = nb;
  if (nb != 0)
    procTable = (procInfo*)checkAlloc(nb,sizeof(procInfo));
}

/****************************************************************************************/
/* Procedure declareProcedure puts a newly parsed procedure into the procedures' table. */
/* The parsed id of the procedure is checked against the declared number of procedures. */
/* procName is the name of the procedure, refArgList is the chained list of its         */
/* 'by reference' arguments type and valArgList is the chained list of its 'by value'   */
/* arguments types. These lists are copied into arrays by copyArgList which returns the */
/* length of the list, and then freed.                                                  */
/****************************************************************************************/
void declareProcedure(int procID, char *procName, argInfo *refArglist, argInfo *valArglist) {
  if (procID >= nbProcedures) {
    sprintf(tmpStr,"Procedure %s has number %d, which makes more than the %d declared procedures",
                   procName, procID, nbProcedures);
    ocError(tmpStr);
  }
  procTable[procID].name = procName;
  
  procTable[procID].nbRefArgs = copyArgList(refArglist,&(procTable[procID].refArgTypes));
  freeArgList(refArglist);
  
  procTable[procID].nbValArgs = copyArgList(valArglist,&(procTable[procID].valArgTypes));
  freeArgList(valArglist);
}

/************************************************************************/
/* Procedure initSignals stores the declared number of signals in the   */
/* global variable nbSignals and allocates memory for the table of      */
/* signals. It is called when the signal header has been parsed in the  */
/* source file.                                                         */
/************************************************************************/
void initSignals(int nb) {
  nbSignals = nb;
  if (nb != 0)
    sigTable = (sigInfo*)checkAlloc(nb, sizeof(sigInfo));
}

/***************************************************************/
/* Function newSignal allocates a new signal descriptor for a  */
/* signal of type theType (pure, single, multiple). It is      */
/* called when a new signal is being parsed.                   */
/***************************************************************/
sigInfo *newSignal(sigType theType)
{
  sigInfo *tmp;
  tmp = (sigInfo*)checkAlloc(1,sizeof(sigInfo));
  tmp->type = theType;
  tmp->name = 0;
  tmp->varId = -1;
  tmp->boolId = -1;
  return tmp;
}

/*************************************************************************************/
/* Procedure declareSignal puts a newly parsed signal into the signals' table.       */
/* The parsed id of the signal is checked against the declared number of signals.    */
/* sigInfo describes the type, kind and storage of the signal.                       */
/*************************************************************************************/
void declareSignal(int sigID, sigInfo *sInfo) {
  if (sigID >= nbSignals) {
    sprintf(tmpStr,"Signal %s has number %d, which makes more than the %d declared signals.",
                   sInfo->name,sigID,nbSignals);
    ocError(tmpStr);
  }
  sigTable[sigID] = *sInfo;     /* We make a copy of sInfo in our table */
  free(sInfo);                  /* so we need to free the original block */
}

/****************************************************************************/
/* Procedure initImplications stores the declared number of implications in */
/* the global variable nbImplications and allocates memory for the table of */
/* implications. It is called when the implication header has been parsed.  */
/****************************************************************************/
void initImplications(int nb) {
  nbImplications = nb;
  if (nb != 0)
    implTable = (implInfo*)checkAlloc(nb, sizeof(implInfo));
}

/*****************************************************************************/
/* Procedure declareImplication puts a newly parsed implication in the table */
/* of implications. The Id of the implication is checked against the declared*/
/* number of implications.                                                   */
/*****************************************************************************/
void declareImplication(int impId, int cause, int effect) {
  if (impId >= nbImplications) {
    sprintf(tmpStr,"Implication with id %d, which makes more than the %d declared implications.",
                                                                          impId, nbImplications);
    ocError(tmpStr);
  }
  implTable[impId].cause = cause;
  implTable[impId].effect = effect;
}

/****************************************************************************/
/* Procedure initExclusions stores the declared number of exclusions in the */
/* global variable nbExclusions and allocates memory for the table of       */
/* exclusions. It is called when the exclusion header has been parsed.      */
/****************************************************************************/
void initExclusions(int nb) {
  nbExclusions = nb;
  if (nb != 0)
    exclTable = (exclInfo*)checkAlloc(nb, sizeof(exclInfo));
}

/*****************************************************************************/
/* Function addSignal adds a signal number to a list of signals. If the list */
/* is empty, it is created. The resulting list is returned.                  */
/*****************************************************************************/
sigList *addSignal(sigList *theList, int sigNum) {
  sigList *tmp;
  
  if (theList == (sigList*)0) {                          /* If the list is empty */
    theList = (sigList*)checkAlloc(1, sizeof(sigList));  /* create a new one */
    tmp = theList;
  } else {
    tmp = theList;                                       /* If it is not empty */
    while(tmp->next != (sigList*)0) {                    /* Go to the end of the list */
      tmp = tmp->next;
    }
    tmp->next = (sigList*)checkAlloc(1, sizeof(sigList));/* and add a new cell */
    tmp = tmp->next;
  }
  tmp->sigNum = sigNum;                                  /* Fill the cell */
  tmp->next = (sigList*)0;                               /* Tell them it's the last one */
  return theList;                                        /* Return the list */
}

/**************************************************************************/
/* Procedure declareExclusion puts a newly parsed exclusion in the table  */
/* of exclusions. The Id of the exclusion is checked against the declared */
/* number of exclusions. The list of signals is compacted into an array   */
/* and then freed.                                                        */
/**************************************************************************/
void declareExclusion(int exclId, sigList *theList) {
  int i, count;
  sigList *tmp, *old;
  
  if (exclId >= nbExclusions) {
    sprintf(tmpStr,"Exclusion with id %d, which makes more than the %d declared exclusions.",
                                                                       exclId, nbExclusions);
    ocError(tmpStr);
  }
  tmp = theList;
  count = 0;
  while(tmp != (sigList*)0) {                /* Get the number of signals in the list */
    count ++;
    tmp = tmp->next;
  }
  exclTable[exclId].nbSigs = count;          /* Store in the exclInfo struct */
  exclTable[exclId].sigs = (int*)checkAlloc(count, sizeof(int));
                                             /* Allocate the Array of signal IDs */
  tmp = theList;
  for (i = 0; i < count; i++) {              /* Fill the array with the values from the list */
    exclTable[exclId].sigs[i] = tmp->sigNum;
    old = tmp;
    tmp = tmp->next;
    free((char*)old);
  }
}

/**************************************************************************/
/* Procedure initVariables stores the declared number of variables in the */
/* global variable nbVariables and allocates memory for the table of      */
/* variables. It is called when the signal header has been parsed in the  */
/* oc source file.                                                        */
/**************************************************************************/
void initVariables(int nb) {
  nbVariables = nb;
  if (nb != 0)
    varTable = (varInfo*)checkAlloc(nb,sizeof(varInfo));
}

/***************************************************************************/
/* Function newVariable allocates a new variable descriptor for a variable */
/* of kind theKind (user variable, value or boolean of signal). It is      */
/* called when a new variable is being parsed.                             */
/***************************************************************************/
varInfo *newVariable(varKind theKind) {
  varInfo *tmp;
  
  tmp = (varInfo*)checkAlloc(1,sizeof(varInfo));
  tmp->kind = theKind;
  tmp->name = 0;
  tmp->sigNum = -1;
  tmp->initValue = (argNode*)0;
  return tmp;
}

/***************************************************************************/
/* Function nextVarCntName() returns a new name for an internal occurrence */
/* counter variable.                                                       */
/***************************************************************************/
char *nextVarCntName(void) {
  static int nextVarCntNum = 1;
  char buf[256];
  char * tmp;

  sprintf(buf,"_compteur_%d",nextVarCntNum++);
  tmp = checkAlloc(strlen(buf)+1,1);
  strcpy(tmp,buf);
  return tmp;
}

/***************************************************************************/
/* Function buildVarName() returns a name for a variable of the module.    */
/* We need to postfix the name with the variable id because multiple local */
/* variable with the same name can be declared in different submodules.    */
/***************************************************************************/
char *buildVarName(char *shortName, int id) {
  char buf[256];
  char * tmp;

  sprintf(buf,"%s_%d",shortName, id);
  tmp = checkAlloc(strlen(buf)+1,1);
  strcpy(tmp,buf);
  return tmp;
}

/*************************************************************************************/
/* Procedure declareVariable puts a newly parsed variable into the variables' table. */
/* The parsed id of the variable is checked against the declared number of variables.*/
/* vInfo describes the type, kind (user variable, signal variable...) and name of    */
/* the variable.                                                                     */
/*************************************************************************************/
void declareVariable(int varId, varInfo *vInfo) {
  char *tmp;
  if (varId >= nbVariables) {
    sprintf(tmpStr,"Variable %s has number %d, which makes more than the %d declared variables.",
                   vInfo->name,varId,nbVariables);
    ocError(tmpStr);
  }
  if (vInfo->kind == vVarVal) {
    tmp = vInfo->name;
    vInfo->name = buildVarName(tmp, varId);
    free(tmp);
  }
  if (vInfo->kind == vVarUnknown) {
    int i;
    for (i = 0; i < nbSignals; i++) {
      if (sigTable[i].varId == varId) {
        vInfo->kind = vSigVal;
        vInfo->sigNum = i;
        break;
      }
      if (sigTable[i].boolId == varId) {
        vInfo->kind = vSigBool;
        vInfo->sigNum = i;
        break;
      }
    }
    if (vInfo->kind == vVarUnknown) {
      vInfo->kind = vVarVal;
      vInfo->name = buildVarName("var", varId);
    }
  }
  varTable[varId] = *vInfo;   /* We make a copy of vInfo in our table */
  free(vInfo);                /* so we need to free the original block */
}

/***************************************************************/
/* Procedure declareNewVariable creates a new variable and     */
/* declares it in one operation. Used for Lustre signal        */
/* variables.                                                  */
/***************************************************************/
void declareNewVariable(int varId, int typeId, varKind theKind) {
  varInfo *vInfo;
  
  vInfo = newVariable(theKind);
  vInfo->type = typeId;
  declareVariable(varId, vInfo);
}

/***************************************************************/
/* Procedure checkVarsAndSigs checks that variables that hold  */
/* signal values or booleans match what is declared for signals*/
/***************************************************************/
void checkVarsAndSigs(void) {
  int i;
  
  for (i = 0; i < nbSignals; i++) {
    if (sigTable[i].boolId >= 0) {
      switch (varTable[sigTable[i].boolId].kind) {
        case vSigBool:
          break;
        case vVarUnknown:
          varTable[sigTable[i].boolId].kind = vSigBool;
          varTable[sigTable[i].boolId].sigNum = i;
          break;
        default:
          sprintf(tmpStr,
                  "Inconsistent tables: bool of signal %d (%s) and variable %d",
                  i, sigTable[i].name, sigTable[i].boolId);
          ocError(tmpStr);
          break;
      }
    }
    if ((sigTable[i].type == sPure) || (sigTable[i].varId < 0)) {
      continue;
    }
    switch (varTable[sigTable[i].varId].kind) {
      case vSigVal:
        break;
      case vVarUnknown:
        varTable[sigTable[i].varId].kind = vSigVal;
        varTable[sigTable[i].varId].sigNum = i;
        break;
      default:
        sprintf(tmpStr,
                "Inconsistent tables: value of signal %d (%s) and variable %d",
                i, sigTable[i].name, sigTable[i].boolId);
        ocError(tmpStr);
        break;
    }
  }
  for (i = 0; i < nbVariables; i++) {
    switch (varTable[i].kind) {
      case vSigBool:
        if (sigTable[varTable[i].sigNum].boolId != i) {
          sprintf(tmpStr,
                  "Inconsistent tables: variable %d and bool of signal %d (%s)",
                  i, varTable[i].sigNum, sigTable[varTable[i].sigNum].name);
          ocError(tmpStr);
        }
        break;
      case vSigVal:
        if (sigTable[varTable[i].sigNum].varId != i) {
          sprintf(tmpStr,
                  "Inconsistent tables: variable %d and value of signal %d (%s)",
                  i, varTable[i].sigNum, sigTable[varTable[i].sigNum].name);
          ocError(tmpStr);
        }
        break;
      default:
        break;
    }
  }
}

/***************************************************************/
/* Procedure initTasks stores the declared number of tasks in  */
/* the global variable nbTasks and allocates memory for the    */
/* table of tasks. It is called when the task header has been  */
/* parsed in the oc source file.                               */
/***************************************************************/
void initTasks(int nb) {
  nbTasks = nb;
  if (nb != 0)
    taskTable = (procInfo*)checkAlloc(nb,sizeof(procInfo));
}

/******************************************************************************/
/* Procedure declareTask puts a newly parsed task into the tasks' table.      */
/* The parsed id of the task is checked against the declared number of tasks. */
/* Argument lists are copied into tables, and the original lists are freed.   */
/******************************************************************************/
void declareTask(int taskID, char *taskName, argInfo *refArglist, argInfo *valArglist) {
  if (taskID >= nbTasks) {
    sprintf(tmpStr,"Task %s has number %d, which makes more than the %d declared tasks",
                   taskName, taskID, nbTasks);
    ocError(tmpStr);
  }
  taskTable[taskID].name = taskName;
  
  taskTable[taskID].nbRefArgs = copyArgList(refArglist,&(taskTable[taskID].refArgTypes));
  freeArgList(refArglist);
  
  taskTable[taskID].nbValArgs = copyArgList(valArglist,&(taskTable[taskID].valArgTypes));
  freeArgList(valArglist);
}

/***************************************************************/
/* Procedure initExecs stores the declared number of execs in  */
/* the global variable nbExecs and allocates memory for the    */
/* table of execs. It is called when the exec header has been  */
/* parsed in the oc source file.                               */
/***************************************************************/
void initExecs(int nb) {
  nbExecs = nb;
  if (nb != 0)
    execTable = (execInfo*)checkAlloc(nb,sizeof(execInfo));
}

/******************************************************************************/
/* Procedure declareExec puts a newly parsed exec into the execs' table.      */
/* The parsed id of the exec is checked against the declared number of execs. */
/******************************************************************************/
void declareExec(int execID, int taskID, int retSigID, args *refArgs, args *valArgs) {
  if (execID >= nbExecs) {
    sprintf(tmpStr,"Exec with id %d exceeds declared number of execs (%d)",execID, nbExecs);
    ocError(tmpStr);
  }
  execTable[execID].task = taskID;
  execTable[execID].retsig = retSigID;
  execTable[execID].refArgList = refArgs;
  execTable[execID].valArgList = valArgs;
}

/******************************************************************/
/* Procedure initActions stores the declared number of actions in */
/* the global variable nbActions and allocates memory for the     */
/* table of actions. It is called when the action header has been */
/* parsed in the oc source file.                                  */
/******************************************************************/
void initActions(int nb) {
  nbActions = nb;
  if (nb != 0)
    actTable = (actInfo*)checkAlloc(nb,sizeof(actInfo));
}

/*********************************************************************************/
/* Procedure declareAction puts a newly parsed action into the action table.     */
/* The parsed id of the action is checked against the declared number of actions.*/
/*********************************************************************************/
void declareAction(int actId, actInfo *theAct) {
  if (actId >= nbActions) {
    sprintf(tmpStr,"Action number %d is out of the specified range (%d).",actId,nbActions);
    ocError(tmpStr);
  }
  actTable[actId] = *theAct;   /* We make a copy of theAct in our table */
  free(theAct);                /* so we need to free the original block */
}

/***************************************************************/
/* Function newAction allocates a new action descriptor for an */
/* action of type theType (call, if, present etc...). It is    */
/* called when a new action is being parsed.                   */
/***************************************************************/
actInfo *newAction(actType theType) {
  actInfo *tmp;
  
  tmp = (actInfo*)checkAlloc(1, sizeof(actInfo));
  tmp->type = theType;
  if (theType == aSeq)
    tmp->body.sequence.next = 0;
  return tmp;
}

/***************************************************************/
/* Function appendToActionSeq appends an action to a sequence  */
/* of actions.                                                 */
/***************************************************************/
void appendToActionSeq(actInfo *list, actInfo *theAct) {
  actInfo *last = list;
  actInfo *tmp = list;

  if ((list == 0) || (list->type != aSeq)) {
    intError("Attempt to append an action to something other than a sequence.");
  }
  while(last->body.sequence.next != 0) {
    last = last->body.sequence.next;
  }
  tmp = newAction(aSeq);
  tmp->body.sequence.action = theAct;
  
  last->body.sequence.next = tmp;
}

/********************************************************************/
/* Function newArgNode allocates a new argument node descriptor for */
/* an argument of type theType (immediate value, variable etc...)   */
/* It is called when a new argument is being parsed.                */
/********************************************************************/
argNode *newArgNode(argType theType) {
  argNode *tmp;
  
  tmp = (argNode*)checkAlloc(1, sizeof(argNode));
  tmp -> type = theType;
  return tmp;
}

/***************************************************************/
/* Function addArgument adds an argument node at the end of    */
/* an argument list. The argument node is put in an arg struct */
/* before being append to the list.                            */
/* If the list is empty, it is created.                        */
/***************************************************************/
args *addArgument(args *arglist, argNode *n) {
  args *tmp;
  
  if (arglist == (args*)0) {                      /* If the list is empty */
    arglist = (args*)checkAlloc(1, sizeof(args)); /* Create a new one */
    tmp = arglist;
  } else {                                        /* If it is not empty */
    tmp = arglist;
    while (tmp->next != (args*)0) {               /* Go to the end of the list */
      tmp = tmp->next;
    }
    tmp->next = (args*)checkAlloc(1,sizeof(args));/* and add a new cell */
    tmp=tmp->next;
  }
  tmp -> node = n;                                /* Store the argNode in the cell */
  tmp -> next = (args*)0;                         /* Tell them it's the last one */
  return arglist;                                 /* and return the list */
}

/*******************************************************************/
/* Procedure initDags stores the declared number of dags in the    */
/* global variable nbDags and allocates memory for the table of    */
/* dags. It is called when the dags header has been parsed in the  */
/* source file.                                                    */
/*******************************************************************/
void initDags(int nb) {
  int i;

  nbDags = nb;
  if (nb != 0)
    dagTable = (stateInfo*)checkAlloc(nb,sizeof(stateInfo));
  for(i = 0; i < nb; i++)
    dagTable[i].isOpenDAG = 0;
}

/****************************************************************************/
/* Procedure declareDag puts a newly parsed dag into the dags' table.       */
/* The parsed id of the dag is checked against the declared number of dags. */
/* theSteps is the linked list of actions to perform in this dag. It is     */
/* compiled into an array of actions called the action table of the dag.    */
/****************************************************************************/
void declareDag(int dagId, stateStep *theSteps, sigList *awaited) {
  stateInfo *curDag;
  sigList *curSig, *oldSig;
  int i, nbAwaited;
  
  if (dagId >= nbDags) {
    sprintf(tmpStr,"Dag number %d is out of the specified range (%d)",dagId, nbDags);
    ocError(tmpStr);
  }
  curDag = &dagTable[dagId];
  compileState(theSteps,curDag);
  
  curSig = awaited;
  nbAwaited = 0;
  while(curSig != 0) {
    nbAwaited++;
    curSig = curSig->next;
  }
  curDag->nbAwaited = nbAwaited;
  if (nbAwaited == 0)
    curDag->awaited = 0;
  else {
    curDag->awaited = (int*)checkAlloc(nbAwaited, sizeof(int));
    for(curSig = awaited, i = 0;
        i < nbAwaited;
        oldSig = curSig, curSig = curSig->next, free(oldSig))
      curDag->awaited[i++] = curSig->sigNum;
  }
}

/************************************************************************/
/* Procedure initStates stores the declared number of states in the     */
/* global variable nbStates and allocates memory for the table of       */
/* states. It is called when the state header has been parsed in the    */
/* source file.                                                         */
/************************************************************************/
void initStates(int nb) {
  nbStates = nb;
  if (nb != 0)
    stateTable = (stateInfo*)checkAlloc(nb,sizeof(stateInfo));
}

/******************************************************************/
/* Procedure setStartState stores the number of the initial state */
/* in the global variable initState. It is called when the        */
/* startpoint: oc statement has been parsed.                      */
/******************************************************************/
void setStartState(int stateId) {
  initState = stateId;
}

/*************************************************************************************/
/* Procedure declareState puts a newly parsed state into the states' table.          */
/* The parsed id of the state is checked against the declared number of states.      */
/* theSteps is the linked list of actions to be performed in this state. It is       */
/* compiled into an array of actions called the action table of the state.           */
/*************************************************************************************/
void declareState(int stateId, stateStep *theSteps, sigList *awaited) {
  stateInfo *curState;
  sigList *curSig, *oldSig;
  int i, nbAwaited;
  
  if (stateId >= nbStates) {
    if (ocVersion > 2) { /* From version 3, we know the number of states */
      sprintf(tmpStr,"State number %d is out of the specified range (%d)",stateId, nbStates);
      ocError(tmpStr);
    } else {             /* With version 2, the number of states grows as we see them */
      nbStates = stateId + 1;
      if (stateId == 0) {
        stateTable = (stateInfo*)checkAlloc(1, sizeof(stateInfo));
        stateTable->next = 0;
        curState = stateTable;
      } else {
        curState = stateTable;
        while(curState->next != 0)
          curState = curState->next;
        curState->next = (stateInfo*)checkAlloc(1, sizeof(stateInfo));
        curState = curState->next;
        curState->next = 0;
      }
    }
  }
  if (ocVersion > 2)
    curState = &stateTable[stateId];
  compileState(theSteps,curState);
  
  curSig = awaited;
  nbAwaited = 0;
  while(curSig != 0) {
    nbAwaited++;
    curSig = curSig->next;
  }
  curState->nbAwaited = nbAwaited;
  if (nbAwaited == 0)
    curState->awaited = 0;
  else {
    curState->awaited = (int*)checkAlloc(nbAwaited, sizeof(int));
    for(curSig = awaited, i = 0;
        i < nbAwaited;
        oldSig = curSig, curSig = curSig->next, free(oldSig))
      curState->awaited[i++] = curSig->sigNum;
  }
}

/************************************************************/
/* Function addStep adds a step to a list of state steps.   */
/* The Yacc parser can't call this function with an empty   */
/* list, so we don't care about the value of 'list'. The    */
/* first stateStep is created by the newStateStep function. */
/************************************************************/
stateStep *addStep(stateStep *list, stateStep *elem) {
  stateStep *tmp;
  
  tmp = list;
  while (tmp->next != (stateStep*)0) {  /* Go to the end of the list */
    tmp = tmp->next;
  }
  tmp->next = elem;                     /* Put the stateStep there */
  elem->next = (stateStep*)0;           /* Tell them it's the last one */
  return list;                          /* and return the list */
}

/***************************************************************/
/* Function newStateStep allocates a new step descriptor for a */
/* step of type theType (sequential or conditional). It is     */
/* called when a new step in a state is being parsed.          */
/***************************************************************/
stateStep *newStateStep(stepType theType) {
  stateStep *tmp;
  
  tmp = (stateStep*)checkAlloc(1,sizeof(stateStep));
  tmp->type = theType;
  tmp->next = (stateStep*)0;  /* Nothing follows : it is not yet a list ! */
  return tmp;
}

/**************************************************************************/
/* Procedure compileState allocates and fills the action table of a state */
/* from the list of steps to do in that state.                            */
/* The actual fill of the action table is done by compileSteps.           */
/**************************************************************************/
void compileState(stateStep *theSteps, stateInfo *sInfo) {
  int s;
  int *tab;
  
  s = stepLength(theSteps);     /* Compute the number of steps in the list */
  sInfo->length = s;            /* Store in for future use */
  sInfo->actions = (int*)checkAlloc(s,sizeof(int));
                                /* Allocate a table to store the action numbers */
  tab = sInfo->actions;
  compileSteps(theSteps,&tab);  /* Compile the step list into the table */
  freeSteps(theSteps);          /* We can now free the step list */
}

/**********************************************************************/
/* Procedure freeSteps iterates through a linked list of stateStep    */
/* (a stateStep describes an action to be performed in a given state) */
/* and frees them. When a conditional step is found, a recursive call */
/* occurs to free both branches.                                      */
/**********************************************************************/
void freeSteps(stateStep *theSteps) {
  stateStep *tmp;
  
  while (theSteps != (stateStep*)0) {  /* While there are steps in the list */
    tmp = theSteps;                    /* Hold the step to free */
    theSteps = theSteps->next;         /* Go to the next one */
    switch (tmp->type) {               /* Let's have a look at the step type */
      case seqStep:                    /* If it's a brave regular seqStep */
      case gotoStep:                   /* or a goto step (oc5) */
      case cdagStep:                   /* or a goto closed dag step (oc5) */
      case odagStep:                   /* or an execute open dag step (oc5) */
        free(tmp);                     /* just get rid of it. */
        break;

      case condStep:                   /* but if it's a conditional step */
        freeSteps(tmp->thenBranch);    /* don't forget to free its list of 'then' */
        freeSteps(tmp->elseBranch);    /* steps and its list of 'else' steps */
        free(tmp);                     /* You can then get rid of this step. */
        break;
    }
  }
}

/***********************************************************************/
/* Procedure compileSteps fills an action table from the list of steps */
/* to follow in a state. stp points to the list of steps, and tab      */
/* points to the action table (which is an array of integers).         */
/* Action numbers other than goto (numbered 0) are incremented by one  */
/* because an additional action numbered one is created to jump over   */
/* branches of conditional actions. (It is labeled branchAlways).      */
/* Conditional steps are implemented by a branchIfFalse step. The code */
/* following the step is the number of steps to jump when the condition*/
/* is false. When the condition is true, the code following the jump   */
/* offset is interpreted. So the sequence of codes is :                */
/* branchIfFalse(condition)/jump offset when false/next code when true */
/***********************************************************************/
void compileSteps(stateStep *stp, int **tab) {
  while (stp != (stateStep*)0) {                  /* While there are steps in the list */
    switch (stp->type) {                          /* watch your steps */
      case seqStep:                               /* If it is a sequential step */
        if (actTable[stp->actNum].type == aGoto) {/* and if its action is a goto */
          **tab = stp->actNum;                    /* just store the action number in the table. */
          (*tab)++;                               /* Go to the next slot in the table */
          stp = stp -> next;                      /* and to the next step in the list. */
          **tab = stp->actNum;                    /* This is the number of the state to go to. */
          (*tab)++;                               /* Go to the next slot in the table. */
        } else {                                  /* If it is not a goto */
          **tab = stp->actNum + incrAction;       /* store the incremented action number */
          (*tab)++;                               /* Go to the next slot in the table. */
        }
        break;
      case gotoStep:                              /* If it is a goto (reserved in oc5) */
        **tab = GotoAction;                       /* Store the reserved action number for */
        (*tab)++;                                 /* goto in the action table. */
        **tab = stp->actNum;                      /* actNum holds the state to go to */
        (*tab)++;
        break;
      case cdagStep:                              /* If it is a goto closed dag (oc5) */
        **tab = CDAGAction;                       /* store the reserved action number for */
        (*tab)++;                                 /* it in the action table. */
        **tab = stp->actNum;                      /* actNum holds the dag to go to. */
        (*tab)++;
        break;
      case odagStep:                              /* If it is an execute open dag (oc5) */
        **tab = ODAGAction;                       /* store the reserved action number for */
        (*tab)++;                                 /* it in the action table. */
        **tab = stp->actNum;                      /* actNum holds the dag to execute. */
        (*tab)++;
        break;
      case condStep:                              /* If it is a conditional step */
        if (stp->thenLength == 0) {               /* if it has no 'then' branch */
          if (stp->elseLength == 0) {             /* and no 'else' branch */
            /* Nothing to do */                   /* it's a curious step ! */
          } else {                                /* If it has only an 'else' branch */
            **tab = stp->actNum + incrAction;     /* store the incremented action number */
            (*tab)++;
            **tab = 3;                            /* If condition is false, jump 3 codes */
            (*tab)++;                             /* to find the first code of the 'else' branch. */
            **tab = JumpAction;                   /* If it's true, jump */
            (*tab)++;
            **tab = stp->elseLength + 1;          /* over the 'else' branch. */
            (*tab)++;
            compileSteps(stp->elseBranch,tab);    /* compile the 'else' branch. */
          }
        } else {                                  /* If it has a 'then' branch */
          if (stp->elseLength == 0) {             /* but no 'else' branch */
            **tab = stp->actNum + incrAction;     /* Store the incremented action number */
            (*tab)++;
            **tab = stp->thenLength + 1;          /* If condition is false, jump over the 'then' branch */
            (*tab)++;
            compileSteps(stp->thenBranch,tab);    /* compile the 'then' branch */
          } else {                                /* If it has a 'then' and a 'else' branch */
            **tab = stp->actNum + incrAction;     /* store the incremented action number */
            (*tab)++;
            **tab = stp->thenLength + 3;          /* When false, jump to the 'else' branch */
            (*tab)++;
            compileSteps(stp->thenBranch,tab);    /* compile the 'then' branch */
            **tab = JumpAction;                   /* After the 'then' branch, jump over ...*/
            (*tab)++;
            **tab = stp->elseLength + 1;          /* ...the 'else' branch. */
            (*tab)++;
            compileSteps(stp->elseBranch,tab);    /* compile the 'else' branch */
          }
        }
        break;
      default:                                    /* Should never happen */
        sprintf(tmpStr,"Invalid type %d for a state step",stp->type);
        intError(tmpStr);
    }
    stp = stp->next;                              /* Go on with the next step */
  }
}

/**************************************************************************/
/* Function stepLength is a recursive function for computing the length   */
/* of the action table of a state. It computes the global length of the   */
/* table, including both branches of conditional actions and the overhead */
/* due to the branch actions inserted to jump over conditional branches.  */
/* This overhead varies according to the kind of the conditional action.  */
/* It is 4 for a if-else action, 2 for a if-then action and 4 for an if-  */
/* then-else action.                                                      */
/**************************************************************************/
int stepLength(stateStep *s) {
  if (s == (stateStep*)0)                       /* The length of an empty list is 0 */
    return 0;
  switch (s->type) {                            /* Look at the type of the step */
    case seqStep:                               /* A sequential step is just 1 step */
      return (1+stepLength(s->next));
      break;
    case gotoStep:                              /* A goto step is a goto + a state number */
    case cdagStep:                              /* A goto closed dag is cdag + dag number */
      return (2+stepLength(s->next));
      break;
    case odagStep:                              /* An execute open dag is odag + dag number */
      return (2+stepLength(s->next));
      break;
    case condStep:                              /* A conditional step is more complex */
      s->thenLength = stepLength(s->thenBranch);/* Compute its 'then' and 'else' branch length */
      s->elseLength = stepLength(s->elseBranch);
      if (s->thenLength == 0) {                 /* If there is no 'then' branch */
        if (s->elseLength == 0)                 /* and no 'else' branch */
          return (stepLength(s->next));         /* the length is the length of what follows! */
        else                                           /* If there is just an 'else' branch */
          return (4+s->elseLength+stepLength(s->next));/* the length is 4 more steps */
          /* two for the test and its offset, two for the jump over the 'else' branch and its offset */
      } else {                                  /* If there is a 'then' branch */
        if (s->elseLength == 0)                 /* but no 'else' branch */
          return (2+s->thenLength+stepLength(s->next)); /* the length is two more steps */
                             /* two for the test and its offset over the 'then' branch. */
        else                                    /* If there are a 'then' and a 'else' branch */
          return (4+s->thenLength+s->elseLength+stepLength(s->next));/* the length is 4 more steps. */
      }
      break;
    default:                                    /* Should never happen. */
      sprintf(tmpStr,"Invalid type %d for a state step",s->type);
      intError(tmpStr);
      return 0;
  }
}

/*********************************************************************/
/* Procedure freeStorage frees all dynamicaly allocated storage used */
/* during oc to cxx translation. It is called to clean up memory     */
/* after translation of a module.                                    */
/*********************************************************************/
void freeStorage(void) {
  int i;
  
  if (nbTypes != 0) {
    for( i= 0; i < nbTypes; i++) {
      free(typeTable[i].name);         /* Free space allocated for type names */
    }
    free(typeTable);                   /* Free table of types */
    typeTable = (typeInfo*)0;
    nbTypes = 0;
  }
  
  if (nbConsts != 0) {
    for(i = 0; i < nbConsts; i++) {
      free(constTable[i].name);        /* Free space allocated for constant names */
    }
    free(constTable);                  /* Free table of constants */
    constTable = (constInfo*)0;
    nbConsts = 0;
  }
  
  if (nbFunctions != 0) {
    for(i = 0; i < nbFunctions; i++) {
      free(funcTable[i].name);         /* Free space allocated for function names */
      free(funcTable[i].argTypes);     /* Free table of argument types */
    }
    free(funcTable);                   /* Free table of functions */
    funcTable = (funcInfo*)0;
    nbFunctions = 0;
  }
  
  if (nbProcedures != 0) {
    for(i = 0; i < nbProcedures; i++) {
      free(procTable[i].name);         /* Free space allocated for procedure names */
      free(procTable[i].refArgTypes);  /* Free table of 'by reference' argument types */
      free(procTable[i].valArgTypes);  /* Free table of 'by value' argument types */
    }
    free(procTable);                   /* Free table of procedures */
    procTable = (procInfo*)0;
    nbProcedures = 0;
  }
  
  if (nbSignals != 0) {
    for(i = 0; i < nbSignals; i++) {
      free(sigTable[i].name);          /* Free space allocated for signal names */
    }
    free(sigTable);                    /* Free table of signals */
    sigTable = (sigInfo*)0;
    nbSignals = 0;
  }
  
  if (nbImplications != 0) {
    free(implTable);                   /* Free table of implications */
    implTable = (implInfo*)0;
    nbImplications = 0;
  }
  
  if (nbExclusions != 0) {
    for(i = 0; i < nbExclusions; i++) {
      free(exclTable[i].sigs);         /* Free table of incompatible signals */
    }
    free(exclTable);                   /* Free table of incompatibilies */
    exclTable = (exclInfo*)0;
    nbExclusions = 0;
  }
  
  if (nbVariables != 0) {
    for(i = 0; i < nbVariables; i++) {
      free(varTable[i].name);          /* Free space allocated for variable names */
    }
    free(varTable);                    /* Free table of variables */
    varTable = (varInfo*)0;
    nbVariables = 0;
  }
  
  if (nbActions != 0) {
    for(i = 0; i < nbActions; i++) {
      switch(actTable[i].type) {
        case aCall:
          freeArgs(actTable[i].body.call.refArgList);   /* Free argument lists */
          freeArgs(actTable[i].body.call.valArgList);
          break;
        case aCombine:
          freeArgNode(actTable[i].body.combine.oneArg); /* Free single argument */
          break;
        case aIf:
          freeArgNode(actTable[i].body.ifAct.oneArg);   /* Free single argument */
          break;
        default:                          /* Other action types : nothing to do */
          break;
      }
    }
    free(actTable);
    actTable = (actInfo*)0;
    nbActions = 0;
  }
}

/*************************************************************/
/* Procedure freeAction frees the storage used by an action. */
/*************************************************************/
void freeAction(actInfo *act) {
  actInfo *prev;
  
  switch(act->type) {
    case aCall:
      freeArgs(act->body.call.refArgList);   /* Free argument lists */
      freeArgs(act->body.call.valArgList);
      break;
    case aCombine:
      freeArgNode(act->body.combine.oneArg); /* Free single argument */
      break;
    case aIf:
      freeArgNode(act->body.ifAct.oneArg);   /* Free single argument */
      break;
    case aSeq:                               /* List of actions */
      freeAction(act->body.sequence.action); /* Free action of first link */
      act = act->body.sequence.next;
      while(act != 0) {                      /* For each link in the list */
        freeAction(act);                     /* free action of link */
        prev = act;
        act = act->body.sequence.next;
        free(prev);                          /* then free the link. */
      }
    default:                          /* Other action types : nothing to do */
      break;
  }
}

/**********************************************************/
/* Procedure freeArgs frees a list of arguments. It calls */
/* freeArgNode to free each argument.                     */
/**********************************************************/
void freeArgs(args* arglist) {
  args *tmp, *old;
  
  tmp = arglist;
  while(tmp != (args*)0) {      /* While the list is not empty */
    old = tmp;                  /* Hold the cell to free */
    tmp = tmp->next;            /* Go to the next cell */
    freeArgNode(old->node);     /* Free the node in the cell */
    free(old);                  /* free the cell */
  }
}

/**********************************************************/
/* Procedure freeArgNode frees a list of arguments. It    */
/* calls freeArgs to free argument list of function calls.*/
/**********************************************************/
void freeArgNode(argNode *node) {
  switch(node->type) {
    case cStringVal:             /* If the node holds a ptrValue */
      free(node->ptrValue);      /* free the data storage */
      break;
    case cFuncVal:               /* If it is a function return value */
      freeArgs(node->arguments); /* free the argument list of the function */
      break;
    default:
      break;
  }
  free(node);                    /* then, free the node */
}
