/* -*-C-*-
 * ###################################################################
 *  occ++ - oc to C++ post-compiler for Esterel
 * 
 *  FILE: "occxxTables.h"
 *                                    created: 03/09/92 00:00:00 
 *                                last update: 26/11/97 18:39:20 
 *  Author: Frdric Boulanger
 *  E-mail: Frederic.Boulanger@supelec.fr
 *    mail: Suplec - Service Informatique
 *          Plateau de Moulon, F-91192 Gif-sur-Yvette cedex
 *     www: http://wwwsi.supelec.fr/fb/fb.html
 *  
 * ========================================================================
 * Permission to use, copy and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that:
 *   - the above copyright notice appears in all copies and,
 *   - both the copyright notice and warranty disclaimer appear in
 *     supporting documentation and,
 *   - any product making use of this software or of parts of this 
 *     software displays the following acknowledgment:
 *       "This product includes software developed by Suplec
 *        and its contributors"
 *     in the documentation or other materials provided with the product 
 *     and in all advertising material mentioning features or use of 
 *     this software.
 *     
 * Warranty disclaimer: this software is provided "as is",
 *   Suplec and Frdric Boulanger disclaim all warranties with regard 
 *   to this software, including all implied warranties of merchantability 
 *   and fitness.
 *   In no event shall Suplec or Frdric Boulanger be liable for any 
 *   special, indirect or consequential damages or any damages whatsoever 
 *   resulting from loss of use, data or profits, whether in an action of 
 *   contract, negligence or other tortuous action, arising out of or in 
 *   connection with the use or performance of this software.
 * ========================================================================
 *  Description: 
 *    Tables and functions used by the oc parser and the C++ code 
 *    generator.
 * 
 *  History
 * 
 *  modified by  rev reason
 *  -------- --- --- -----------
 *  03/09/92 FBO 1.0 original
 *  26/11/97 FBO 4.8 support for C++ types and operators (mangled)
 * ###################################################################
 */

#include "parseTypes.h"


extern int lineNb;                           /* oc source line number */
extern int ocVersion;                        /* Version of oc language */
extern char *moduleName;                     /* Name of parsed module */
extern int rootModule;                       /* Id of root submodule */
extern int nbInstances;                      /* Number of submodules */
extern moduleInfo *moduleTable;              /* Table of modules */

extern int nbTypes;                          /* Number of user defined types */
extern typeInfo *typeTable;                  /* Table of user defined types */

extern int nbConsts;                         /* Number of constants */
extern constInfo *constTable;                /* Table of constants */

extern int nbFunctions;                      /* Number of functions */
extern funcInfo *funcTable;                  /* Table of functions */

extern int nbProcedures;                     /* Number of procedures */
extern procInfo *procTable;                  /* Table of procedures */

extern int nbSignals;                        /* Number of signals */
extern sigInfo *sigTable;                    /* Table of signals */

extern int nbImplications;                   /* Number of implications */
extern implInfo *implTable;                  /* Table of implications */
extern int nbExclusions;                     /* Number of exclusions */
extern exclInfo *exclTable;                  /* Table of exclusions */

extern int nbVariables;                      /* Number of variables */
extern varInfo *varTable;                    /* Table of variables */

extern int nbTasks;                          /* Number of tasks */
extern procInfo *taskTable;                  /* Table of tasks (same struct as procs) */

extern int nbExecs;                          /* Number of execs */
extern execInfo *execTable;                  /* Table of execs */

extern int nbActions;                        /* Number of actions */
extern actInfo *actTable;                    /* Table of actions */
extern int GotoAction;                       /* Code of the Goto action */
extern int ODAGAction;                       /* Code for the execute open dag action */
extern int CDAGAction;                       /* Code for the goto closed dag action */
extern int JumpAction;                       /* Code of the Jump Action */
extern int incrAction;                       /* increment for automaton actions */

extern int nbDags;                           /* Number of dags */
extern stateInfo *dagTable;                  /* Table of dags */

extern int nbStates;                         /* Number of states */
extern stateInfo *stateTable;                /* Table of states */
extern int initState;                        /* Initial state */
extern int hasAwaitedInfo;                   /* Awaited signals info present? */

#define nbSysTypes 5                         /* Number of predefined Esterel types */
#define nbSysConsts 2                        /* Number of predefined Esterel constants */
#define nbSysFuncs 52                        /* Number of predefined functions */
#define nbSysProcs 5                         /* Number of predefined procedures */
#define nbOperators 22                       /* Number of operators */
extern typeInfo sysTypeTable[nbSysTypes];    /* Table of predefined types */
extern constInfo sysConstTable[nbSysConsts]; /* Table of predefined constants */
extern funcInfo* sysFuncTable[nbSysFuncs];   /* Table of predefined functions */
extern procInfo* sysProcTable[nbSysProcs];   /* Table of predefined procedures */
extern char *tagName[nbOperators];           /* Tag names */
extern char *opName[nbOperators];            /* Corresponding operator names */
extern char *cPlusTagName[];                 /* C++ operators Tag names */
extern char *cPlusOpName[];                  /* Corresponding C++ operator names */
extern int cPlusArity[];                     /* Corresponding C++ operator arities */



/*************************************************************/
/* See occxxTables.c for comments on the following functions */
/*************************************************************/

void initInstances(int nb);
void declareModule(int modID, char *modName, char *srcFile);
void initTypes(int nb);
typeInfo *newType(int assignP, int equalF, int diffF, int condF);
void declareType(int typeID, char *typeName, typeInfo *theInfo);
void initConstants(int nb);
void declareConstant(int constID, int constType, char *constName);
void initFunctions(int nb);
argInfo *addArgType(argInfo *arglist,int theType);
void declareFunction(int funcID, char *funcName, argInfo *arglist, int retType);
void initProcedures(int nb);
void declareProcedure(int procID, char *procName, argInfo *refArgList, argInfo *valArgList);
void initSignals(int nb);
sigInfo *newSignal(sigType theType);
void declareSignal(int sigID, sigInfo *sInfo);
void initImplications(int nb);
void declareImplication(int implId, int cause, int effect);
void initExclusions(int nb);
sigList *addSignal(sigList *theList, int sigNum);
void declareExclusion(int exclId, sigList *theList);
void initVariables(int nb);
varInfo *newVariable(varKind theKind);
char *nextVarCntName(void);
void declareVariable(int varId, varInfo *vInfo);
void declareNewVariable(int varId, int typeId, varKind theKind);

void checkVarsAndSigs(void);

void initTasks(int nb);
void declareTask(int taskID, char *taskName, argInfo *refArglist, argInfo *valArglist);
void initExecs(int nb);
void declareExec(int execID, int taskID, int retSigID, args *refArgs, args *valArgs);

void initActions(int nb);
void declareAction(int actId, actInfo *theAct);
actInfo *newAction(actType theType);
void appendToActionSeq(actInfo *list, actInfo *theAct);
argNode *newArgNode(argType theType);
args *addArgument(args *arglist, argNode *n);
void initDags(int nb);
void declareDag(int dagId, stateStep *theSteps, sigList *awaited);
void initStates(int nb);
void setStartState(int stateId);
void declareState(int stateId, stateStep *theSteps, sigList *awaited);
stateStep *addStep(stateStep *list, stateStep *elem);
stateStep *newStateStep(stepType theType);
void compileState(stateStep *theSteps, stateInfo *sInfo);
void compileSteps(stateStep *stp, int **tab);
int stepLength(stateStep *s);
void freeSteps(stateStep *theSteps);
void freeStorage(void);
void freeArgs(args *arglist);
void freeArgNode(argNode *node);
