/* -*-C-*-
 * ###################################################################
 *  occ++ - oc to C++ post-compiler for Esterel
 * 
 *  FILE: "parseTypes.h"
 *                                    created: 03/09/92 00:00:00
 *                                last update: 26/11/97 18:43:54 
 *  Author: Frdric Boulanger
 *  E-mail: Frederic.Boulanger@supelec.fr
 *    mail: Suplec - Service Informatique
 *          Plateau de Moulon, F-91192 Gif-sur-Yvette cedex
 *     www: http://wwwsi.supelec.fr/fb/fb.html
 *  
 * ========================================================================
 * Permission to use, copy and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that:
 *   - the above copyright notice appears in all copies and,
 *   - both the copyright notice and warranty disclaimer appear in
 *     supporting documentation and,
 *   - any product making use of this software or of parts of this 
 *     software displays the following acknowledgment:
 *       "This product includes software developed by Suplec
 *        and its contributors"
 *     in the documentation or other materials provided with the product 
 *     and in all advertising material mentioning features or use of 
 *     this software.
 *     
 * Warranty disclaimer: this software is provided "as is",
 *   Suplec and Frdric Boulanger disclaim all warranties with regard 
 *   to this software, including all implied warranties of merchantability 
 *   and fitness.
 *   In no event shall Suplec or Frdric Boulanger be liable for any 
 *   special, indirect or consequential damages or any damages whatsoever 
 *   resulting from loss of use, data or profits, whether in an action of 
 *   contract, negligence or other tortuous action, arising out of or in 
 *   connection with the use or performance of this software.
 * ========================================================================
 *  Description: 
 *    Type definitions for oc parsing.
 * 
 *  History
 * 
 *  modified by  rev reason
 *  -------- --- --- -----------
 *  03/09/92 FBO 1.0 original
 *  26/11/97 FBO 4.8 support for C++ types and operators (mangled)
 * ###################################################################
 */

typedef struct _argInfo {        /* Structure of lists of argument types */
  int type;                      /* ID of the type or class of the argument */
  struct _argInfo *next;         /* Pointer to next cell in list */
} argInfo;

typedef struct {                 /* Structure for module info in parse tables */
  char *name;                    /* Name of the module */
  char *srcfile;                 /* Source file of the module */
} moduleInfo;

typedef struct {                 /* Structure for type info in parse tables */
  char *name;                    /* name of the type or class */
  int assignProc;                /* Assignement procedure (not useful for C++)*/
  int equalFunc;                 /* Equality test function */
  int diffFunc;                  /* Non-equality test function */
  int condFunc;                  /* Condition(bool, type, type)->type returns second */
                                 /* arg if first arg is true, else returns third arg. */
} typeInfo;

typedef struct {                 /* Structure for const info in parse tables */
  int type;                      /* ID of the type or class of the constant */
  char *name;                    /* Name of the constant */
} constInfo;

typedef struct {                 /* Struct for function info in parse tables */
  int resType;                   /* ID of return value type */
  char *name;                    /* name of function */
  int nbArgs;                    /* Number of arguments */
  int *argTypes;                 /* Array of IDs for the arguments type */
} funcInfo;

typedef struct {                 /* Struct for procedure info in parse tables */
  char *name;                    /* name of procedure */
  int nbRefArgs;                 /* Number of arguments passed by reference */
  int nbValArgs;                 /* Number of arguments passed by value */
  int *refArgTypes;              /* Array of the IDs of by reference arg type */
  int *valArgTypes;              /* Array of the IDs of by value args type */
} procInfo;

typedef enum {sInput, sInputOutput,sOutput,sSensor,sLocal,sException,sReturn} sigKind;
                                 /* Kinds of signal */
             
typedef enum {sPure,sSingle,sMultiple} sigType;
                                 /* Types of signal */

typedef struct {                 /* Structure for signal info in parse tables */
  sigKind kind;                  /* Kind of the signal : input, output, ... */
  sigType type;                  /* Type of the signal : single, multiple ... */
  char *name;                    /* name of the signal */
  int varId;                     /* ID of var holding the signal's value */
  int boolId;                    /* ID of var holding the signal's boolean */
  int combFunc;                  /* ID of func used to combine simultaneous */
                                 /* values for this signal */
} sigInfo;

typedef struct _sigList {        /* Structure for parsing lists of signals */
  int sigNum;                    /* ID of the signal */
  struct _sigList *next;         /* Pointer to next cell in list */
} sigList;

typedef struct {                 /* Structure for info on implication rules */
  int cause;                     /* If signal 'cause' is present, signal */
  int effect;                    /* 'effect' must be present too */
} implInfo;

typedef struct {                 /* Structure for info on exclusion rules */
  int nbSigs;                    /* Number of signals in the rule */
  int *sigs;                     /* Array of the IDs of signals that can't */
                                 /* be simultaneously present */
} exclInfo;

typedef enum {
  cVarVal, cSysVal, cConstVal, 
  cSysConstVal, cImmedVal, 
  cStringVal, cSigVal, cFuncVal
} argType;
                                 /* Types of arguments : variable, predefined */
                                 /* value, constant, predefined constant, */
                                 /* immediate value, string, signal value */
                                 /* or return value of a function. */
              
struct _args;                    /* struct _args will be defined later */

typedef struct {                 /* Structure for an argument to a call */
  argType type;                  /* Type of the argument : const, var etc... */
  int value;                     /* Value for an immediate value, */
                                 /* ID for a function call,... */
  void *ptrValue;                /* Pointer to data, for instance for strings */
  struct _args *arguments;       /* List of args if result of a function call */
} argNode;

typedef struct _args {           /* Structure for lists of arguments */
  argNode *node;                 /* The current argument */
  struct _args *next;            /* Pointer to next cell in list */
} args;

typedef enum {
  vSigBool, vSigVal, vVarVal, vVarCnt,
  vVarWire, vVarRegister, vVarUnknown
} varKind;
                                 /* Kinds of variables : boolean of a signal */
                                 /* value of a signal */
                                 /* user variable, internal counter,. */
                                 /* boolean circuit wire or boolean circuit */
                                 /* register. */

typedef struct {                 /* Structure for information on variables */
  varKind kind;                  /* Kind of the variable */
  int type;                      /* ID of the type or class of the variable */
  char *name;                    /* Name of the variable */
  int sigNum;                    /* ID of the signal this variable is bound to*/
  argNode *initValue;            /* Initial value (for oc5 registers) */
} varInfo;

typedef struct  {                /* Structure for info on execs */
  int task;                      /* Task id */
  int retsig;                    /* Return signal id */
  args *refArgList;              /* Reference argument list */
  args *valArgList;              /* Value argument list */
} execInfo;

typedef enum {
  aGoto, aAct, aPresent, aOutput,
  aCall, aCombine, aReset,
  aIf, aDsz, aSeq,
  aStart, aKill, aSuspend,
  aActivate, aReturn
} actType;
                                 /* Types of actions : state transition, other action call, */
                                 /* test for the presence of a signal, output of a */
                                 /* signal, procedure call, combination of 2 */
                                 /* values for a signal, reset of a signal, */
                                 /* conditional action, decr and jump if zero, */
                                 /* or sequence of actions. */

typedef struct {                 /* Struct for a simple action: goto or reset */
  int sigNum;                    /* ID of signal, or state number for a goto */
} simpleAction;

typedef struct {                 /* Structure for a procedure call */
  int procNum;                   /* ID of the procedure */
  args *refArgList;              /* List of by reference arguments */
  args *valArgList;              /* List of by value arguments */
} callAction;

typedef struct {                 /* Structure for the combination of values */
  int sigNum;                    /* Id of the signal */
  argNode *oneArg;               /* Argument to the combination function */
}combineAction;

typedef struct {                 /* Structure for a conditional action */
  argNode *oneArg;               /* Condition (always yields a boolean) */
}ifAction;

struct _actInfo;

typedef struct {
  struct _actInfo *action;
  struct _actInfo *next;
}seqAction;

typedef struct _actInfo {        /* Structure used to represent any action */
  actType type;                  /* Type of the action : goto, reset, if ... */
  union {                        /* Union of the structures seen above */
    simpleAction simple;
    callAction call;
    combineAction combine;
    ifAction ifAct;
    seqAction sequence;
  } body;
} actInfo;

typedef enum {seqStep, condStep, gotoStep, cdagStep, odagStep} stepType;
                                 /* Types of steps to do in a state */

typedef struct _stateStep {      /* Structure for a step in a state */
  stepType type;                 /* Type of step : sequential or conditional */
  int actNum;                    /* ID of the action to perform in this step */
  struct _stateStep *next;       /* Pointer to next step (sequential step) */
  struct _stateStep *thenBranch; /* Pointer to the 'then' branch (condStep) */
  struct _stateStep *elseBranch; /* Pointer to the 'else' branch (condStep) */
  int thenLength;                /* Number of steps in the 'then' branch */
  int elseLength;                /* Number of steps in the 'else' branch */
} stateStep;                     /* This structure is used to create the */
                                 /* action tables of the state machine */
                                 /* implementing the synchronous module. */

typedef struct _stateInfo {      /* Structure for a state of the automaton */
  int length;                    /* Number of steps in the state */
  int isOpenDAG;                 /* Tells wether the dag is open or not (for dags) */
  int *actions;                  /* Array of the actions IDs */
  struct _stateInfo *next;       /* Link to next for oc2 (does not tell the state number) */
  int nbAwaited;                 /* Number of awaited signals */
  int *awaited;                  /* Array of awaited signal */
} stateInfo;

