#include "tasksDef.H"

// This task merely computes its result in the tstart method, so it should
// work on all systems. It just serves as an example of what you could write
// to build a correct task class for occ++.

tasksTask_add::tasksTask_add(unsigned taskNum, unsigned execNum,
                             const char *name)
             : tkNum(taskNum), exNum(execNum),
               hasReturned(false),
               suspended(false),
               killed(false),
               running(false),
               fReturn(0,"return") {
}

void tasksTask_add::tstart(integer& x, integer a, integer b) {
  running = true;
  suspended = false;
  killed = false;
  hasReturned = false;
  
  ref = &x;           // Store the reference argument
  res = x;            // Make o copy of it to work on
  
  res = a + b;        // Compute the result of the task
  hasReturned = true; // The task is finished

  // Debugging stuff
  cerr << "start of (" << tkNum << "," << exNum << "): (" << x << "," << a << "," << b << ")" << endl;
}

void tasksTask_add::tkill() {       // Nothing to do
  if (running)
    killed = true;                  // Just rember we've been killed...
}

void tasksTask_add::tsuspend() {    // Nothing to do
  if (running)
    suspended = true;               // Just remember we are suspended
}

void tasksTask_add::tactivate() {   // Nothing to do
  if (running)
    suspended = false;              // Just stop remembering we're suspended
}

void tasksTask_add::treturn() {
  *ref = res;           // Update reference argument
  hasReturned = false;  // No need to remember about task completion any longer
  running = false;
  // Debugging stuff
  cerr << "doReturn (" << tkNum << "," << exNum << "). res=" << res << ", *ref =" << *ref << endl;
}

OutSignal<integer>& tasksTask_add::returnSig() {  // Just return our return signal
  return fReturn;
}

void tasksTask_add::setReturn() { // If we are running, have returned, are not suspended
	if (running && hasReturned && !suspended && !killed) { // and not killed, then
	  fReturn.value() = res;        // emit the return signal
	} else
	  fReturn.value().reset();      // Else, just say nothing.
}

