#include "tasksDef.H"
#include <signal.h>
#include <sys/wait.h>
#include <unistd.h>
#include <sys/stat.h>

// This task is an example of use of the exec primtive to control of forked
// Unix process. It just yields the number of links to its argument.
tasksTask_countlinks::tasksTask_countlinks(unsigned taskNum, unsigned execNum,
                             const char *name)
             : tkNum(taskNum), exNum(execNum),
               hasReturned(false),
               suspended(false),
               killed(false),
               running(false),
               fReturn(0,"return") {
}

void tasksTask_countlinks::do_it(SString dir) { // Do the real job
	struct stat s;
	if (stat(dir,&s) == 0)
    _exit(s.st_nlink);  // Get number of links from stat structure
  else
  	_exit(-1);
}

void tasksTask_countlinks::tstart(integer& x, SString a) {
  ref = &x;           // Store the reference argument
  res = x;            // Make o copy of it to work on
  
  process = fork();
  if (process != 0) { // Father: go on with the main program
    running = true;
    suspended = false;
    killed = false;
    hasReturned = false;
    // Debugging stuff
    cerr << "start of (" << tkNum << "," << exNum;
    cerr << "): (" << x << "," << a << "). PID=" << process << endl;
  } else {           // Child: do what the task should do
    do_it(a);
  }
}

void tasksTask_countlinks::tkill() {
  if (running) {
    killed = true;                  // rember we've been killed...
  	kill(process,SIGKILL);          // attempt to kill the real process
  }
}

void tasksTask_countlinks::tsuspend() {    // Nothing to do
  if (running)
    suspended = true;               // Just remember we are suspended
}

void tasksTask_countlinks::tactivate() {   // Nothing to do
  if (running)
    suspended = false;              // Just stop remembering we're suspended
}

void tasksTask_countlinks::treturn() {
  *ref = res;           // Update reference argument
  hasReturned = false;  // No need to remember about task completion any longer
  running = false;
  // Debugging stuff
  cerr << "doReturn (" << tkNum << "," << exNum;
  cerr << "). res=" << res << ", *ref =" << *ref << endl;
}

PureOutSignal& tasksTask_countlinks::returnSig() {  // Just return our return signal
  return fReturn;
}

void tasksTask_countlinks::setReturn() {
	int status;
	
	if (waitpid(process,&status,WNOHANG) == process) {
	  hasReturned = true;
	  res = (signed char)(WEXITSTATUS(status));
	}
	
	if (running && hasReturned && !suspended && !killed) { // and not killed, then
	  fReturn.value() = true;       // emit the return signal
	} else
	  fReturn.value().reset();      // Else, just say nothing.
}

